eval '(exit $?0)' && eval 'exec perl -S $0 ${1+"$@"}'
  & eval 'exec perl -S $0 $argv:q'
  if 0;

#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2009 - 2014.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#
##############################################################################

=head1 NAME

cqtsadmin.pl

=head1 SYNOPSIS

B<cqtsadmin.pl> B<[-man]> B<[-help]>

=head1 OPTIONS

=over 4

=item -help

Print a brief help message.

=item -man

Print the man page.

=back

=head1 DESCRIPTION

=head1 AUTHOR

=cut


use CQPerlExt;

use CQTsAdmin::L10N;

use strict;
use warnings;
use IO::File;
use XML::Parser;
use HTML::Entities;
use File::Copy;
use File::Find;
use File::Path;
use File::Spec;
use Getopt::Long;
use Sys::Hostname;
use Data::Dumper;
use Encode;
use LWP::UserAgent;


my $langMsgText = CQTsAdmin::L10N->get_handle()
  || die "Cannot find an acceptable language module!?!";

# replace PERL's print function so it can log as well
my $logFileHandler;				# the file handle to the log file
my $cqtsadminLogFile = "";		# the log file name
my $logDataBacklog = "";		# holds log data in memory till when the log file is ready

sub printAndLog;
sub openLogger;
sub closeLogger;


require "CQPerlExt.pm" || die;

# get localized message strings
my $stat_START = $langMsgText->maketext("START_MSG_TEXT");
my $stat_DONE = $langMsgText->maketext("DONE_MSG_TEXT");

my $msgLabel_INFO = $langMsgText->maketext("INFO_MSG_TEXT");
my $msgLabel_WARN = $langMsgText->maketext("WARN_MSG_TEXT");
my $msgLable_ERROR = $langMsgText->maketext("ERROR_MSG_TEXT");
my $msgLable_oS_ERROR = $langMsgText->maketext("ERROR_MSG_TEXT_OS");

my $ENABLE_CQWEB_FTS = 1;
my $DISABLE_CQWEB_FTS = 0;

my $gVersion_cqtsadmin = "2.1.0.00-D140614";

my $gSolrVersion_120 = "1.2.0";	# Using Solr 1.2.0 with CQ prior to 8.0.0.5
my $gSolrVersion_361 = "3.6.1";	# Using Solr 3.6.1 with CQ 8.0.0.5 and later

my %gCQSolrVersionInfo =
	(
		# SolrVersion => [ SolrTemplateLocation, SolrWARFileLocation, [ListOfFromSolrVerWillRequireReIndex] ]
		$gSolrVersion_120 => ["CQ.Solr/SAMPL", "cqweb/apps/apache-solr-".$gSolrVersion_120.".war", []], 
		$gSolrVersion_361 => ["CQ.Solr_".$gSolrVersion_361."/SAMPL", "cqweb/apps/apache-solr-".$gSolrVersion_361.".war", [$gSolrVersion_120]], 
	);

my $gSolrUpgradeToVersion = $gSolrVersion_361;	# The Solr version to upgrade to


my $gAboutThisFTSFile = "";

my $gDieKind_must_match = "must-match";
my $gDieKind_must_not_match = "must-not-match";
my $gDieKind_none = "none";
my $gTerminationErrorMsg = $msgLable_ERROR."*************** ".$langMsgText->maketext("TERMINATION_MSG")." ****************";
my $gMaxWASProfileLen = length("123456789012345678901234567");

my $gExternalCmdMarker = $ENV{"CQ_FTS_XCMD_MARKER"};						# Not documented
my $gSuppressExternalCmdOutput = $ENV{"CQ_FTS_SUPPRESS_XCMD_OUTPUT"};   	# Not documented, must be set to "yes" to take effect
my $gAllow64BitWAS = $ENV{"CQ_FTS_ALLOW_64BIT_WAS"};						# Not documented, must be set to "true" to allow 64Bit WAS
if (!defined($gAllow64BitWAS))
{
	$gAllow64BitWAS = "false";
}						

my $gCqFtsEncoding = $ENV{"CQ_FTS_ENCODING"};
if (!defined($gCqFtsEncoding))
{
	$gCqFtsEncoding = $langMsgText->maketext("CODEPAGE_ENCODING");
}

my $gIsReplicated = 0;
my $gTypeaheadCmd = "";
my $gWinServiceStartUpType = "";
my $gBackupDestinationDir = "";
my $gScrubOplogBeforeDate = "";
my $gPrepUpgdFtsWASProfiles = "";
my $gRestoreFtsWASProfiles = "";
my $gShowScenarios = "";
my $gShowScenariosOption = 0;
my $gCaptureExeOutputData = "";
my $gShowMustReIndexMsg = 0;
my $gInitCqFTSOption = 0;
my $gGenerateEntityFileOption = 0;
my $gGenerateFTSFilesOption = 0;
my $gIncludeUsersGroups = 0;   

my @gCmdLineArgvList;
my @gCmdToExecute;
my %gDispOrdFieldList = ();

my $gTopLevelSate = "";
my %gFtsScenarios = ();

my $gVerboseOutput = 1;		# for this release, we are always going to be verbose

my $optUserName = "";						# --username=s						[takes a string]
my $optPassword = "";						# --password=s						[takes a string]
my $optDbSet = "";							# --dbset=s							[takes a string]
my $optuserDB = "";							# --userdb=s						[takes a string]
my $optCqFtsHomeDir = "";					# --ftshome=s						[takes a string]
my $optInitCqFTS = "";						# --init_cq_fts
my $optSetupCqFTS = "";						# --setup_cq_fts
my $optCopyFTSTemplate = "";				# --copy_fts_template
my $optGenerateEntityFile = "";				# --gen_entity_file
my $optGenerateFTSFiles = "";				# --gen_fts_files
my $optCustomizeCQSearchPropertyFile = "";	# --cust_fts_files
my $optCustomizeSolrFile = "";				# --cust_solr_files
my $optStopWASSearchProfile = "";			# --stop_fts_was_profile
my $optSetSolrHome = "";					# --set_solr_home
my $optStartWASSearchProfile = "";			# --start_fts_was_profile
my $optRunInstallUtil = "";					# --enable_cqweb_fts
my $optRunBatchIndex = "";					# --run_batch_idx
my $optEnableUpdateIndex = "";				# --start_update_idx
my $optUpdateGenFiles = "";					# --update_fts_prop_files
my $optDisableFTS = "";						# --disable_cqweb_fts
my $optSetWASMaxMem = "";					# --set_was_max_mem
my $optStopUpdateIndex = "";				# --stop_update_idx
my $optFreshRunBatchIndex = "";				# --fresh_batch_idx
my $optFreshRunUpdateIndex = "";			# --fresh_update_idx
my $optAddRecordType = ""; 					# --add_record_type
my $optRemoveRecordType = "";				# --remove_record_type
my $optCreateCqFtsWasProfile = "";			# --create_fts_was_profile=s		[takes a string]
my $optFtsWasProfileHome = "";				# --fts_was_profile_home=s			[takes a string]
my $optFtsWasProfPortsFile = "";			# --was_profile_ports_file=s		[takes a string]
my $optDeleteCqFtsWasProfile = "";			# --delete_fts_was_profile
my $optOptimizeIndex = "";					# --optimize_idx
my $optRemoveLuceneIdxLock = "";			# --remove_lucene_idx_lock
my $optRepairRecords = "";					# --repair_records
my $optScrubOplog = "";						# --scrub_oplog
my $optBackupCqFTS = "";					# --backup_fts=s					[takes a string]
my $optArchiveCqFTS = "";					# --archive_fts
my $optClearState = "";						# --clear_state
my $optLockCqFts = "";						# --lock_cq_fts
my $optUnlockCqFts = "";					# --unlock_cq_fts
my $optValidateCqFtsDeplo = "";				# --validate_cq_fts_deplo
my $optPrepUpgdFtsWASProfiles = "";			# --prep_upgd_was_profiles=s		[takes a string]
my $optRestoreFtsWASProfiles = "";			# --restore_was_profiles=s			[takes a string]
my $optUpgradeSolrApp = "";					# --upgrade_solr_app
my $optTypeaheadCmd = "";					# --typeahead_cmd=s					[takes a string]
my $optGatherDiagnosticData = "";			# --gather_diagnostic_data
my $optShowScenarios = "";					# --show_scenarios=s				[takes a string]
my $optVerbose = "";						# --verbose
my $optHelp = "";							# --help

# get the order of the comand line options
foreach my $argnum (0 .. $#ARGV)
{
	if ($ARGV[$argnum] =~ m/^-/)
	{
		my $tmp = $ARGV[$argnum];
		$tmp =~ s/^-//;
		$tmp =~ s/^-//;
		
		if ($tmp eq "show_scenarios")
		{
			$gShowScenariosOption = 1;
			next;
		}
		if ($tmp eq "verbose")
		{
			next;
		}
		if ($tmp eq "init_cq_fts")
		{
			$gInitCqFTSOption = 1;
		}
		# If run 'gen_entity_file' and 'gen_fts_files' separately, also need to add IncludeUsersGroups to the commands.
		if ($tmp eq "gen_entity_file")
		{
			$gGenerateEntityFileOption = 1;
		}	
		if ($tmp eq "gen_fts_files")
		{
			$gGenerateFTSFilesOption = 1;
		}
			
		push(@gCmdLineArgvList, $tmp);
	}
}

# Get flag values
my $result = GetOptions(
						 "username=s" => \$optUserName, 
						 "password=s" => \$optPassword, 
						 "dbset=s" => \$optDbSet, 
						 "userdb=s" => \$optuserDB, 
						 "ftshome=s" => \$optCqFtsHomeDir, 
						 
						 "init_cq_fts:s" => \$optInitCqFTS, # The value is "IncludeUsersGroups" or ""
						 "setup_cq_fts" => \$optSetupCqFTS, 
						 
						 "copy_fts_template" => \$optCopyFTSTemplate, 
						 "gen_entity_file:s" => \$optGenerateEntityFile, # The value is "IncludeUsersGroups" or ""
						 
						 "gen_fts_files:s" => \$optGenerateFTSFiles, # The value is "IncludeUsersGroups" or ""
						 "cust_fts_files" => \$optCustomizeCQSearchPropertyFile, 
						 "cust_solr_files" => \$optCustomizeSolrFile, 
						 "stop_fts_was_profile" => \$optStopWASSearchProfile, 
						 "set_solr_home" => \$optSetSolrHome, 
						 "start_fts_was_profile" => \$optStartWASSearchProfile, 
						 "enable_cqweb_fts" => \$optRunInstallUtil, 
						 "run_batch_idx" => \$optRunBatchIndex, 
						 "start_update_idx" => \$optEnableUpdateIndex, 
						 
						 "update_fts_prop_files" => \$optUpdateGenFiles, 
						 "disable_cqweb_fts" => \$optDisableFTS, 
						 "set_was_max_mem" => \$optSetWASMaxMem, 
						 "stop_update_idx" => \$optStopUpdateIndex, 
						 "fresh_batch_idx" => \$optFreshRunBatchIndex, 
						 "fresh_update_idx" => \$optFreshRunUpdateIndex, 
						 
						 "add_record_type" => \$optAddRecordType, 
						 "remove_record_type" => \$optRemoveRecordType, 
						 
						 "create_fts_was_profile=s" => \$optCreateCqFtsWasProfile, 
 						 "fts_was_profile_home=s" => \$optFtsWasProfileHome, 
 						 "was_profile_ports_file=s" => \$optFtsWasProfPortsFile, 
						 "delete_fts_was_profile" => \$optDeleteCqFtsWasProfile, 
						 
						 "optimize_idx" => \$optOptimizeIndex, 
						 "remove_lucene_idx_lock" => \$optRemoveLuceneIdxLock, 
						 "repair_records" => \$optRepairRecords, 
						 "scrub_oplog=s" => \$optScrubOplog,
						 "backup_fts=s" => \$optBackupCqFTS, 
						 "archive_fts" => \$optArchiveCqFTS, 
						 
						 "clear_state" => \$optClearState,
						 
						 "lock_cq_fts" => \$optLockCqFts, 
						 "unlock_cq_fts" => \$optUnlockCqFts, 
						 
						 "validate_cq_fts_deplo" => \$optValidateCqFtsDeplo, 
						 
						 "prep_upgd_was_profiles=s" => \$optPrepUpgdFtsWASProfiles, 
						 "restore_was_profiles=s" => \$optRestoreFtsWASProfiles, 
						 
						 "upgrade_solr_app" => \$optUpgradeSolrApp, 
						 
						 "typeahead_cmd=s" => \$optTypeaheadCmd, 
						 
						 "gather_diagnostic_data" => \$optGatherDiagnosticData, 
						 
						 "show_scenarios:s" => \$optShowScenarios, 
						 
						 "verbose" => \$optVerbose, 
						 
						 "help" => \$optHelp
						);


# they are asking for help, show it and then exit
if ($optHelp eq "1")
{
	Usage("", 1);
}

# they are asking for scenarios, show it and then exit
if (($optShowScenarios ne "") or ($gShowScenariosOption eq 1))
{
	$gShowScenarios = $optShowScenarios;
	$optShowScenarios = "1";
	Func_show_scenarios();
	exit 1;
}

# deal with the commands "--init_cq_fts", "--gen_entity_file" and "--gen_fts_files" and their parameters
if ($gInitCqFTSOption eq 1)
{
	if (lc($optInitCqFTS) eq "includeusersgroups")
	{
		$gIncludeUsersGroups = 1;
	}
	$optInitCqFTS = "1";
}
else
{
	$optInitCqFTS = "";
}
if ($gGenerateEntityFileOption eq 1)
{
	if (lc($optGenerateEntityFile) eq "includeusersgroups")
	{
		$gIncludeUsersGroups = 1;
	}
	$optGenerateEntityFile = "1";
}
else
{
	$optGenerateEntityFile = "";
}
if ($gGenerateFTSFilesOption eq 1)
{
	if (lc($optGenerateFTSFiles) eq "includeusersgroups")
	{
		$gIncludeUsersGroups = 1;
	}
	$optGenerateFTSFiles = "1";
}
else
{
	$optGenerateFTSFiles = "";
}

# do they want verbose output?
if ($optVerbose ne "")
{
	Func_verbose();
	$optVerbose = "";
}


# verify they used proper command line options (i.e.: they didn't just pass in no option(s))
if ($result eq "")
{
	my $dieMsg1 = $langMsgText->maketext("INVALID_COMMAND_LINE_OPTION");
	my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
	my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
	dealWithError($dieMsg, 1);
}
else
{
	my $haveRequiredOptions = 0;
	my $haveAtLeastOneOption = 0;
	
	# make sure the min. required command-line options and at least one additional option is used
	foreach my $item (@gCmdLineArgvList)
	{
		if (($item eq "username") or ($item eq "password") or
			($item eq "dbset") or ($item eq "userdb") or
			($item eq "ftshome"))
		{
			$haveRequiredOptions++;
		}
		else
		{
			$haveAtLeastOneOption++;
		}
	}
	
	# are all required command-line options given?
	if ($haveRequiredOptions ne 5)
	{
		my $dieMsg1 = $langMsgText->maketext("MISSING_REQUIRED_OPTIONS");
		my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
		my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
		dealWithError($dieMsg, 1);
	}
	
	# is at least one additional command-line option given?
	if ($haveAtLeastOneOption eq 0)
	{
		my $dieMsg1 = $langMsgText->maketext("NO_OPERATIONAL_COMMAND_OPTION");
		my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
		my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
		dealWithError($dieMsg, 1);
	}
}

# deal with some exceptions with the command line option:
# if init_cq_fts, or setup_cq_fts is used, nothing else is allowed
if (($optInitCqFTS eq "1") or ($optSetupCqFTS eq "1"))
{
	foreach my $item (@gCmdLineArgvList)
	{
		my $foundAt = index($item, '=');	# if the command takes arg, remove the the arg part

		if ($foundAt ne -1)
		{
			$item = substr($item, 0, $foundAt);
		}
		if (($item eq "username") or ($item eq "password") or
			($item eq "dbset") or ($item eq "userdb") or
			($item eq "ftshome") or ($item eq "init_cq_fts") or
			($item eq "setup_cq_fts") or ($item eq "create_fts_was_profile") or
			($item eq "fts_was_profile_home") or ($item eq "was_profile_ports_file"))
		{
			next;
		}
		else
		{
			my $dieMsg1 = $langMsgText->maketext("NOT_SUPPORTED_COMMAND_OPTION");
			my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
			my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
			dealWithError($dieMsg, 1);
		}
	}
}

# deal with command line options that accept parameters
if ($optTypeaheadCmd ne "")
{
	# deal with: --typeahead_cmd
	$gTypeaheadCmd = $optTypeaheadCmd;
	$optTypeaheadCmd = "1";
}
if ($optCreateCqFtsWasProfile ne "")
{
	# deal with: --create_fts_was_profile
	$optCreateCqFtsWasProfile = lc($optCreateCqFtsWasProfile);
	if (($optCreateCqFtsWasProfile eq "manual") or 
		($optCreateCqFtsWasProfile eq "automatic") or 
		($optCreateCqFtsWasProfile eq "disabled"))
	{
		$gWinServiceStartUpType = $optCreateCqFtsWasProfile;
		$optCreateCqFtsWasProfile = "1";
	}
	else
	{
		my $dieMsg1 = $langMsgText->maketext("INVALID_CREATE_PROFILE");
		my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
		my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
		dealWithError($dieMsg, 1);
	}
}
if ($optBackupCqFTS ne "")
{
	$gBackupDestinationDir = $optBackupCqFTS;
	$optBackupCqFTS = "1";
}
if ($optScrubOplog ne "")
{
	$gScrubOplogBeforeDate = $optScrubOplog;
	$optScrubOplog = "1";
}
if ($optPrepUpgdFtsWASProfiles ne "")
{
	$gPrepUpgdFtsWASProfiles = $optPrepUpgdFtsWASProfiles;
	$optPrepUpgdFtsWASProfiles = "1";
}
if ($optRestoreFtsWASProfiles ne "")
{
	$gRestoreFtsWASProfiles = $optRestoreFtsWASProfiles;
	$optRestoreFtsWASProfiles = "1";
}


# Print command-line value
printAndLog ""."\n";
printAndLog ""."\n";
printAndLog "username ->               : $optUserName"."\n";
printAndLog "password ->               : ********"."\n";
printAndLog "dbset ->                  : $optDbSet"."\n";
printAndLog "userdb ->                 : $optuserDB"."\n";
printAndLog "ftshome ->                : $optCqFtsHomeDir"."\n";
printAndLog "init_cq_fts ->            : $optInitCqFTS"."\n";
printAndLog "setup_cq_fts ->           : $optSetupCqFTS"."\n";
printAndLog "copy_fts_template ->      : $optCopyFTSTemplate"."\n";
printAndLog "gen_entity_file ->        : $optGenerateEntityFile"."\n";
printAndLog "gen_fts_files ->          : $optGenerateFTSFiles"."\n";
printAndLog "cust_fts_files ->         : $optCustomizeCQSearchPropertyFile"."\n";
printAndLog "cust_solr_files ->        : $optCustomizeSolrFile"."\n";
printAndLog "stop_fts_was_profile ->   : $optStopWASSearchProfile"."\n";
printAndLog "set_solr_home ->          : $optSetSolrHome"."\n";
printAndLog "start_fts_was_profile ->  : $optStartWASSearchProfile"."\n";
printAndLog "enable_cqweb_fts ->       : $optRunInstallUtil"."\n";
printAndLog "run_batch_idx ->          : $optRunBatchIndex"."\n";
printAndLog "start_update_idx ->       : $optEnableUpdateIndex"."\n";
printAndLog "update_fts_prop_files ->  : $optUpdateGenFiles"."\n";
printAndLog "disable_cqweb_fts ->      : $optDisableFTS"."\n";
printAndLog "set_was_max_mem ->        : $optSetWASMaxMem"."\n";
printAndLog "stop_update_idx ->        : $optStopUpdateIndex"."\n";
printAndLog "fresh_batch_idx ->        : $optFreshRunBatchIndex"."\n";
printAndLog "fresh_update_idx ->       : $optFreshRunUpdateIndex"."\n";
printAndLog "add_record_type ->        : $optAddRecordType"."\n";
printAndLog "remove_record_type ->     : $optRemoveRecordType"."\n";
printAndLog "create_fts_was_profile -> : $optCreateCqFtsWasProfile"."\n";
printAndLog "fts_was_profile_home ->   : $optFtsWasProfileHome"."\n";
printAndLog "was_profile_ports_file -> : $optFtsWasProfPortsFile"."\n";
printAndLog "delete_fts_was_profile -> : $optDeleteCqFtsWasProfile"."\n";
printAndLog "optimize_idx ->           : $optOptimizeIndex"."\n";
printAndLog "remove_lucene_idx_lock -> : $optRemoveLuceneIdxLock"."\n";
printAndLog "repair_records ->         : $optRepairRecords"."\n";
printAndLog "scrub_oplog ->            : $optScrubOplog"."\n";
printAndLog "backup_fts ->             : $optBackupCqFTS"."\n";
printAndLog "archive_fts ->            : $optArchiveCqFTS"."\n";
printAndLog "clear_state ->            : $optClearState"."\n";
printAndLog "lock_cq_fts ->            : $optLockCqFts"."\n";
printAndLog "unlock_cq_fts ->          : $optUnlockCqFts"."\n";
### ### printAndLog "validate_cq_fts_deplo ->  : $optValidateCqFtsDeplo"."\n";
printAndLog "prep_upgd_was_profiles -> : $optPrepUpgdFtsWASProfiles"."\n";
printAndLog "restore_was_profiles ->   : $optRestoreFtsWASProfiles"."\n";
printAndLog "upgrade_solr_app ->       : $optUpgradeSolrApp"."\n";
printAndLog "typeahead_cmd ->          : $optTypeaheadCmd"."\n";
printAndLog "gather_diagnostic_data -> : $optGatherDiagnosticData"."\n";
printAndLog "show_scenarios ->         : $optShowScenarios"."\n";
### ### printAndLog "verbose ->                : $optVerbose"."\n";
printAndLog "help ->                   : $optHelp"."\n";
printAndLog "**result's ->             : $result"."\n";
printAndLog ""."\n";
printAndLog ""."\n";


# fix '--ftshome'
$optCqFtsHomeDir = File::Spec->rel2abs($optCqFtsHomeDir);
$optCqFtsHomeDir =~ s/\\/\//sg;

if (-d $optCqFtsHomeDir)
{
	# we are given a directory
	if ((substr($optCqFtsHomeDir, -1) ne "/") or (substr($optCqFtsHomeDir, -1) ne "\\"))
	{
		$optCqFtsHomeDir .= "/";
	}
}

# fix '--fts_was_profile_home'
if ($optFtsWasProfileHome ne "")
{
	$optFtsWasProfileHome = File::Spec->rel2abs($optFtsWasProfileHome);
	$optFtsWasProfileHome =~ s/\\/\//sg;
	
	# we are given a directory
	if ((substr($optFtsWasProfileHome, -1) ne "/") or (substr($optFtsWasProfileHome, -1) ne "\\"))
	{
		$optFtsWasProfileHome .= "/";
	}
}

# fix '--was_profile_ports_file'
if ($optFtsWasProfPortsFile ne "")
{
	$optFtsWasProfPortsFile = File::Spec->rel2abs($optFtsWasProfPortsFile);
	$optFtsWasProfPortsFile =~ s/\\/\//sg;
}

my $cqVersion = "";

my %defValues = ();
my %newValues = ();

my $tag 		 = "";
my $defaultValue = "";
my $newValue	 = "";

my $isFTSEnabled = 0;
my $isInDefaultValue = 0;
my $isInNewValue = 0;

my $element = "";

sub trim;
sub compareArrays;
sub cleanArrayList;
sub dealWithError;
sub chdirAndValidate;
sub validateFileGenerated;
sub executeExternalCommand;
sub createAboutThisFts;
sub resolveFTSDataDir;
sub setupCqFts;
sub createNewCqFtsSetup;
sub getSolrCopyFieldData;
sub getSolrFieldData;
sub findExistingCqFtsSetup;
sub copyFiles;
sub copyFilesMatch;
sub validateUser;
sub cleanupConfigUserData;
sub fixPathSlashes;
sub findFtsWebXMLFolder;
sub findFtsServerXMLFolder;
sub getHeapSizeAndJVMArg;
sub getEntityFileWithoutDispFieldInfo;
sub buildScenariosFromXMLFile;

sub Func_copy_fts_template;
sub Func_gen_entity_file;
sub Func_gen_fts_files;
sub Func_cust_fts_files;
sub Func_cust_solr_files;
sub Func_stop_fts_was_profile;
sub Func_set_solr_home;
sub Func_enable_cqweb_fts;
sub Func_disable_cqweb_fts;
sub Func_run_batch_idx;
sub Func_start_update_idx;
sub Func_update_fts_prop_files;
sub Func_stop_update_idx;
sub Func_start_fts_was_profile;
sub Func_fresh_batch_idx;
sub Func_fresh_update_idx;
sub Func_repair_records;
sub Func_optimize_idx;
sub Func_remove_lucene_idx_lock;
sub Func_add_record_type;
sub Func_create_fts_was_profile;
sub Func_delete_fts_was_profile;
sub Func_scrub_oplog;
sub Func_backup_fts;
sub Func_archive_fts;
sub Func_clear_state;
sub Func_lock_cq_fts;
sub Func_unlock_cq_fts;
sub Func_validate_cq_fts_deplo;
sub Func_prep_upgd_was_profiles;
sub Func_restore_was_profiles;
sub Func_upgrade_solr_app;
sub Func_typeahead_cmd;
sub Func_gather_diagnostic_data;
sub Func_show_scenarios;
sub Func_set_was_max_mem;
sub Func_verbose;	# no-op

sub insertTypeaheadPropertyData;
sub extractTypeaheadPropertyData;
sub updateFTSPropFileIdxSetting;
sub customizeFtsFiles;
sub enableDisableCQWebFTS;
sub dieIfCqFtsIsLocked;
sub dieIf64BitWAS;
sub dieIfInState;
sub printValues;
sub setupCQFts;
sub doNothing;
sub getUserValue;
sub copyDataFile;
sub profileExist;
sub findNextAvailablePortNumber;
sub findPortNumberForProfile;
sub setDefaultValues;
sub runMultiutilOnUnix;
sub removeRecordType;
sub createDeplySolrScript;
sub deleteDeplySolrScript;
sub findProfilePath_cqweb;
sub createFTSConfigXMLFile;
sub isThisRecordTypeIndexed;
sub isThisFieldIndexed;
sub isRecordTypeExist;
sub isFieldListExist;
sub getListOfProfilesName;
sub getProfilePath;
sub isThisFtsProfile;


#
# deal with Unix / Windows
#
my $Q = "\'";	# assume Unix
my $gIsWindows = 0;
if ($^O =~ /MSWin/)
{
	$gIsWindows = 1;
	$Q = "\"";
}

my $ext = ($gIsWindows eq 1) ? ".bat" : ".sh";


#
# Create and set our global variables that are not config file driven
#
my $ftsClearQuestHomeDir = "";
$ftsClearQuestHomeDir = $ENV{'CLEARQUEST_HOME'};	# on Windows
if (!defined($ftsClearQuestHomeDir) or $ftsClearQuestHomeDir eq "")
{
	# don't give up just yet
	$ftsClearQuestHomeDir = $ENV{'CQ_HOME'};	# on Unix
	if (!defined($ftsClearQuestHomeDir) or $ftsClearQuestHomeDir eq "")
	{
		# keep trying, one last time ...
		$ftsClearQuestHomeDir = $ENV{'RATIONAL_HOME'};
		if (defined($ftsClearQuestHomeDir) and $ftsClearQuestHomeDir ne "")
		{
			# 'RATIONAL_HOME' doesn't have 'ClearQuest"
			$ftsClearQuestHomeDir .= "/ClearQuest";
		}
	}
}
if ((substr($ftsClearQuestHomeDir, -1) ne "/") or (substr($ftsClearQuestHomeDir, -1) ne "\\"))
{
	$ftsClearQuestHomeDir .= "/";
}
my $ftsRationalCommon = "";
$ftsRationalCommon = $ENV{'RATIONAL_COMMON'};
if (!defined($ftsRationalCommon) or $ftsRationalCommon eq "")
{
	$ftsRationalCommon = $ftsClearQuestHomeDir."../common";
}
if ((substr($ftsRationalCommon, -1) ne "/") or (substr($ftsRationalCommon, -1) ne "\\"))
{
	$ftsRationalCommon .= "/";
}
my $ftsJREHomeDir = "";
$ftsJREHomeDir = $ENV{'JRE_HOME'};
if ((!defined($ftsJREHomeDir) or $ftsJREHomeDir eq "") and $gIsWindows eq 0)
{
	$ftsJREHomeDir = $ftsClearQuestHomeDir."../common/java/jre";
}
if ((substr($ftsJREHomeDir, -1) ne "/") or (substr($ftsJREHomeDir, -1) ne "\\"))
{
	$ftsJREHomeDir .= "/";
}
my $ftsFtsWasAppSrvDir = "";
$ftsFtsWasAppSrvDir = $ENV{'CQFTS_AppServer_HOME'}; ### ###
if (!defined($ftsFtsWasAppSrvDir) or $ftsFtsWasAppSrvDir eq "")
{
	# we need to get the AppServer's path from IM's file
	$ftsFtsWasAppSrvDir = findWASAppServerDir();
}
if (!((substr($ftsFtsWasAppSrvDir, -1) eq "/") or (substr($ftsFtsWasAppSrvDir, -1) eq "\\")))
{
	$ftsFtsWasAppSrvDir .= "/";
}

my $ftsWASProfileHomeDir = $optFtsWasProfileHome;
my $ftsWASProfilePortsDefPropFile = $optFtsWasProfPortsFile;
my $ftsFtsWasAppSrvBinDir = $ftsFtsWasAppSrvDir."bin/";
my $ftsCQtsdbcrawlerPL = $ftsClearQuestHomeDir."cqtsdbcrawler.pl";
my $ftsCqtssetupPL = $ftsClearQuestHomeDir."cqtssetup.pl";
my $ftsCQtacfgPL = $ftsClearQuestHomeDir."cqtacfg.pl";
my $ftsSetSolrHomePY = $ftsClearQuestHomeDir."cqweb/scripts/setSolrHome.py";
my $ftsJavaEXE = $ftsJREHomeDir."bin/java";
my $ftsCQPerl = "";
my $ftsLockFile = "";
my $ftsStateFile = "";
my $ftsKeyFile = "";
my $ftsPWDFile = "";
my $ftsSourceDir = "";
my $ftsDBSetName = $optDbSet;
my $ftsUserDbName = $optuserDB;
my $ftsSiteName = "<local>";
my $ftsIsReplicated = 0;
my $ftsIsSiteWorkingMaster = 0;
my $ftsLocalReplicaName = $ftsSiteName;
my $ftsReplicaNames = "";
my $ftsReplicaNamesList = "";
my $ftsEnabledAtReplicaNamesList = "";
my $ftsUserName = $optUserName;
my $ftsUserPassword = $optPassword;
my $ftsTargetDir = "";
my $ftsWASProfileName;
my $ftsFtsWASProfilePort;
my $ftsFtsWASProfileCellName;
my $ftsFtsWASProfileNodeName;
my $ftsFtsServerName = lc(hostname);
my $ftsUserDefinedEntityFile;
my $ftsDataDir = "";
my $ftsSolrFile = "";
my $ftsSolrTypeaheadFile = "";
my $ftsLuceneIndexDir = "";
my $ftsEntityFileName;
my $ftsEntityFile;
my $ftsCQPropertyFileName;
my $ftsCQPropertyFile;
my $ftsSolrIntermediateFileName;
my $ftsSolrIntermediateFile;
my $ftsIsEnableCQFTS;
my $ftsDoBatchIndex;
my $ftsStartupdateIndex;
my $ftsBatchIndexJVMParm;
my $ftsWasProfileMaxHeapSize;
my $ftsBatchIndexBatchSize;
my $ftsBatchIndexDelay;
my $ftsBatchIndexThreads;
my $ftsUpdateIndexBatchSize;
my $ftsUpdateIndexDelay;
my $ftsUpdateIndexLoginInterval;
my $ftsAddRecordType;
my $ftsRemoveRecordType;
my $ftsDeleteFtsWASProfiles;
my $ftsDeleteIndex;
my $ftsEnableTypeahead;
my $ftsTypeAheadCacheFillTries;
my $ftsTypeAheadCacheSize;
my $ftsTypeAheadPageSize;
my $ftsSolrVersion;
my $ftsListOfIndxedRecordTypes;
my %ftsHashOfIndexedFields = ();
my @ftsExistingProfileNames;
my %ftsEntityFieldList = ();
my $installUtilMsgFile;
my $multiutilMsgFile;


fixPathSlashes();		# changes any "\" to "/"
validateUser(); 		# process only if user has proper CQ permission


##############################################################################
# it's time to do real work, where data may get modified
##############################################################################

# first, check to see if they are asking us to initialize, for the first time, a CQ FTS deployment
if ($optInitCqFTS eq "1")	# --init_cq_fts
{
	# figure out our profile home
	if (!defined($optFtsWasProfileHome) or $optFtsWasProfileHome eq "")
	{
		# no WAS profile home was given by the user, so set to the default of %RATIONAL_HOME%/clearquest/cqweb/cqwebprofile
		$ftsWASProfileHomeDir = $ftsClearQuestHomeDir."cqweb/";
	}
	else
	{
		$ftsWASProfileHomeDir = $optFtsWasProfileHome;
	}
	
	# figure out our profile ports
	if (!defined($optFtsWasProfPortsFile) or $optFtsWasProfPortsFile eq "")
	{
		# no WAS profile ports was given by the user, it means we will auto generate them
		$ftsWASProfilePortsDefPropFile = "";
	}
	else
	{
		$ftsWASProfilePortsDefPropFile = $optFtsWasProfPortsFile;
	}
	
	# we are asked to create a new depolyment
	initCqFTS();
	
	# with the --init_cq_fts option, they can also specify --create_fts_was_profile
	if ($optCreateCqFtsWasProfile eq "1")
	{
		push(@gCmdToExecute, ["Func_"."create_fts_was_profile"]);		# --create_fts_was_profile
	}
}
else
{
	# we are going after a pre-existing depolyment, read its values
	$optCqFtsHomeDir = findExistingCqFtsSetup();
}

readConfigData($optCqFtsHomeDir);	# read the XML config file into memory
setDefaultValues();					# fill in defaults if missing from the config file


#
# Now that we have read our config data, set our global variables
#
$ftsSolrVersion = getUserValue("solrVersion");
$ftsSourceDir = getCQSolrTemplateDir($ftsSolrVersion);
$ftsDBSetName = getUserValue("dbSet");
$ftsUserDbName = getUserValue("userDB");
$ftsSiteName = HTML::Entities::decode(getUserValue("siteName"));
$ftsEnabledAtReplicaNamesList = HTML::Entities::decode(getUserValue("enableAtSites"));
if ($ftsEnabledAtReplicaNamesList eq "")
{
	$ftsEnabledAtReplicaNamesList = $ftsReplicaNamesList;
}
else
{
	$ftsEnabledAtReplicaNamesList = cleanArrayList($ftsEnabledAtReplicaNamesList);
}
$ftsTargetDir = getUserValue("destination");
$ftsUserDefinedEntityFile = getUserValue("entityFile");
$ftsDataDir = resolveFTSDataDir();
$ftsLockFile = $ftsDataDir."cqfts.lock";
$ftsStateFile = $ftsDataDir."cqfts.state";
$ftsKeyFile = $ftsDataDir."key.txt";
$ftsPWDFile = $ftsDataDir."pwd.txt";
$ftsSolrFile = $ftsDataDir."Solr/solr/conf/schema.xml";
$ftsSolrTypeaheadFile = $ftsDataDir."Solr/typeahead/conf/schema.xml";
$ftsLuceneIndexDir = $ftsDataDir."Solr/solr/data/index/";
$ftsWASProfileName = getUserValue("ftsWASProfileName");
$ftsWASProfileHomeDir = getUserValue("ftsWASProfileHome");
$ftsFtsWASProfilePort = getUserValue("ftsWASProfilePort");
$ftsWASProfilePortsDefPropFile = getUserValue("ftsWASProfilePortsDefPropFile");
$ftsFtsServerName = getUserValue("ftsServerName");
$ftsEntityFileName = "Entity-".$ftsDBSetName."-".$ftsUserDbName.".txt";
$ftsEntityFile = $ftsDataDir.$ftsEntityFileName;
$ftsCQPropertyFileName = "CQ-".$ftsDBSetName."-".$ftsUserDbName.".xml";
$ftsCQPropertyFile = $ftsDataDir.$ftsCQPropertyFileName;
$ftsSolrIntermediateFileName = "Solr-".$ftsDBSetName."-".$ftsUserDbName.".xml";
$ftsSolrIntermediateFile = $ftsDataDir.$ftsSolrIntermediateFileName;
$ftsIsEnableCQFTS = getUserValue("enableCQFTS");
$ftsDoBatchIndex = getUserValue("runBatchIndex");
$ftsStartupdateIndex = getUserValue("startUpdateIndex");
$ftsBatchIndexJVMParm = getUserValue("batchIndexJVMParm");
$ftsWasProfileMaxHeapSize = getUserValue("ftsWASProfileMaxHeapSize");
$ftsBatchIndexBatchSize = getUserValue("batchIndexBatchSize");
$ftsBatchIndexDelay = getUserValue("batchIndexDelay");
$ftsBatchIndexThreads = getUserValue("batchIndexThreads");
$ftsUpdateIndexBatchSize = getUserValue("updateIndexBatchSize");
$ftsUpdateIndexDelay = getUserValue("updateIndexDelay");
$ftsUpdateIndexLoginInterval = getUserValue("updateIndexLoginInterval");
$ftsAddRecordType = getUserValue("addRecordType");
$ftsRemoveRecordType = getUserValue("removeRecordType");
$ftsDeleteFtsWASProfiles = getUserValue("deleteFtsWASProfiles");
$ftsDeleteIndex = getUserValue("deleteIndex");
$ftsEnableTypeahead = getUserValue("enableTypeahead");
$ftsTypeAheadCacheFillTries = getUserValue("typeAheadCacheFillTries");
$ftsTypeAheadCacheSize = getUserValue("typeAheadCacheSize");
$ftsTypeAheadPageSize = getUserValue("typeAheadPageSize");

fixPathSlashes();			# changes any "\" to "/", do this again to clean up what we may have read from our config file
cleanupConfigUserData();	# the user typed data in the config file may need some cleanup
openLogger();				# we can now write to the log file

$installUtilMsgFile = $ftsDataDir."EnablingCQWebFTS.txt";
$multiutilMsgFile = $ftsDataDir."ScruboplogFTS.txt";

printAndLog $msgLabel_INFO."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQTSADMIN_VER", $gVersion_cqtsadmin)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SOLR_VERSION", $ftsSolrVersion)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CONFIG_XML_FILE", $optCqFtsHomeDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQ_HOME", $ftsClearQuestHomeDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_COMMON_HOME", $ftsRationalCommon)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_JRE_HOME", $ftsJREHomeDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQ_USER_NAME", $ftsUserName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQ_USER_PW", "********")."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_FTS_SOURCE_TEMPL_DIR", $ftsSourceDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_FTS_TARGET_DIR", $ftsTargetDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_FTS_DATA_DIR", $ftsDataDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_DB_SET_NAME", $ftsDBSetName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_USER_DB_NAME", $ftsUserDbName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SITE_NAME", $ftsSiteName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SITE_NAME_LIST", $ftsReplicaNamesList)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENABLE_AT_SITES", $ftsEnabledAtReplicaNamesList)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_WAS_APPSRV_HOME", $ftsFtsWasAppSrvDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_HOME", $ftsWASProfileHomeDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_NAME", $ftsWASProfileName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_PORT", $ftsFtsWASProfilePort)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_PORT_FILE", $ftsWASProfilePortsDefPropFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_XMX_SETTING", $ftsWasProfileMaxHeapSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SERVER_NAME", $ftsFtsServerName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_USER_ENTITY_FILE", $ftsUserDefinedEntityFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SOLR_SCHEMA_FILE", $ftsSolrFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_LUCENE_IDX_DIR", $ftsLuceneIndexDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENTITY_FILE", $ftsEntityFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROPERTY_FILE", $ftsCQPropertyFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SOLR_INTERMEDIATE_FILE", $ftsSolrIntermediateFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PW_KEY_FILE", $ftsKeyFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENCRYPTED_PW_FILE", $ftsPWDFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CRAWLER_FILE", $ftsCQtsdbcrawlerPL)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SETUP_FILE", $ftsCqtssetupPL)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SETSOLRHOME_FILE", $ftsSetSolrHomePY)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQTACFG_FILE", $ftsCQtacfgPL)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_JAVA_EXE_FILE", $ftsJavaEXE)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENABLE_CQ_FTS", $ftsIsEnableCQFTS)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_START_BATCH_IDX", $ftsDoBatchIndex)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_START_UPDATE_IDX", $ftsStartupdateIndex)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_JVM_BATCH_MODE", $ftsBatchIndexJVMParm)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_BATCH_SIZE", $ftsBatchIndexBatchSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_BATCH_DELAY", $ftsBatchIndexDelay)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_BATCH_THREADS", $ftsBatchIndexThreads)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_UPDATE_SIZE", $ftsUpdateIndexBatchSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_UPDATE_DELAY", $ftsUpdateIndexDelay)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_UPDATE_LOGIN_INTERVAL", $ftsUpdateIndexLoginInterval)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ADD_RECORD", $ftsAddRecordType)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_REMOVE_RECORD", $ftsRemoveRecordType)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_TYPE_AHEAD_ENABLED", $ftsEnableTypeahead)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_TYPE_AHEAD_CACHE_FILL_TRIES", $ftsTypeAheadCacheFillTries)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_TYPE_AHEAD_CACHE_SIZE", $ftsTypeAheadCacheSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_TYPE_AHEAD_PAGE_SIZE", $ftsTypeAheadPageSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_DELETE_PROFILES", $ftsDeleteFtsWASProfiles)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_DELETE_INDEX", $ftsDeleteIndex)."\n";

# Before continuing, check to see if user has elevated privilage
validateUserElevation(1);	# check to see if the user has right permission to execute in this CMD / Shell


# the next logic depends on the following rules which was checked earlier.
# at this point, they either have used --init_cq_fts or --setup_cq_fts or 
# --create_fts_was_profile, which are the only 3 that we allow them to mix.
# since we already processed --init_cq_fts, lets see if they have --setup_cq_fts
# if so, process it, if not, and if --init_cq_fts was not given, then we process
# what ever they gave us on the command line

# if they are asking for a default end-to-end setup, do so now
if ($optSetupCqFTS eq "1")	# --setup_cq_fts
{
	# use default end-to-end setup
	setupCQFts();
}
# otherwise, if they have not used '--init_cq_fts', than it means they are passing their own command(s), so run it/them
elsif ($optInitCqFTS ne "1")	# --init_cq_fts
{
	# build the list of command-line options the user wanted to execute
	@gCmdToExecute = ();
	
	foreach my $item (@gCmdLineArgvList)
	{
		# skip those non-executable command line options
		if (($item eq "username") or ($item eq "password") or
			($item eq "dbset") or ($item eq "userdb") or
			($item eq "ftshome") or ($item eq "init_cq_fts") or
			($item eq "setup_cq_fts") or ($item eq "fts_was_profile_home") or
			($item eq "was_profile_ports_file"))
		{
			next;
		}
		else
		{
			my $foundAt = index($item, '=');	# if the command takes arg, remove the the arg part
			if ($foundAt ne -1)
			{
				$item = substr($item, 0, $foundAt);
			}
			push(@gCmdToExecute, [ "Func_".$item ]);
		}
	}
}


#
# we are now ready to execute each command in the order given
#
foreach my $function (@gCmdToExecute)
{
	my $argsStr = "";
	my @args = ();
	my $skipFuncName = 1;
	foreach my $arg (@$function)
	{
		if ($skipFuncName eq 1)
		{
			$skipFuncName = 0;
			next;
		}
		
		push(@args, $arg);		# build the function args
		
		if (length($argsStr) != 0)
		{
			$argsStr .= ", ";
		}
		$argsStr .= $arg;
	}
	
	my $funcName = @$function[0];
	my %funcCaller = ( $funcName => \&$funcName );
	
	printAndLog $msgLabel_INFO."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("CALLING_TEXT").$funcName."(".$argsStr.")"."\n";
	
	$funcCaller{$funcName}->(@args);	# make the function call
}


#
# If the 'AboutThisFTS.txt' file doesn't exist, create it and tell the user about it
#
$gAboutThisFTSFile = $ftsDataDir."AboutThisFTS.txt";
if (!(-f $gAboutThisFTSFile))
{
	# if we just executed the 'archive_fts' command, than we can't create the 'about' file
	if ($optArchiveCqFTS ne "1")
	{
		createAboutThisFts($gAboutThisFTSFile);
	}
}


if ($ftsIsSiteWorkingMaster eq 0)	# give them a warning about working master site
{
	printAndLog $msgLabel_WARN."\n", 1;
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("AT_WOKING_MASTER_L1")."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("AT_WOKING_MASTER_L2")."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("AT_WOKING_MASTER_L3")."\n", 1;
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN."\n", 1;
}


if ($optArchiveCqFTS ne "1")	# don't show this msg if we are archiving
{
	printAndLog $msgLabel_INFO."\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SEE_FILE", $gAboutThisFTSFile)."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FOR_IMPORTANT_NOTES")."\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	printAndLog $msgLabel_INFO."\n";
}

if (($optUpgradeSolrApp eq "1") and ($gShowMustReIndexMsg == 1))	# Re-index is required
{
	# Re-index is required; let the user know
	printAndLog $msgLabel_WARN."\n", 1;
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("REINDEX_IS_REQUIRED_1")."\n";
	printAndLog $msgLabel_WARN.$langMsgText->maketext("REINDEX_IS_REQUIRED_2")."\n";
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN."\n", 1;
}

if($gIncludeUsersGroups eq 1)
{
	# Show the warning message if include users and groups
	printAndLog $msgLabel_WARN."\n", 1;
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("INCLUDE_USERS_GROUPS")."\n", 1;
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN."\n", 1;
}

closeLogger();			# close the logger

exit 0;		# exit script, all is well




##############################################################################
# Close the logger
sub closeLogger
{
	if ($cqtsadminLogFile ne "")
	{
		# finish off our log
		printAndLog $msgLabel_INFO.$langMsgText->maketext("LOG_ENDED_ON", "".localtime())."\n";
		close logFileHandler;
	}
	elsif ($logDataBacklog ne "")
	{
		if ($optArchiveCqFTS eq "")
		{
			# we run into issues creating our log file, so attempt to flush the log data (in memory) to local file
			my $flushLogFile = $ftsDataDir."log-flush.log";
			if (!open(FH, ">".$flushLogFile))	# yes, it's OK to write to the file as utf-8, it's what we want
			{
				warn $msgLabel_WARN.""."\n", 1;
				warn $msgLabel_WARN."**************** ".$langMsgText->maketext("NON_FATAL_ERROR")." ****************"."\n", 1,
				warn $msgLabel_WARN.$langMsgText->maketext("CANNOT_CREATE_LOG_FILE", $flushLogFile)." ".$msgLable_oS_ERROR.$!, 1;
				warn $msgLabel_WARN.$langMsgText->maketext("LOST_LOG_MSG_LINE_1")."\n", 1;
				warn $msgLabel_WARN.$langMsgText->maketext("LOST_LOG_MSG_LINE_2")."\n", 1;
				warn $msgLabel_WARN.""."\n", 1;
			}
			else
			{
				print FH $logDataBacklog;
				close(FH);
				warn $msgLabel_WARN.$langMsgText->maketext("FLUSH_LOG", $flushLogFile)."\n", 1;
			}
		}
	}
	
	$cqtsadminLogFile = "";	# close the logger
}


##############################################################################
# Creates the log file and opens it.
sub openLogger
{
	my $logDir = $ftsDataDir."logs/";
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	
	$cqtsadminLogFile = $logDir."cqtsadmin_log-".$ftsDBSetName."_".$ftsUserDbName."-".$timeData.".log";
	
	my $tmp = "";
	
	if (!(-e $logDir))
	{
		$tmp = $msgLabel_INFO."mkpath -> [".$logDir."]"."\n";
		File::Path::mkpath $logDir;
	}
	
	if (!open(logFileHandler, ">:utf8", $cqtsadminLogFile))	
	{
		warn $msgLabel_WARN.""."\n", 1;
		warn $msgLabel_WARN."**************** ".$langMsgText->maketext("NON_FATAL_ERROR")." ****************"."\n", 1;
		warn $msgLabel_WARN.$langMsgText->maketext("CANNOT_CREATE_LOG_FILE", $cqtsadminLogFile)."\n", 1;
		warn $msgLabel_WARN.$msgLable_oS_ERROR.$!."\n", 1;
		warn $msgLabel_WARN.$langMsgText->maketext("FLUSH_LOG_ATTEMPT_LINE_1", $ftsDataDir."log-flush.log")."\n", 1;
		warn $msgLabel_WARN.$langMsgText->maketext("FLUSH_LOG_ATTEMPT_LINE_2")."\n", 1;
		warn $msgLabel_WARN.""."\n", 1;
		
		$cqtsadminLogFile = "";
	}
	else
	{
		logFileHandler->autoflush(1);
	}
	
	printAndLog $tmp;
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOG_STARTED_ON", "".localtime())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("WRITTING_TO_LOG_FILE", $cqtsadminLogFile)."\n";
}


##############################################################################
# Prints a message tot he screen and logs it
sub printAndLog
{
	my ($msg, $sendToScreen) = @_;
	
	if (!defined($sendToScreen))
	{
		$sendToScreen = 0;	# if we were not told to display the text, we won't
	}
	
	if (($sendToScreen eq 0) and ($msg =~ m/====>/))
	{
		$sendToScreen = 1;	# always display banner text
	}
	
	if (($gVerboseOutput eq 1) or ($sendToScreen eq 1))
	{
		my $screenMsg = $msg;
		
		if (!defined($::PRINTANDLOG_ENC) and defined($gCqFtsEncoding))
		{
			if ($gCqFtsEncoding ne "")
			{
				$screenMsg = Encode::encode($gCqFtsEncoding, $screenMsg);
			}
		}

		print $screenMsg;		# output the message to the screen.
	}
	
	if (defined($::PRINTANDLOG_ENC) and defined($gCqFtsEncoding))
	{
		if ($gCqFtsEncoding ne "")
		{
			$msg = Encode::decode($gCqFtsEncoding, $msg);
		}
	}
	
	if ($cqtsadminLogFile eq "")
	{
		# the log file is not ready yet, cache log data into memory
		$logDataBacklog .= $msg;
	}
	else
	{
		# the log file is ready, write to it
		if ($logDataBacklog ne "")
		{
			# but first flush what we haev in memory
			print logFileHandler $logDataBacklog;
			$logDataBacklog = "";
		}
		
		print logFileHandler $msg;
	}
}


##############################################################################
# Reports an error message and terminates the script (if told so)
sub dealWithError
{
	my ($errMsg, $dieOnError) = @_;
	
	if ($dieOnError)
	{
		printAndLog "\n".$gTerminationErrorMsg."\n", 1;		# always show error messages, independet of if "verbose" is set or not
		printAndLog "\n".$errMsg, 1;
		
		die "\n"; ### ### .$errMsg;
	}
	else
	{
		printAndLog $msgLabel_WARN."*********vvvvvv ".$langMsgText->maketext("IGNORE_SAFE_ERROR")." vvvvvv*********"."\n", 1;
		printAndLog $errMsg, 1;
		printAndLog $msgLabel_WARN."*********^^^^^^ ".$langMsgText->maketext("IGNORE_SAFE_ERROR")." ^^^^^^*********"."\n", 1;
	}
}


##############################################################################
# Changes the current working directory and verifies the change
sub chdirAndValidate
{
	my ($chdirTo, $dieOnError) = @_;
	
	printAndLog $msgLabel_INFO."chdir -> [".$chdirTo."]"."\n";
	
	my $status = chdir $chdirTo;
	if (! $status)
	{
		dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  chdir ".$chdirTo."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Checks to see if a file was created or not
sub validateFileGenerated
{
	my ($executedCmd, $expectedFile, $dieOnError) = @_;
	
	# mask out password string
	my $msg = $executedCmd;
	
	$msg =~ s/-p \Q$ftsUserPassword\E/-p ********/sg;
	$msg =~ s/$ftsUserName \Q$ftsUserPassword\E/$ftsUserName ********/sg;
	
	my $status = 1;
	
	if (!(-f $expectedFile))
	{
		my $errMsg = $langMsgText->maketext("ERROR_DURING_EXE")."\n"."  ".$msg."\n\n".$langMsgText->maketext("FILE_NOT_GENERATED")."\n";
		
		dealWithError($errMsg, $dieOnError);
		
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Executes an external program / command; checks for it's success / completion, 
# and terminates if it fails (if instructed to)
sub executeExternalCommand
{
	my ($cmd, $matchingString, $dieOnErrorMatchKind, $dieOnError, $captureOutput) = @_;
	
	if (!defined($captureOutput))
	{
		$captureOutput = 0;	# use default
	}
	
	# mask out password string from the command to be executed
	my $maskedCmd = $cmd;
	
	$maskedCmd =~ s/-p \Q$ftsUserPassword\E/-p ********/sg;
	$maskedCmd =~ s/$ftsUserName \Q$ftsUserPassword\E/$ftsUserName ********/sg;
	$maskedCmd =~ s/-password \Q$ftsUserPassword\E/-password ********/sg;
	
	printAndLog $msgLabel_INFO."system -> [".$maskedCmd."]"."\n";
	
	my $status = 1;
	
	# execute the command
	if (open(FH,  "$cmd 2>&1 |"))
	{
		my $foundMatchingString = 0;
		my $line;
		
		while ($line = <FH>)
		{
			# output to the screen and log file unless if they asked not to (undocumented)
			if (not (defined($gSuppressExternalCmdOutput) and $gSuppressExternalCmdOutput eq "yes"))
			{
				if (defined($gExternalCmdMarker) and $gExternalCmdMarker ne "")
				{
					printAndLog $gExternalCmdMarker;
				}
				
				{
					local $::PRINTANDLOG_ENC = $gCqFtsEncoding;
					printAndLog $line;
				}
			}
			
			foreach my $matchItem (@$matchingString)
			{
				if (($matchItem ne "") and ($line =~ /$matchItem/))
				{
					$foundMatchingString = 1;
				}
			}
			
			if ($captureOutput == 1)
			{
				$gCaptureExeOutputData .= $line;
			}
		}
		close FH;
		
		if ($matchingString ne "")
		{
			if (($foundMatchingString eq 1) and ($dieOnErrorMatchKind eq $gDieKind_must_not_match))
			{
				# we found the required string, but we are not suppose to, so we will die
				dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$maskedCmd."]"."\n", $dieOnError);
				$status = 0;
			}
			elsif (($foundMatchingString eq 1) and ($dieOnErrorMatchKind eq $gDieKind_must_match))
			{
				# we found the required string, which we are required to, so we will not die
			}
			elsif (($foundMatchingString eq 0) and ($dieOnErrorMatchKind eq $gDieKind_must_not_match))
			{
				# we did not find our matching string, which we are not suppose to, so we don't die
			}
			elsif (($foundMatchingString eq 0) and ($dieOnErrorMatchKind eq $gDieKind_must_match))
			{
				# we did not find our matching string, which we are suppose to, so we will die
				dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$maskedCmd."]"."\n", $dieOnError);
				$status = 0;
			}
			elsif (($foundMatchingString eq 0) and ($dieOnErrorMatchKind eq $gDieKind_none))
			{
				# we did not find the required string, which we are not required to, so we don't die
			}
			else
			{
				# this is an unknown state ... 
				dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$maskedCmd."]"."\n", $dieOnError);
				$status = 0;
			}
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_EXE", $cmd)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Creates the file, "AboutThisFTS.txt' file
sub createAboutThisFts
{
	my ($aboutThisFtsFile) = @_;
	
	if (!open(FH, ">:utf8", $aboutThisFtsFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $aboutThisFtsFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	
	print FH $langMsgText->maketext("ABOUT_INFO")."\n";
	print FH "\n";
	print FH $langMsgText->maketext("ABOUT_CREATED_ON", "".localtime())."\n";
	print FH $langMsgText->maketext("ABOUT_OS", $^O)."\n";
	print FH $langMsgText->maketext("ABOUT_CFG_XML_FILE", $optCqFtsHomeDir)."\n";
	print FH $langMsgText->maketext("ABOUT_SOURCE_TEMPL_DIR", $ftsSourceDir)."\n";
	print FH $langMsgText->maketext("ABOUT_DB_SET_NAME", $ftsDBSetName)."\n";
	print FH $langMsgText->maketext("ABOUT_USER_DB_NAME", $ftsUserDbName)."\n";
	print FH $langMsgText->maketext("ABOUT_SITE_NAME", $ftsSiteName)."\n";
	print FH $langMsgText->maketext("ABOUT_SITE_LIST_NAME", $ftsReplicaNamesList)."\n";
	print FH $langMsgText->maketext("ABOUT_ENABLED_SITE_LIST_NAME", $ftsEnabledAtReplicaNamesList)."\n";
	print FH $langMsgText->maketext("ABOUT_SERVER_HOST", $ftsFtsServerName)."\n";
	print FH $langMsgText->maketext("ABOUT_FTS_TARGET_DIR", $ftsTargetDir)."\n";
	print FH $langMsgText->maketext("ABOUT_WAS_APPSRV_HOME", $ftsFtsWasAppSrvDir)."\n";
	print FH $langMsgText->maketext("ABOUT_PROFILE_HOME", $ftsWASProfileHomeDir)."\n";
	print FH $langMsgText->maketext("ABOUT_PROFILE_NAME", $ftsWASProfileName)."\n";
	print FH $langMsgText->maketext("ABOUT_PROFILE_PORT", $ftsFtsWASProfilePort)."\n";
	print FH $langMsgText->maketext("ABOUT_PROFILE_PORT_FILE", $ftsWASProfilePortsDefPropFile)."\n";
	print FH $langMsgText->maketext("ABOUT_SOLR_ADMIN_URL", $ftsFtsServerName, $ftsFtsWASProfilePort)."\n";
	print FH $langMsgText->maketext("ABOUT_LUCENE_INDEX_LOCATION", $ftsLuceneIndexDir)."\n";
	print FH $langMsgText->maketext("ABOUT_DATA_DIR", $ftsDataDir)."\n";
	print FH $langMsgText->maketext("ABOUT_ENTITY_FILE", $ftsEntityFile)."\n";
	print FH $langMsgText->maketext("ABOUT_PROPERTY_FILE", $ftsCQPropertyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_SOLR_INTERM_FILE", $ftsSolrIntermediateFile)."\n";
	print FH $langMsgText->maketext("ABOUT_PW_KEY_FILE", $ftsKeyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_ENCRPTED_PW_FILE", $ftsPWDFile)."\n";
	print FH $langMsgText->maketext("ABOUT_SOLR_FILE", $ftsSolrFile)."\n";
	print FH "\n"."\n";
	print FH $langMsgText->maketext("ABOUT_POST_SETUP_INFO")."\n";
	print FH "\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_1", $ftsFtsServerName, $ftsFtsWASProfilePort)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_2", $ftsCQPropertyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_3", $ftsKeyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_4", $ftsWASProfileName)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_5")."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_6")."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_7")."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_8")."\n";
	
	if ($gIsWindows eq 0)
	{
		print FH $langMsgText->maketext("ABOUT_POST_INFO_9", $multiutilMsgFile)."\n";
		runMultiutilOnUnix();
	}
	
	print FH "\n"."\n";
	print FH $langMsgText->maketext("ABOUT_SEE_INFO_CENTER")."\n";
	
	close FH;
}


##############################################################################
# Fixes up the FTS home path
sub resolveFTSDataDir
{
	if ($ftsDataDir eq "/")
	{
		my $foundAt = rindex($optCqFtsHomeDir, "/") + 1;
		if ($foundAt eq -1)
		{
			$foundAt = rindex($optCqFtsHomeDir, "\\") + 1;
		}
		if ($foundAt ne -1)
		{
			$ftsDataDir = substr($optCqFtsHomeDir, 0, $foundAt);
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_DETERMINE_FTS_DIR", $optCqFtsHomeDir)."\n", 1);
		}
	}
	
	return ($ftsDataDir);
}


##############################################################################
# Creates a new FTS setup
sub createNewCqFtsSetup
{
	if (substr($optCqFtsHomeDir, -1) ne "/")
	{
		$optCqFtsHomeDir .= "/";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATE_NEW_CQFTS", $optCqFtsHomeDir)."\n";
	
	$ftsTargetDir = $optCqFtsHomeDir;	# this is now our FTS target setup directory
	
	$ftsDataDir = $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName."/";
	printAndLog $msgLabel_INFO."mkpath -> [".$ftsDataDir."]"."\n";
	File::Path::mkpath $ftsDataDir;
	
	$optCqFtsHomeDir = $ftsDataDir."cqtsadmin"."-".$ftsDBSetName."-".$ftsUserDbName.".xml";
	createFTSConfigXMLFile($optCqFtsHomeDir);
}


##############################################################################
# Looks to see if an existing FTS setup exist or not
sub findExistingCqFtsSetup
{
	my $result = "";
	
	$ftsTargetDir = $optCqFtsHomeDir;	# this is now our FTS target setup directory
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOCATE_CQFTS", $ftsTargetDir)."\n";
	
	# directory exist, look for the default file CQ FTS XML configuration file
	$ftsDataDir = $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName."/";
	$result = $ftsDataDir."cqtsadmin"."-".$ftsDBSetName."-".$ftsUserDbName.".xml";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOOKING_FOR_CQFTS", $result)."\n";
	if (!(-f $result))		# "/CQ.Search/TextSearch_SAMPL/cqtsadmin-TextSearch-SAMPL.xml"
	{
		$result = "";
	}
	
	return ($result);
}


##############################################################################
# Changes '\' to '/' and adds a '/' to the end of the path (if one doesn't exist)
sub fixPathSlashes
{
	$ftsClearQuestHomeDir =~ s/\\/\//sg;
	$ftsWASProfileHomeDir =~ s/\\/\//sg;
	$ftsRationalCommon =~ s/\\/\//sg;
	$ftsJREHomeDir =~ s/\\/\//sg;
	$ftsSourceDir =~ s/\\/\//sg;
	$ftsTargetDir =~ s/\\/\//sg;
	$ftsDataDir =~ s/\\/\//sg;
	$ftsCqtssetupPL =~ s/\\/\//sg;
	$ftsCQtsdbcrawlerPL =~ s/\\/\//sg;
	$ftsCQtacfgPL =~ s/\\/\//sg;
	$ftsSetSolrHomePY =~ s/\\/\//sg;
	$ftsSolrFile =~ s/\\/\//sg;
	$ftsSolrTypeaheadFile =~ s/\\/\//sg;
	$ftsLuceneIndexDir =~ s/\\/\//sg;
	$ftsCQPerl =~ s/\\/\//sg;
	$ftsJavaEXE =~ s/\\/\//sg;
	$gPrepUpgdFtsWASProfiles =~ s/\\/\//sg;
	$gRestoreFtsWASProfiles =~ s/\\/\//sg;
	$ftsWASProfilePortsDefPropFile =~ s/\\/\//sg;
	$ftsFtsWasAppSrvBinDir =~ s/\\/\//sg;
	$ftsFtsWasAppSrvDir =~ s/\\/\//sg;
	
	if (($ftsClearQuestHomeDir ne "") and (substr($ftsClearQuestHomeDir, -1) ne "/"))
	{
		$ftsClearQuestHomeDir .= "/";
	}
	if (($ftsWASProfileHomeDir ne "") and (substr($ftsWASProfileHomeDir, -1) ne "/"))
	{
		# deal with pre 8.0 issue where profile-home is not given
		if (length($ftsWASProfileHomeDir) != 0)
		{
			$ftsWASProfileHomeDir .= "/";
		}
	}
	if (($ftsRationalCommon ne "") and (substr($ftsRationalCommon, -1) ne "/"))
	{
		$ftsRationalCommon .= "/";
	}
	if (($ftsJREHomeDir ne "") and (substr($ftsJREHomeDir, -1) ne "/"))
	{
		$ftsJREHomeDir .= "/";
	}
	if (($ftsSourceDir ne "") and (substr($ftsSourceDir, -1) ne "/"))
	{
		$ftsSourceDir .= "/";
	}
	if (($ftsTargetDir ne "") and (substr($ftsTargetDir, -1) ne "/"))
	{
		$ftsTargetDir .= "/";
	}
	if (($ftsDataDir ne "") and (substr($ftsDataDir, -1) ne "/"))
	{
		$ftsDataDir .= "/";
	}
	if (($gPrepUpgdFtsWASProfiles ne "") and (substr($gPrepUpgdFtsWASProfiles, -1) ne "/"))
	{
		$gPrepUpgdFtsWASProfiles .= "/";
	}
	if (($gRestoreFtsWASProfiles ne "") and (substr($gRestoreFtsWASProfiles, -1) ne "/"))
	{
		$gRestoreFtsWASProfiles .= "/";
	}
}


##############################################################################
# After reading cqfts admin data, clean it up before usage
sub cleanupConfigUserData
{
	# if they didn't give us any passowrd, it means no password which is a blank quote
	if ($ftsUserPassword eq "")
	{
		$ftsUserPassword = "\"\"";
	}
	
	# see if they want to associate the db-set and user-db name with the profile name
	$ftsWASProfileName =~ s/\[user-db-name\]/$ftsUserDbName/s;
	$ftsWASProfileName =~ s/\[db-set-name\]/$ftsDBSetName/s;
	
	# if the profile home is not given, set it to the default
	if ($ftsWASProfileHomeDir eq "")
	{
		$ftsWASProfileHomeDir = $ftsClearQuestHomeDir."cqweb/";
	}
	
	# work out the port for the WAS profile
	if ($ftsFtsWASProfilePort eq "[*]")
	{
		# the CQ FTS config XML file doesn't have a port number, figure it out
		if ($optCreateCqFtsWasProfile eq "1")
		{
			# we are createing a new profile, and are asked to figure out it's port number
			$ftsFtsWASProfilePort = findNextAvailablePortNumber(1);
		}
		else
		{
			# we are asked to get the port number of a given profile
			$ftsFtsWASProfilePort = findPortNumberForProfile($ftsWASProfileName, 1);
			if ($ftsFtsWASProfilePort eq "-1")
			{
				printAndLog "\n".$gTerminationErrorMsg."\n", 1;
				dealWithError($langMsgText->maketext("RETRIEVE_WAS_PORT_NUMBER", $ftsWASProfileName)."\n", 1);
			}
		}
	}
	
	if (length($ftsWASProfileName) > $gMaxWASProfileLen)
	{
		printAndLog "\n".$gTerminationErrorMsg."\n", 1;
		dealWithError($langMsgText->maketext("PROFILE_NAME_TOO_LONG", $ftsWASProfileName, $gMaxWASProfileLen)."\n", 1);
	}
	
	# remove any spaces that maybe in the list of record-types to add / remove
	$ftsAddRecordType = trim($ftsAddRecordType);
	$ftsRemoveRecordType = trim($ftsRemoveRecordType);
	$ftsDeleteFtsWASProfiles = trim($ftsDeleteFtsWASProfiles);
	$ftsDeleteIndex = trim($ftsDeleteIndex);
	$ftsEnableTypeahead = trim($ftsEnableTypeahead);
	$ftsTypeAheadCacheFillTries = trim($ftsTypeAheadCacheFillTries);
	$ftsTypeAheadCacheSize = trim($ftsTypeAheadCacheSize);
	$ftsTypeAheadPageSize = trim($ftsTypeAheadPageSize);
	$ftsSolrVersion = trim($ftsSolrVersion);
}


##############################################################################
# Verify that the CQ user has proper permission
sub validateUser
{
	my $user = $ftsUserName;
	my $pwd = $ftsUserPassword;
	my $dbSet = $ftsDBSetName;
	my $dbName = $ftsUserDbName;
	
	if ($pwd eq "\"\"")
	{
		$pwd = "";
	}
	
	eval
	{
		my $adminSessionObj = CQAdminSession::Build();
		
		$adminSessionObj->Logon($user, $pwd, $dbSet);
		
		$ftsIsReplicated = $adminSessionObj->IsReplicated();
		$ftsReplicaNames = $adminSessionObj->GetReplicaNames();
		$ftsLocalReplicaName = $adminSessionObj->GetLocalReplicaName();
		$ftsIsSiteWorkingMaster = $adminSessionObj->IsSiteWorkingMaster();
		
		CQAdminSession::Unbuild($adminSessionObj);
	};
	die "\n"."$@\n" if $@; $@ = "";
	
	if ($ftsIsReplicated eq 1)
	{
		foreach my $replicName (@$ftsReplicaNames)
		{
			if ($ftsReplicaNamesList ne "")
			{
				$ftsReplicaNamesList .= ",";
			}
			$ftsReplicaNamesList .= trim($replicName);
		}
	}
	else
	{
		if ($ftsLocalReplicaName eq "<local>")
		{
			$ftsReplicaNamesList = "";
		}
		else
		{
			$ftsReplicaNamesList = $ftsLocalReplicaName;
		}
	}
	
	$ftsSiteName = $ftsReplicaNamesList;
	
	# User session, for doing actual work
	my $sessionObj;
	
	eval
	{
		$sessionObj = CQSession::Build();
	
		$sessionObj->UserLogon($user, $pwd, $dbName, $dbSet);
	};
	die "\n"."$@\n" if $@; $@ = "";
	
	if ($sessionObj->IsUserSuperUser() eq 0)
	{
		dealWithError($langMsgText->maketext("NO_PRIVILAGE")."\n", 1);
	}
	
	my $productInfo = $sessionObj->GetProductInfo();	
	$cqVersion = $productInfo->GetFullProductVersion();
	
	my $indexedRecordTypes = "";
	
	$isFTSEnabled = $sessionObj->IsTextSearchEnabled();
	if ($isFTSEnabled)
	{
		my $textSearchDef = $sessionObj->BuildTextSearch();
		$ftsListOfIndxedRecordTypes = $textSearchDef->GetIndexedEntityDefNames();
		
		my $recordType;
		my $indexedFiledNames;
		foreach $recordType (@$ftsListOfIndxedRecordTypes)
		{
			if ($indexedRecordTypes ne "")
			{
				$indexedRecordTypes .= ", ";
			}
			$indexedRecordTypes .= $recordType;
			$indexedFiledNames = $textSearchDef->GetIndexedFieldDefNames($recordType);
			$ftsHashOfIndexedFields{$recordType} = $indexedFiledNames;
		}
		
		$indexedRecordTypes = "[".$indexedRecordTypes."]";
	}
	else
	{
		$indexedRecordTypes = $langMsgText->maketext("FTS_NOT_ENABLED");
		$ftsListOfIndxedRecordTypes = "";
	}
	
	# check to see if CQ FTS is installed
	if (!(-f $ftsCqtssetupPL))
	{
		dealWithError($langMsgText->maketext("CQ_FTS_NOT_INSTALLED")."\n", 1);
	}
	
	{
		my $entityDefNames = $sessionObj->GetEntityDefNames();
		
		# Iterate over the record types
		foreach my $entityName ( @$entityDefNames )
		{
			if(($entityName eq "history") || ($entityName eq "attachments") || ($entityName eq "ratl_replicas"))
			{
				next;  # record type 'history', 'attachments' and 'ratl_replicas' are not FTS indexed
			}
			my $entDef = $sessionObj->GetEntityDef($entityName);
			
			my $fieldsListString = "";
			my ($fieldNamesRef) = $entDef->GetFieldDefNames();
			foreach my $fieldName (@$fieldNamesRef)
			{
				if ($fieldName eq "history")
				{
					next;	# 'history' field is not FTS indexed
				}
				
				$fieldsListString = $fieldsListString.",".$fieldName;
			}
			
			if ($gCqFtsEncoding eq "")
			{
				$ftsEntityFieldList{$entDef->GetName()} = $fieldsListString;
			}
			else
			{
				$ftsEntityFieldList{Encode::decode($gCqFtsEncoding, $entDef->GetName())} = Encode::decode($gCqFtsEncoding, $fieldsListString);
			}
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FTS_INDEXED_RECORDS_TYPES", $indexedRecordTypes)."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("CQ_PROD_INFO")."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FULL_PROD_VER", $productInfo->GetFullProductVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("PROD_VER", $productInfo->GetProductVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("PATCH_VER", $productInfo->GetPatchVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LICE_VER", $productInfo->GetLicenseVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LICE_FEATURE", $productInfo->GetLicenseFeature())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("WEB_LICE_VER", $productInfo->GetWebLicenseVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_NAME", $productInfo->GetCompanyName())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_FULL_NAME", $productInfo->GetCompanyFullName())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_EMAIL_ADDESS", $productInfo->GetCompanyEmailAddress())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_WEB_ADDRESS", $productInfo->GetCompanyWebAddress())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BUILD_NUM", $productInfo->GetBuildNumber())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("STAGE_LABEL", $productInfo->GetStageLabel())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DEFAULT_DBSET_NAME", $productInfo->GetDefaultDbSetName())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("OM_VER_MAJOR", $productInfo->GetObjectModelVersionMajor())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("OM_VER_MINOR", $productInfo->GetObjectModelVersionMinor())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SUITE_PROD_VER", $productInfo->GetSuiteProductVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SUITE_BUILD_ID", $productInfo->GetSuiteBuildId())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FEATURE_LEVEL", $sessionObj->GetSessionFeatureLevel())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("IS_REPLICATED", $sessionObj->IsReplicated())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("IS_FTS_ENABLED", $sessionObj->IsTextSearchEnabled())."\n";
	
	my $FL = $sessionObj->GetSessionFeatureLevel();
	my $gIsReplicated = $sessionObj->IsReplicated();
	if (($FL < 7) and ($gIsReplicated eq 0))
	{
		$gIsReplicated = ($gIsReplicated eq 1) ? $langMsgText->maketext("IS_MULTISITED") : $langMsgText->maketext("IS_NOT_MULTISITED");
		
		printAndLog $msgLabel_WARN.""."\n", 1;
		printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
		printAndLog $msgLabel_WARN.$langMsgText->maketext("DB_FL", $FL)."\n", 1;
		printAndLog $msgLabel_WARN.$gIsReplicated."\n", 1;
		printAndLog $msgLabel_WARN.$langMsgText->maketext("NEED_FL7_OR_MULTISITE")."\n", 1;
		printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
		printAndLog $msgLabel_WARN.""."\n", 1;
	}
	
	CQSession::Unbuild($sessionObj);
	
	my @val = split(/\./, $cqVersion);
	if ($val[0] < 8)		# make sure the tool is being run on the correct version of CQ
	{
		dealWithError($langMsgText->maketext("WRONG_CQ_VERSION", $cqVersion)."\n", 1);
	}
}


##############################################################################
# check to see if the user has right permission to execute in this CMD / Shell
sub validateUserElevation
{
	my ($dieOnError) = @_;
	
	my $isElevated = 1;		# assume elevation is granted
	
	if ($gIsWindows eq 1)
	{
		createAdminElevationTestScript($dieOnError);
		
		my $cmd = $ftsDataDir."adminTest".$ext;
		my @mustMatch = ("Administrator PRIVILEGES Detected!");
		$isElevated = executeExternalCommand($cmd, \@mustMatch, $gDieKind_must_match, 0);
		
		deleteAdminElevationTestScript();
	}
	else
	{
		if ($> != 0)
		{
			$isElevated = 0;	# not admin
		}
	}
	
	if ($isElevated == 0)
	{
		dealWithError($langMsgText->maketext("ADMIN_ELEVATION_REQUIRED")."\n", $dieOnError);
	}
}


##############################################################################
# Helper function to recursively copy files
sub copyDataFile
{
	# make the data directory if it doesn't exist
	if (!(-e $ftsDataDir))
	{
		printAndLog $msgLabel_INFO."mkpath -> [".$ftsDataDir."]"."\n";
		File::Path::mkpath $ftsDataDir;
	}
	
	# copy the data file over
	my @fileName = split(/\//, $optCqFtsHomeDir);
	my $to = $ftsDataDir.$fileName[@fileName - 1];
	if (!(-e $to))
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CONFIG_IS_HERE", $to)."\n";
		if (!copy($optCqFtsHomeDir, $to))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $optCqFtsHomeDir, $msgLable_oS_ERROR)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
}


##############################################################################
# Checks to see if a WAS profile exist or not
sub profileExist
{
	my ($theProfileName) = @_;
	
	if (@ftsExistingProfileNames eq 0)
	{
		getListOfProfilesName(1);
	}
	
	my $doesProfileExist = 0;
	
	foreach my $profileName (@ftsExistingProfileNames)
	{
		if ($profileName eq $theProfileName)
		{
			$doesProfileExist = 1;
			last;
		}
	}
	
	return $doesProfileExist;
}


##############################################################################
# Returns the list of profiles name
sub getListOfProfilesName
{
	my ($dieOnError) = @_;
	
	my $status = 1;
	
	chdirAndValidate($ftsFtsWasAppSrvBinDir, $dieOnError);
	
	$gCaptureExeOutputData = "";		# remove any old output data
	my @mustMatch = $langMsgText->maketext("CANNOT_RETRIVE_PROFILE_PATH");
	my $cmd = $Q.$ftsFtsWasAppSrvBinDir."manageprofiles".$ext.$Q." -listProfiles";
	$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, $dieOnError, 1);
	
	if ($status eq 1)
	{
		# All is well
		my $numOfLines = 0;
		my @data = split(/\n/, $gCaptureExeOutputData);
		foreach my $line ( @data )
		{
			if ($numOfLines ne 0)
			{
				# We have more than one line; something went wrong
				dealWithError($langMsgText->maketext("UNEXPECTED_DATA", $gCaptureExeOutputData)."\n", $dieOnError);
			}
			$numOfLines++;
		}
		
		# Remove the leading "[" and trainling "]"
		$gCaptureExeOutputData = trim($gCaptureExeOutputData);
		my $foundAt;
		$foundAt = index($gCaptureExeOutputData, "[");
		if ($foundAt ne 0)
		{
			# We have more than one line; something went wrong
			dealWithError($langMsgText->maketext("UNEXPECTED_DATA", $gCaptureExeOutputData)."\n", $dieOnError);
		}
		else
		{
			$gCaptureExeOutputData = substr($gCaptureExeOutputData, 1);
		}
		$foundAt = rindex($gCaptureExeOutputData, "]");
		if ($foundAt ne length($gCaptureExeOutputData) - 1)
		{
			# We have more than one line; something went wrong
			dealWithError($langMsgText->maketext("UNEXPECTED_DATA", $gCaptureExeOutputData)."\n", $dieOnError);
		}
		else
		{
			$gCaptureExeOutputData = substr($gCaptureExeOutputData, 0, length($gCaptureExeOutputData) - 1);
		}
		
		@data = split(/,/, $gCaptureExeOutputData);
		foreach my $item ( @data )
		{
			push(@ftsExistingProfileNames, trim($item));
		}
	}
	
	return $status;
}


##############################################################################
# Get a profile's path
sub getProfilePath
{
	my ($profileName, $dieOnError) = @_;
	
	my $status = 1;
	my $profilePath;
	
	chdirAndValidate($ftsFtsWasAppSrvBinDir, $dieOnError);
	
	$gCaptureExeOutputData = "";		# remove any old output data
	my @mustMatch = $langMsgText->maketext("CANNOT_RETRIVE_PROFILE_PATH");
	my $cmd = $Q.$ftsFtsWasAppSrvBinDir."manageprofiles".$ext.$Q." -getPath -profileName ".$profileName;
	$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, $dieOnError, 1);
	
	if ($status eq 1)
	{
		# All is well
		my $numOfLines = 0;
		my @data = split(/\n/, $gCaptureExeOutputData);
		foreach my $line ( @data )
		{
			if ($numOfLines ne 0)
			{
				# We have more than one line; something went wrong
				dealWithError($langMsgText->maketext("UNEXPECTED_DATA", $gCaptureExeOutputData)."\n", $dieOnError);
				$status = 0;
			}
			$profilePath = $line."/";
			$profilePath =~ s/\\/\//sg;
			$numOfLines++;
		}
	}
	
	return ($status, $profilePath);
}


##############################################################################
# Returns 1 if this is an FTS profile, 0 otherwise
sub isThisFtsProfile
{
	my ($profileName) = @_;
	
	my $webXML = findFtsWebXMLFolder($profileName);
	my $serverXML = findFtsServerXMLFolder($profileName);
	my $val = 0;
	my $status = 0;
	
	if (index($webXML, "cqtsdbcrawler.war/WEB-INF/") ne -1)
	{
		$val++;
	}
	if (index($serverXML, $profileName."/servers/server1/") ne -1)
	{
		$val++;
	}
	
	if ($val == 2)
	{
		$status = 1;
	}
	
	return ($status);
}


##############################################################################
# Returns a list of existing profiles directory
sub getListOfProfilesDir
{
	my ($rootDir) = @_;
	
	my @profilesDirList = ();
	my $DH;
	my $dir;
	
	unless (opendir $DH, $rootDir)
	{
		dealWithError($langMsgText->maketext("NO_SUCH_DIR_FILE", $rootDir)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	
	while ($dir = readdir $DH)
	{
		next if (($dir eq ".") or ($dir eq "..") or (-f $rootDir.$dir));
		push(@profilesDirList, $dir);
	}
	
	return @profilesDirList;
}


##############################################################################
# Find the next WAS prfile port number
sub findNextAvailablePortNumber
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("NEXT_AVAILABLE_PORT")."\n";
	
	if (@ftsExistingProfileNames == 0)
	{
		getListOfProfilesName(1);
	}
	
	my $nextPortNumber = -1;
	my $portNumber = -1;
	
	# loop over the list of profiles to find the 'next' availalbe port number
	foreach my $profileName (@ftsExistingProfileNames)
	{
		$portNumber = findPortNumberForProfile($profileName, $dieOnError);
		
		if ($portNumber > $nextPortNumber)
		{
			$nextPortNumber = $portNumber;
		}
	}
	
	if (($nextPortNumber == -1) and (@ftsExistingProfileNames == 0))
	{
		$nextPortNumber = 15080;	# there are no WAS profiles, default to first one
	}
	else
	{
		$nextPortNumber += 1000;	# use the 'next' port number
		if ($nextPortNumber < 14080)
		{
			$nextPortNumber = 14080;	# start at 14080
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FOUND_NEXT_PORT", $nextPortNumber)."\n";
	
	return $nextPortNumber;
}


##############################################################################
# Returns the port number of a WAS profile
sub findPortNumberForProfile
{
	my ($profileName, $dieOnError) = @_;
	my $status = 0;
	my $profilePath = "";
	
	($status, $profilePath) = getProfilePath($profileName, $dieOnError);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOOKING_IN", $profilePath)."\n";
	
	my $transportPortMatch = $langMsgText->maketext("HTTP_TRANSPORT_PORT")->[0];
	my $aboutFile = $profilePath."/logs/AboutThisProfile.txt";
	my $line;
	my $portNumber = -1;
	
	if (!open(FH, $aboutFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $aboutFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	while ($line = <FH>)
	{
		if ($gCqFtsEncoding ne "")
		{
			$line = Encode::decode($gCqFtsEncoding, $line);
		}
		
		if ($line =~ /$transportPortMatch/)
		{
			my $foundAt = rindex($line, " ");
			if ($foundAt ne -1)
			{
				$line = substr($line, $foundAt);
				
				$portNumber = trim($line);
				last;
			}
		}
	}
	close FH;
	
	return $portNumber;
}

##############################################################################
# Copies files that match a given extension
sub copyFilesMatch
{
	my ($fromDir, $toDir, $extension, $dieOnError) = @_;
	
	my $status = 1;
	
	my @files; 
	if (opendir(DIR, $fromDir))
	{
		while (my $fileName = readdir(DIR))
		{
			if ($fileName =~ m/.$extension$/)
			{
				my $fromFile;
				my $toFile;
				
				$fromFile = $fromDir."/".$fileName;
				$toFile = $toDir."/".$fileName;
				
				if (-f $fromFile)
				{
					if (!copy($fromFile, $toFile))		# returns 0 on failure
					{
						dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $fromFile, $toFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
						$status = 0;
					}
				}
			}
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $fromDir)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}

##############################################################################
# Recursively, copies files and sub-directories from source to destination
sub copyFiles
{
	my ($fromDir, $toDir, $dieOnError) = @_;
	
	my $status = 1;
	
	printAndLog $msgLabel_INFO."mkpath -> [".$toDir."]"."\n";
	File::Path::mkpath $toDir;
	
	if (opendir(DIR, $fromDir))
	{
		my @files = readdir DIR;
		closedir (DIR);	
		
		my $fromFile;
		my $toFile;
		my $file;
		
		foreach $file (@files)
		{
			$fromFile = $fromDir.$file;
			$toFile = $toDir.$file;
			if (-f $fromFile)
			{
				if (!copy($fromFile, $toFile))		# returns 0 on failure
				{
					dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $fromFile, $toFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
					$status = 0;
				}
			}
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $fromDir)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Normalizes the Entity file so that any white spaces are eliminated
sub cleanupEntityFile
{
	my ($entityFile) = @_;
	
	my @data;
	if (!open(FH, $entityFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $entityFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return 0;
	}
	@data = <FH>;
	close(FH);
	
	my $resultData;
	
	my $line;
	foreach $line (@data)
	{
		my @list = split(/=/, $line);
		
		if (scalar(@list) ne 2)
		{
			next;
		}
		
		my $entity = trim($list[0]);
		
		if ($entity eq "")
		{
			next;
		}
		
		my $fields = cleanupFieldsList($list[1]);
		
		$fields =~ s/,+/,/g;	# remove any extra double "," (blank field-names)
		$fields =~ s/^,//g;		# remove any extra start of ','
		$fields =~ s/,$//g;		# remove any extra ending of ','
		
		$resultData .= $entity."=".$fields."\n";
	}
	
	# Now write the changes to the file
	return safeFileUpdate($entityFile, $resultData, 0);
}


##############################################################################
# If the field-list of the Entity file has '&', clean it
sub cleanupFieldsList
{
	my ($fieldsList) = @_;
	
	my $cleanFieldsList = "";
	
	my @fields = split(/,/, $fieldsList);
	
	foreach my $field (@fields)
	{
		my $cleanField = trim($field);
		
		if (($cleanField eq "") or ($cleanField eq "&"))
		{
			next;
		}
		
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		
		$cleanField =~ s/^&\s+/&/;	# remove any space that maybe between "&" and the field name
		
		$cleanFieldsList .= $cleanField;
	}
	
	# if missing, add required fields: dbid, record_type, version
	if (index($cleanFieldsList, "dbid") eq -1)
	{
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		$cleanFieldsList .= "dbid";
	}
	if (index($cleanFieldsList, "record_type") eq -1)
	{
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		$cleanFieldsList .= "record_type";
	}
	if (index($cleanFieldsList, "version") eq -1)
	{
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		$cleanFieldsList .= "version";
	}
	
	return $cleanFieldsList;
}


##############################################################################
# Determines if the list of fields exist (in the CQ DB) or not
sub addDefaultFieldsList
{
	my ($entityName) = @_;
	
	my $user = $ftsUserName;
	my $pwd = $ftsUserPassword;
	my $dbSet = $ftsDBSetName;
	my $dbName = $ftsUserDbName;
	
	if ($pwd eq "\"\"")
	{
		$pwd = "";
	}
	
	my $sessionObj;
	
	eval
	{
		$sessionObj = CQSession::Build();
		
		$sessionObj->UserLogon($user, $pwd, $dbName, $dbSet);
	};
	die "\n"."$@\n" if $@; $@ = "";
	
	if ($sessionObj->IsUserSuperUser() eq 0)
	{
		dealWithError($langMsgText->maketext("NO_PRIVILAGE")."\n", 1);
	}
	
	my $entDef = $sessionObj->GetEntityDef($entityName);
	
	my $fieldsListString = "";
	my ($fieldNamesRef) = $entDef->GetFieldDefNames();
	foreach my $fieldName (@$fieldNamesRef)
	{
		if ($fieldName eq "history")
		{
			next;	# 'history' field is not FTS indexed
		}
		
		$fieldsListString = $fieldsListString.",".$fieldName;
	}
	
	return $fieldsListString;
}


##############################################################################
# Determines if the list of fields exist (in the CQ DB) or not
sub isFieldListExist
{
	my ($recordTypeName, $fieldsList, $dieOnError) = @_;
	
	my @userGivenfieldsList = split(/,/, $fieldsList);
	
	my $found = 0;
	
	my $item;
	foreach  $item (keys %ftsEntityFieldList) 
	{
		my $entityName = $item;
		my $fieldList = $ftsEntityFieldList{$item};
		
		if ($entityName eq "history")
		{
			next;	# 'history' record-type is not indexed
		}
		
		if ($recordTypeName eq $entityName)
		{
			my @dbfieldsList = split(/,/, $fieldList);
			
			my $rt1;
			foreach $rt1 (@userGivenfieldsList)	# walk over the user list
			{
				$rt1 =~ s/^&//;		# remove "&" from the start filed name (if any)
				
				my $rt2;
				foreach $rt2 (@dbfieldsList)		# make sure it's in the DB's list
				{
					if ($rt1 eq $rt2)
					{
						$found++;
						last;
					}
				}
			}
			
			if ($found ne scalar @userGivenfieldsList)
			{
				$found = 0;
			}
		}
	}
	
	if (($found eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("FIELD_NAME_NOT_IN_SCHEMA", $recordTypeName)."\n", 1);
	}
	
	return $found;
}


##############################################################################
# Determines if the list of record-types exist (in the CQ DB) or not
sub isRecordTypeExist
{
	my ($recordTypeName, $dieOnError) = @_;
	
	my $found = 0;
	
	my $item;
	foreach  $item (keys %ftsEntityFieldList) 
	{
		my $entityName = $item;
		
		if ($entityName eq "history")
		{
			next;	# 'history' field is not indexed
		}
		
		if ($recordTypeName eq $entityName)
		{
			$found = 1;
			last;
		}
	}
	
	if (($found eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("RECORD_TYPE_NOT_IN_SCHEMA", $recordTypeName)."\n", 1);
	}
	
	return $found;
}


##############################################################################
# Determines if given record-type is indexed or not
sub isThisRecordTypeIndexed
{
	my ($lookingFor, $dieOnError) = @_;
	
	my $found = 0;
	my $recordType;
	my $indexedRecordTypes = "";
	foreach $recordType (@$ftsListOfIndxedRecordTypes)
	{
		if ($recordType eq $lookingFor)
		{
			$found = 1;
			last;
		}
	}
	
	if (($found eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("RECORD_TYPE_NOT_INDEXED", $lookingFor)."\n", 1);
	}
	
	return $found;
}

##############################################################################
# Determines if given field in the given record-type is indexed or not
sub isThisFieldIndexed
{
	my ($fieldLookingFor, $recordType, $dieOnError) = @_;
	
	my $found = 0;
	my $field;
	my $indexedFields = $ftsHashOfIndexedFields{$recordType};
	foreach $field (@$indexedFields)
	{
		if ($field eq $fieldLookingFor)
		{
			$found = 1;
			last;
		}
	}
	
	if (($found eq 0) and ($dieOnError eq 1))
	{
		# Need add new error message to en.pm
		dealWithError($langMsgText->maketext("FIELD_NOT_INDEXED", $fieldLookingFor, $recordType)."\n", 1);
	}
	
	return $found;
}

##############################################################################
# Updates a file in a safe way; it will make a backup of the current file, 
# write to it, and delete the backup file.
sub safeFileUpdate
{
	my ($fileToUpdate, $data, $dieOnError) = @_;
	
	my $backupFile = $fileToUpdate."backup";
	my $status = 0;		# assume we will fail
	
	# first make sure the file exist
	if (!(-f $fileToUpdate))
	{
		dealWithError($langMsgText->maketext("UPDATE_FILE_NOT_EXIST", $fileToUpdate)."\n".$msgLable_oS_ERROR.$!."\n", 0);
	}
	else
	{
		# make a backup copy of the current file (the one we will be modifying)
		if (!copy($fileToUpdate, $backupFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE_RESTORE_FILE", $fileToUpdate, $backupFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
		else
		{
			# now open the existing file and write to it the new data
			if (open(FH, ">$fileToUpdate"))
			{
				print FH $data;		# write to it and close it
				close FH;
				
				$status = 1;	# all is well
			}
			else
			{
				dealWithError($langMsgText->maketext("CANNOT_WRITE_TO_FILE", $fileToUpdate)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				
				# restore from the backup file
				if (!copy($backupFile, $fileToUpdate))
				{
					my $msg = $langMsgText->maketext("CANNOT_RESTORE_BACKUP_FILE", $backupFile, $fileToUpdate)."\n".$msgLable_oS_ERROR.$!."\n";
					$msg .= "\n\n".$langMsgText->maketext("MANUALY_RESTORE_FILE", $backupFile)."\n";
					
					dealWithError($msg, 0);
				}
			}
			
			# delete the backup file
			if (unlink($backupFile) ne 1)	# returns the number of files deleted (1 in our case)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $backupFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
	}
	
	if (($status eq 0) and ($dieOnError eq 1))
	{
		# we are asked to die, so we will (with an error message)
		dealWithError($langMsgText->maketext("CANNOT_UPDATE_FILE", $fileToUpdate)."\n", 1);
	}
	
	return $status;
}


##############################################################################
# Removes record-types from an Entity file
sub removeRecordTypeFromEntity
{
	my ($recordTypesToRemove) = @_;
	
	my $line;
	my $data;
	if (!open(FH, $ftsEntityFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsEntityFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return 0;
	}
	
	while ($line = <FH>)
	{
		$line =~ s/ //g;	# remove any spaces that maybe in the data of the line
		
		if ($line =~ m/^(.+=)/)
		{
			my $fromEntityRecordType = $1;
			
			$fromEntityRecordType =~ s/=//g;	# remove '=' from the end of the string
			
			# if this record-type (from the file) is in our remove list, then skip it
			my $foundIt = 0;
			my $recordType;
			foreach $recordType (@$recordTypesToRemove)
			{
				if ($recordType eq $fromEntityRecordType)
				{
					printAndLog $msgLabel_INFO.$langMsgText->maketext("REMOVE_FROM_ENTITY", $recordType)."\n";
					
					$foundIt = 1;
					last;
				}
			}
			
			if ($foundIt eq 0)
			{
				$data .= $line;
			}
		}
		
	}
	close FH;
	
	# Now write the changes to the file
	return safeFileUpdate($ftsEntityFile, $data, 0);
}


##############################################################################
# Adds record-types to an Entity file
sub addRecordTypeToEntity
{
	my ($recordTypesAndFieldsToAdd, $dieOnError) = @_;
	
	my $status;
	
	# first clean up the Enity file (so that the addition won't have any problems)
	$status = cleanupEntityFile($ftsEntityFile);
	
	if ($status)
	{
		my @data;
		if (!open(FH, $ftsEntityFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsEntityFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
			return 0;
		}
		@data = <FH>;
		close(FH);
		
		my $alreadyExit = 0;
		foreach my $addLine (@$recordTypesAndFieldsToAdd)
		{
			my ($addRT, $nu1) = split("=", $addLine);
			
			foreach my $fromLine (@data)
			{
				my ($fromRT, $nu2) = split("=", $fromLine);
				
				if ($fromRT eq $addRT)
				{
					$alreadyExit = 1;
					last;
				}
			}
		}
		
		if ($alreadyExit eq 0)
		{
			my $line;
			foreach $line (@$recordTypesAndFieldsToAdd)
			{
				$line .= "\n";
				push(@data, $line);
			}
			
			my $resultData;
			foreach $line (@data)
			{
				$resultData .= $line;
			}
			
			# Now write the changes to the file
			my $status = safeFileUpdate($ftsEntityFile, $resultData, $dieOnError);
			if ($status)
			{
				cleanupEntityFile($ftsEntityFile);
			}
		}
		else
		{
			dealWithError($langMsgText->maketext("ADD_ENTITY_EXIST")."\n", $dieOnError);
		}
	}
	
	if (($status eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("FAILED_TO_ADD_RECORD", $ftsEntityFile)."\n", $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Adds record-types to the CQ search property file
sub addRecordTypeToProperty
{
	my ($recordTypesAndFieldsToAdd, $dieOnError) = @_;
	
	my $status;
	
	# first clean up the Enity file (so that the addition won't have any problems)
	$status = cleanupEntityFile($ftsEntityFile);
	
	my $tempPropertyFileName = $ftsCQPropertyFileName.".added-record-type.temp";
	
	if ($status)
	{
		$status = generateCQSearchPropertyFile($ftsDataDir.$ftsEntityFileName, $ftsDataDir.$tempPropertyFileName, 1);
	}
	
	if (($status eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("FAILED_TO_ADD_RECORD", $ftsEntityFile)."\n", $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Removes record-types from the CQ search property file
sub removeRecordTypeFromCQSearchProperty
{
	my ($recordTypesToRemove) = @_;
	
	my @data;
	if (!open(FH, $ftsCQPropertyFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsCQPropertyFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return 0;
	}
	@data = <FH>;
	close(FH);
	
	my $line;
	my $recordType;
	foreach $recordType (@$recordTypesToRemove)
	{
		my @mData;
		my $matched = 0;
		foreach $line ( @data )
		{
			if ($matched eq 0)
			{
				if ($line =~ m/(.*<CQEntity.*name="$recordType".*)/)
				{
					$matched = 1;
				}
			}
			else
			{
				if ($line =~ m/(.*<\/CQEntity>)/)
				{
					$matched = 0;
					$line = "";
				}
			}
			
			if ($matched eq 0)
			{
				push(@mData, $line);
			}
		}
		
		@data = @mData;
	}
	
	my $resultData = join("", @data);
	
	# Now write the changes to the file
	return safeFileUpdate($ftsCQPropertyFile, $resultData, 0);
}


##############################################################################
# Remove record-types from the Lucene index
sub removeRecordTypeFromIndex
{
	my ($recordTypesToRemove, $whichIndex, $dieOnError) = @_;
	
	my $status = 1;
	
	if ($whichIndex ne "")
	{
		$whichIndex .= "/";
	}
	
	my $recordType;
	foreach $recordType (@$recordTypesToRemove)
	{
		if ($status eq 0)
		{
			last;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("REMOVE_FROM_INDEX", $recordType)."\n";
		
		# An example of what we need to send: http://localhost:14080/solr/update/<delete><query>record_type:????</query></delete>
		# Here "????" is the record-type name
		my $ua = LWP::UserAgent->new;
		
		my $httpPost = "http://".$ftsFtsServerName.":".$ftsFtsWASProfilePort."/solr/".$whichIndex."update";
		
		# Add POST data to HTTP request body
		my $solrDeleteQuery = "<delete><query>record_type:".$recordType."</query></delete>";
		my $postData = { "stream.body" => $solrDeleteQuery, "commit" => "true" };
		my $postLogData = $httpPost."?"."stream.body=".$solrDeleteQuery."&commit=true";
		
		printAndLog $msgLabel_INFO."system -> [".$postLogData."]"."\n";
		
		my $resp = $ua->post($httpPost, $postData);
		if ($resp->is_error())
		{
			printAndLog $resp->status_line."\n", 1;
			dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$postLogData."]"."\n", 0);
			$status = 0;
		}
	}
	
	return $status;
}


##############################################################################
# Creates setup file for add-record-type operation
sub addRecordType_SetupFile
{
	my ($workingDir, $recordTypesAndFieldsToAdd, $dieOnError) = @_;
	
	my $entityFile = $workingDir."addRecordType_Entity.txt";
	my $status = 1;
	
	my $originalCQPropertyData;
	my $addRecordTypeCQPropertyData;
	
	# write the in-memory Entity data to file
	my $data;
	my $line;
	foreach $line (@$recordTypesAndFieldsToAdd)
	{
		$line = trim($line);
		$data .= $line."\n";
	}
	
	if (open(FH, ">".$entityFile))
	{
		print FH $data;
		close(FH);
	}
	else
	{
		$status = 0;
	}
	
	# create the temp CQ Search property file (we only care about the "<CQEntityList>" block from this file)
	if ($status)
	{
		$status = generateCQSearchPropertyFile($entityFile, $workingDir.$ftsCQPropertyFileName, 0);
		if ($status)
		{
			customizeFtsFiles(0, $workingDir.$ftsCQPropertyFileName, $entityFile, 0);
		}
	}
	
	if ($status)
	{
		# open and read the original CQ search property file
		if (open(FH, $ftsCQPropertyFile))
		{
			my $line;
			while ($line = <FH>)
			{
				$originalCQPropertyData .= $line;
			}
			close FH;
		}
		else
		{
			$status = 0;
		}
	}
	
	if ($status)
	{
		# open and read the 'add' record-type CQ search property file
		if (open(FH, $workingDir.$ftsCQPropertyFileName))
		{
			my $line;
			while ($line = <FH>)
			{
				$addRecordTypeCQPropertyData .= $line;
			}
			close FH;
		}
		else
		{
			$status = 0;
		}
	}
	
	#
	# at this point, both the 'add' and 'original' CQ property file are in memory
	#
	
	my $finalCQPropertyData = $originalCQPropertyData;
	my $finalAddCQPropertyData = $originalCQPropertyData;
	
	if ($status)
	{
		# extract "<CQEntityList>" from the 'add' record-type CQ search property file
		my $addBlock = "";
		if ($addRecordTypeCQPropertyData =~ /<CQEntityList>.+<\/CQEntityList>/sg)
		{
			$addBlock = $&;
			
			# some minor clean up
			$addBlock = trim($addBlock);
			$addBlock =~ s/^<CQEntityList>//;
			$addBlock =~ s/<\/CQEntityList>$//;
			$addBlock = "    ".trim($addBlock)."\n";
		}
		else
		{
			$status = 0;
		}
		
		if ($addBlock ne "")
		{
			# insert the new record-type(s) into the original property file
			my $addedCQEntityData = "";
			my $tempData = $originalCQPropertyData;
			
			if ($tempData =~ s/<CQEntityList>.+<\/CQEntityList>//s)
			{
				$addedCQEntityData = $&;
				
				# some minor clean up followed with add-record-type data post-fixing
				$addedCQEntityData = trim($addedCQEntityData);
				$addedCQEntityData =~ s/^<CQEntityList>//;
				$addedCQEntityData =~ s/<\/CQEntityList>$//;
				$addedCQEntityData = "    ".trim($addedCQEntityData)."\n".$addBlock;
				$addedCQEntityData = "<CQEntityList>\n".$addedCQEntityData."  <\/CQEntityList>";
				
				$finalCQPropertyData =~ s/<CQEntityList>.+<\/CQEntityList>/$addedCQEntityData/s;
			}
			else
			{
				$status = 0;
			}
			
			# add the new record-type(s) to the add-property file.
			$addedCQEntityData = "";
			$tempData = $originalCQPropertyData;
			
			if ($tempData =~ s/<CQEntityList>.+<\/CQEntityList>//s)
			{
				$addedCQEntityData = "<CQEntityList>\n".$addBlock."  <\/CQEntityList>";
				
				$finalAddCQPropertyData =~ s/<CQEntityList>.+<\/CQEntityList>/$addedCQEntityData/s;
			}
			else
			{
				$status = 0;
			}
		}
		else
		{
			$status = 0;
		}
	}
	
	#
	# if we are here with no error, it means we can write our 'add' and 'original' CQ property file
	#
	
	if ($status)
	{
		# now write the new 'add' CQ search property file to the file system
		if (open(FH, ">".$workingDir.$ftsCQPropertyFileName))
		{
			print FH $finalAddCQPropertyData;
			close(FH);
		}
		else
		{
			$status = 0;
		}
		
		# final step, add the added record-type to the original CQ property file
		if ($status eq 1)
		{
			$status = safeFileUpdate($ftsCQPropertyFile, $finalCQPropertyData, 0);
		}
	}
	
	return $status;
}


##############################################################################
# This function copies the CQ FTS source template for this deployment
# --copy_fts_template
sub Func_copy_fts_template
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FTS_TEMPLATE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("copy_fts_template");
	
	my $tmp = $ftsTargetDir;	# remember it so we can restore it
	
	printAndLog $msgLabel_INFO."mkpath -> [".$ftsTargetDir."]"."\n";
	File::Path::mkpath $ftsTargetDir;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FTS_TEMPLATE_AT", $ftsTargetDir)."\n";
	
	$ftsTargetDir = $ftsDataDir;
	
	followDir($ftsSourceDir, \&copyDir);
	
	$ftsTargetDir = $tmp;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FTS_TEMPLATE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function generates the Entity file
# --gen_entity_file
sub Func_gen_entity_file
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GENERATE_ENTITY", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("gen_entity_file");
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $entityParm = $parm." -o Entity";
	
	# If add "IncludeUsersGroups", add "-g" to cmd
	if($gIncludeUsersGroups eq 1)
	{
		$entityParm = $entityParm." -g";
	}
	
	my @mustMatch = $langMsgText->maketext("GEN_ENTITY_FILE");
	executeExternalCommand($cmd.$entityParm, @mustMatch, $gDieKind_must_match, 1);
	validateFileGenerated($cmd.$entityParm, $ftsClearQuestHomeDir.$ftsEntityFileName, 1);
	
	if (!move($ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName))
	{
		dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE", $ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName).
						"\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GENERATE_ENTITY", $stat_DONE)."\n", 1;
}


##############################################################################
# This function collects data about the CQ FTS deployment for diagnostic
# --gather_diagnostic_data
sub Func_gather_diagnostic_data
{
	# [<is dbname supported>, <keyword>, <is sitename supported>]
	my @listOfPropertyKeys = (
			[0, "ALLOWED_CLIENT_VERSIONS", 0],
			[1, "ALLOW_INLINE_EDIT", 0],
			[0, "DBID_ALLOC_RETRY_LIMIT", 0],
			[0, "EDIT_MODE_TIMEOUT", 0],
			[0, "ENFORCE_VISIBLE_IN_QUERY", 0],
			[1, "HARD_QUERY_LIMIT", 0],
			[0, "HIDE_CLEAR_SELECT_ALL", 0],
			[0, "HOOKS_SEE_ALL_RECORDS", 0],
			[0, "HOOKS_SEE_ALL_USERS", 0],
			[1, "OSLC_CQREST", 0],
			[1, "OSLC_CQSERVERCONN", 0],
			[1, "OSLC_FRIENDSCONFIG", 0],
			[1, "OSLC_MAPPINGS", 0],
			[1, "OSLC_OAUTHCONSUMERS", 0],
			[1, "OSLC_PROJECTAREALINKS", 0],
			[0, "REMEMBER_PASSWORD", 0],
			[0, "SECURITY_FILTER_REFERENCES", 0],
			[1, "SOFT_QUERY_LIMIT", 0],
			[1, "TEXT_SEARCH_PROPERTY", 1],
			[1, "TYPEAHEAD_PARAMS", 0],
			[0, "USER_READABLE_PROPERTIES", 0]
		);
	
	my $adminSessionObj = 0;
	my $pwd = $ftsUserPassword;
		
	if ($pwd eq "\"\"")
	{
		$pwd = "";
	}
	eval
	{
		$adminSessionObj = CQAdminSession::Build();
		$adminSessionObj->Logon($ftsUserName, $pwd, $ftsDBSetName);
	};
	if ($@)
	{
		my $err = $@;
		dealWithError("\n".$err."\n", 0);
		$adminSessionObj = 0;
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GATHER_DIAG_DATA", $stat_START)."\n", 1;
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	my $diagFolder = $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName.".Diag-".$timeData."/";
	my $diagFolderProfiles = $diagFolder."profiles/";
	
	my $status = 0;
	my $profilePath = "";
	
	($status, $profilePath) = getProfilePath($ftsWASProfileName, 0);
	
	my $offset = length($profilePath) - length($ftsWASProfileName) - 1;
	
	
	# make a copy of the files in "CQ.Search/SAMPL/"
	$status = copyFiles($ftsDataDir, $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/", 0);
	
	# make a copy of the files in "CQ.Search/<dbset>_<userdb>/Solr/solr/conf/"
	$status = copyFiles($ftsDataDir."Solr/solr/conf/", $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/Solr/solr/conf/", 0);
	
	# make a copy of the files in "CQ.Search/<dbset>_<userdb>/Solr/typeahead/conf/"
	$status = copyFiles($ftsDataDir."Solr/typeahead/conf/", $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/Solr/typeahead/conf/", 0);
	
	# make a copy of the file "CQ.Search/<dbset>_<userdb>/Solr/solr/solr.xml"
	copy($ftsDataDir."Solr/solr/solr.xml", $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/Solr/solr/solr.xml");
	
	# make a copy of the file "CQ.Search/<dbset>_<userdb>/Solr/typeahead/solr.xml"
	copy($ftsDataDir."Solr/typeahead/solr.xml", $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/Solr/typeahead/solr.xml");
	
	# make a copy of the files in "CQ.Search/<dbset>_<userdb>/logs/"
	$status = copyFiles($ftsDataDir."logs/", $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/logs/", 0);
	
	# make a copy of "web.xml"
	my $webXMLFolder = findFtsWebXMLFolder($ftsWASProfileName);
	$status = copyFiles($webXMLFolder, $diagFolderProfiles.substr($webXMLFolder, $offset), 0);
	
	# make a copy of "server.xml"
	my $serverXMLFolder = findFtsServerXMLFolder($ftsWASProfileName);
	$status = copyFiles($serverXMLFolder, $diagFolderProfiles.substr($serverXMLFolder, $offset), 0);
	
	# make a copy of CQ Search profile logs
	my $cqsearchprofileLogsFolder = $profilePath."logs/server1/";
	$status = copyFiles($cqsearchprofileLogsFolder, $diagFolderProfiles.substr($cqsearchprofileLogsFolder, $offset), 0);
	
	$cqsearchprofileLogsFolder = $profilePath."logs/";
	$status = copyFiles($cqsearchprofileLogsFolder, $diagFolderProfiles.substr($cqsearchprofileLogsFolder, $offset), 0);
	
	# make a copy of CQ Search profile 'ffdc' if it exist
	my $cqsearchprofileLogsFolderFfdc = $profilePath."logs/ffdc/";
	if (-e $cqsearchprofileLogsFolderFfdc)
	{
		$status = copyFiles($cqsearchprofileLogsFolderFfdc, $diagFolderProfiles.substr($cqsearchprofileLogsFolderFfdc, $offset), 0);
	}
	
	# copy over any .txt, .trc, and .dmp files that are core dump files from FTS
	copyFilesMatch($profilePath, $diagFolderProfiles.substr($profilePath, $offset), ".txt", 0);
	copyFilesMatch($profilePath, $diagFolderProfiles.substr($profilePath, $offset), ".dmp", 0);
	copyFilesMatch($profilePath, $diagFolderProfiles.substr($profilePath, $offset), ".trc", 0);
	
	# get a list of profiles on this machine and write them to a file
	my @profileList = ();
	getListOfProfilesName(0);
	foreach my $profileName (@ftsExistingProfileNames)
	{
		my ($s, $pPath) = getProfilePath($profileName, 0);
		
		push(@profileList, $profileName.": ".$pPath);
	}
	
	# write the data
	my $profileListFile = $diagFolderProfiles."WASProfilesList.txt";
	eval
	{
		my $status = 1;
		if (!open(FH, ">$profileListFile"))
		{
			dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $profileListFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			$status = 0;
		}
		if ($status eq 1)
		{
			foreach my $item (@profileList)
			{
				print FH $item."\n";
			}
			close FH;
		}
	};
	if ($@)
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $profileListFile)."\n", 0);
	}
	
	# get the WAS version and write it to a file
	my $wasAppVerFile = $diagFolderProfiles."WASVersionInfo.txt";
	my $wasAppVerCmd = $Q.$ftsFtsWasAppSrvDir."bin/versionInfo".$ext.$Q;
	my @mustMatch = ();
	$gCaptureExeOutputData = "";
	executeExternalCommand($wasAppVerCmd, \@mustMatch, $gDieKind_must_not_match, 0, 1);
	eval
	{
		my $status = 1;
		if (!open(FH, ">$wasAppVerFile"))
		{
			dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $wasAppVerFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			$status = 0;
		}
		if ($status eq 1)
		{
			print FH $gCaptureExeOutputData."\n";
			close FH;
		}
	};
	if ($@)
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $wasAppVerFile)."\n", 0);
	}
	
	# create a dump of all master-properties and their value
	if ($adminSessionObj ne 0)
	{
		# update the list to include all needed combinations
		@listOfPropertyKeys = updateMasterPropertiesList(@listOfPropertyKeys);
		
		# get the master property data for each master property key
		my @masterPropertiesData = ();
		my $masterPropertiesFile = $diagFolder."MasterProperties.txt";
		foreach my $subArray (@listOfPropertyKeys)
		{
			my $keyName = @$subArray[1];
			my $outputName = "";
			my $dbName = "";
			my $siteName = "";
			my $propertyValue = "";
			
			if (@$subArray[0] eq 1)
			{
				$dbName = $ftsUserDbName;
				$outputName = $dbName.".";
			}
			
			$outputName = $outputName.$keyName;
			
			if (@$subArray[2] eq 1)
			{
				$siteName = $ftsSiteName;
				if ($ftsSiteName eq "<local>")
				{
					$siteName = "";
				}
				$outputName = $outputName.".".$siteName;
			}
			eval
			{
				$propertyValue = $adminSessionObj->GetMasterProperty($keyName, $dbName, $siteName);				
			};
			if ($@)
			{
				$propertyValue = "";
				$outputName = "**".$outputName;
				dealWithError("\nCannot retrieve Master Property: ".$keyName."\n", 0);
			}
			push(@masterPropertiesData, $outputName.": ".$propertyValue."\n====================================");	
		}
		
		# write the data
		eval
		{
			my $status = 1;
			if (!open(FILE, ">$masterPropertiesFile"))
			{
				dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $masterPropertiesFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				$status = 0;
			}
			if ($status eq 1)
			{
				foreach my $item (@masterPropertiesData)
				{
					print FILE $item."\n";
				}
				close FILE;
			}
		};
		if ($@)
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $masterPropertiesFile)."\n", 0);	
		}
		CQAdminSession::Unbuild($adminSessionObj);
	}
	
	# find out if CQ Web is also on this machine, if so collect some data about it
	my ($cqwebProfileName, $cqwebProfileHome) = findProfilePath_cqweb();
	
	if ($cqwebProfileHome ne "")	# if CM is on this machine too, copy some of it's diagnostic data too
	{
		# make a copy of CM profile logs
		my $cmprofileLogsFolder = $cqwebProfileHome."logs/server1/";
		$status = copyFiles($cmprofileLogsFolder, $diagFolderProfiles.$cqwebProfileName."/".substr($cmprofileLogsFolder, length($cqwebProfileHome)), 0);
		
		$cmprofileLogsFolder = $cqwebProfileHome."logs/";
		$status = copyFiles($cmprofileLogsFolder, $diagFolderProfiles.$cqwebProfileName."/".substr($cmprofileLogsFolder, length($cqwebProfileHome)), 0);
		
		# make a copy of CM profile 'ffdc' if it exist
		my $cmprofileLogsFolderFfdc = $cqwebProfileHome."logs/ffdc/";
		if (-e $cmprofileLogsFolderFfdc)
		{
			$status = copyFiles($cmprofileLogsFolderFfdc, $diagFolderProfiles.$cqwebProfileName."/".substr($cmprofileLogsFolderFfdc, length($cqwebProfileHome)), 0);
		}
		
		# copy over any .txt, .trc, and .dmp files that are core dump files from CM
		my $cmprofileCrashFiles = $cqwebProfileHome;
		copyFilesMatch($cmprofileCrashFiles, $diagFolderProfiles.$cqwebProfileName."/".substr($cmprofileCrashFiles, length($cqwebProfileHome)), ".txt", 0);
		copyFilesMatch($cmprofileCrashFiles, $diagFolderProfiles.$cqwebProfileName."/".substr($cmprofileCrashFiles, length($cqwebProfileHome)), ".dmp", 0);
		copyFilesMatch($cmprofileCrashFiles, $diagFolderProfiles.$cqwebProfileName."/".substr($cmprofileCrashFiles, length($cqwebProfileHome)), ".trc", 0);
	}
	
	if ($status eq 0)
	{
		dealWithError($langMsgText->maketext("DIAG_COLLECTION_ERROR")."\n", 0);
	}
	
	printAndLog $msgLabel_INFO."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DIAG_DATA_LOCATION", $diagFolder)."\n";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GATHER_DIAG_DATA", $stat_DONE)."\n", 1;
}


##############################################################################
# This function displays the headline of all scenarios, the headline and the 
# body of a scenario, or the headline and body of all scenarios
# --show_scenarios
sub Func_show_scenarios
{
	my $dataFile = $ftsClearQuestHomeDir."CQTsAdmin/L10N/".$langMsgText->maketext("LANGUAGE_CODE")."-scenarios.xml";
	
	my $parser = new XML::Parser(ErrorContext => 3, Style => "Tree");
	my $tree;
	
	eval
	{
		$tree = $parser->parsefile($dataFile);
	};
	
	buildScenariosFromXMLFile("", $tree);
	
	my $idToShow;
	if (!defined($gShowScenarios) or ($gShowScenarios eq ""))
	{
		$idToShow = "headline";
	}
	else
	{
		$idToShow = $gShowScenarios;
	}
	
	$idToShow = lc($idToShow);
	
	foreach my $key (sort keys %gFtsScenarios)
	{
		my $value = $gFtsScenarios{$key};
		
		if (($idToShow eq "all") or ($idToShow eq "headline") or ($idToShow eq lc($value->{"id"})))
		{
			printAndLog "\n\n", 1;
			printAndLog $langMsgText->maketext("SCENARIO_ID", $value->{"id"}), 1;
			printAndLog "\n", 1;
			printAndLog $langMsgText->maketext("SCENARIO_HEADLINE", $value->{"headline"}), 1;
			printAndLog "\n\n", 1;
		}
		
		if (($idToShow eq "all") or $idToShow eq lc($value->{"id"}))
		{
			printAndLog $value->{"body"}, 1;
			printAndLog "\n\n", 1;
		}
	}
	
	printAndLog "\n", 1;
	
	if ($@)
	{
		# report any error that stopped parsing, or announce success
		$@ =~ s/at \/.*?$//s;
		my $err = $@;
		dealWithError($langMsgText->maketext("ERROR_IN_CFG_FILE", $dataFile)."\n".$err."\n", 1);
	}
}


##############################################################################
# This function will enable verbose output (always on for this release)
# --verbose
sub Func_verbose
{
	# no-op
	$gVerboseOutput = 1;
}


##############################################################################
# This function will scrub CQ oplogs (if multisite is not enabled)
# --scrub_oplog
sub Func_scrub_oplog
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SCRUB_OPLOG", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("scrub_oplog");
	
	if ($gIsWindows eq 1)
	{
		if ($gIsReplicated eq 0)
		{
			chdirAndValidate($ftsClearQuestHomeDir, 1);
			
			my @mustMatch = $langMsgText->maketext("SCRUB_OPLOG");
			my $cmd = "multiutil.exe scruboplog -dbset ".$ftsDBSetName." -family ".$ftsUserDbName." -user ".$ftsUserName." -password ".$ftsUserPassword." -before ".$gScrubOplogBeforeDate;
			executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, 1);
		}
		else
		{
			dealWithError($langMsgText->maketext("WILL_NOT_SCRUB")."\n", 1);
		}
	}
	else
	{
		runMultiutilOnUnix();
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SCRUB_OPLOG", $stat_DONE)."\n", 1;
}


##############################################################################
# This function makes a backup copy of the CQ FTS deployment
# --backup_fts
sub Func_backup_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_BACKUP_CQFTS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("");
	
	$gBackupDestinationDir = File::Spec->rel2abs($gBackupDestinationDir);
	$gBackupDestinationDir =~ s/\\/\//sg;
	
	if (!(-d $gBackupDestinationDir))
	{
		dealWithError($langMsgText->maketext("WRONG_BACKUP_DESTINATION")."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
		# we are given a directory
	if ((substr($gBackupDestinationDir, -1) ne "/") or (substr($gBackupDestinationDir, -1) ne "\\"))
	{
		$gBackupDestinationDir .= "/";
	}
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	
	my $backupDirName = $gBackupDestinationDir.$ftsDBSetName."_".$ftsUserDbName.".Backup-".$timeData."/";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DISABLING_UPDATE_MODE")."\n";
	Func_stop_update_idx(1);
	
	my $tmp_ftsTargetDir = $ftsTargetDir;	# remember it so we can restore it
	my $tmp_ftsSourceDir = $ftsSourceDir;
	
	$ftsTargetDir = $backupDirName;
	$ftsSourceDir = $ftsDataDir;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BACKING_UP_TO", $backupDirName)."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BACKUP_TIME_DEPENDENCY")."\n";
	followDir($ftsSourceDir, \&copyDir);
	
	$ftsTargetDir = $tmp_ftsTargetDir;
	$ftsSourceDir = $tmp_ftsSourceDir;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("ENABLING_UPDATE_MODE")."\n";
	Func_start_update_idx(1);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_BACKUP_CQFTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This functions archives the current CQ FTS depolyment
# --archive_fts
sub Func_archive_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ARCHIVE_CQFTS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("");
	
	# delete the WAS profile
	Func_delete_fts_was_profile(1);
	
	# disable CQ Web FTS
	Func_disable_cqweb_fts(1);
	
	# rename CQ FTS folder
	my $oldDirName = $ftsDataDir;
	
	if (substr($oldDirName, -1) eq "/")
	{
		$oldDirName = substr($oldDirName, 0, length($oldDirName) -1);
	}
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	
	my $archivedDirName = $oldDirName.".Archived-".$timeData;
	printAndLog $msgLabel_INFO.$langMsgText->maketext("RENAMING_TO", $oldDirName, $archivedDirName)."\n";
	
	printAndLog $msgLabel_WARN.$langMsgText->maketext("LOGGER_STOP_FOR_RENAME")."\n", 1;
	closeLogger();
	
	if (!move($oldDirName, $archivedDirName))
	{
		# we can't recover or log on this error, so just die with an error
		print "\n".$gTerminationErrorMsg."\n";
		print "\n";
		die $langMsgText->maketext("FAILED_TO_MOVE_FOLDER", $oldDirName, $archivedDirName)."\n".$msgLable_oS_ERROR.$!."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO."\n";
		printAndLog $msgLabel_INFO.$langMsgText->maketext("ARCHIVED_TO", $archivedDirName)."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ARCHIVE_CQFTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function removes the lock file 'write.lock', if one exist, from the 
# Lucene index
# --remove_lucene_idx_lock
sub Func_remove_lucene_idx_lock
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_LUCENE_IDX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("");
	
	my $lockFile = $ftsLuceneIndexDir."write.lock";
	if (-f $lockFile)
	{
		# the lock file exist, delete it
		if (unlink($lockFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $lockFile)."\n", 1);
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("IDX_IS_UnLOCKED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("IDX_IS_NOT_LOCKED")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_LUCENE_IDX", $stat_DONE)."\n", 1;
}


##############################################################################
# This locks the CQ FTS deployment so that any command line option that may 
# modify any setting is prevented
# --lock_cq_fts
sub Func_clear_state
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CLEAR_STATE", $stat_START)."\n", 1;
	
	if (-f $ftsStateFile)
	{
		# the state file exist, delete it
		if (unlink($ftsStateFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $ftsStateFile)."\n", 1);
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("STATE_IS_CLEARED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("NO_STATE_TO_CLEAR")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CLEAR_STATE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function locks the CQ FTS deployment so that any command line option 
# that may modify any setting is prevented.
# --lock_cq_fts
sub Func_lock_cq_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_LOCK_CQ_FTS", $stat_START)."\n", 1;
	
	if (!(-f $ftsLockFile))
	{
		if (!open(FH, ">$ftsLockFile"))
		{
			dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $ftsLockFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		print FH "";
		close FH;
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_LOCKED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_ALREADY_LOCKED")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_LOCK_CQ_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function Unlocks the CQ FTS deployment so that any command line option 
# that may modify any setting is allowed.
# --unlock_cq_fts
sub Func_unlock_cq_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_CQ_FTS", $stat_START)."\n", 1;
	
	if (-f $ftsLockFile)
	{
		# the lock file exist, delete it
		if (unlink($ftsLockFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $ftsLockFile)."\n", 1);
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_UnLOCKED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_ALREADY_UNLOCKED")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_CQ_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function validates if the CQ FTS deployment is valid or not
# --validate_cq_fts_deplo
sub Func_validate_cq_fts_deplo
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_VALIDATE", $stat_START)."\n", 1;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_VALIDATE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function will prepare a CQ FTS deployment for a CQ upgrade
# --prep_upgd_was_profiles
sub Func_prep_upgd_was_profiles
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_PREP_UPGD_PROFILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("prep_upgd_was_profiles");
	
	# make sure the backup directory exist, if not, create it
	if (!(-d $gPrepUpgdFtsWASProfiles))
	{
		printAndLog $msgLabel_INFO."mkpath -> [".$gPrepUpgdFtsWASProfiles."]"."\n";
		File::Path::mkpath $gPrepUpgdFtsWASProfiles;
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BACKUP_PROFILES_TO", $gPrepUpgdFtsWASProfiles)."\n";
	
	# get list of CQ FTS WAS profiles
	getListOfProfilesName(1);
	
	foreach my $profileName (@ftsExistingProfileNames)
	{
		my $status = 0;
		my $profilePath = "";
		
		($status, $profilePath) = getProfilePath($profileName, 0);
		
		if ($status eq 1)
		{
			# the profile's path which we got back ends with the profile-name, remove it
			my $foundAt = rindex($profilePath, $profileName) + 1;
			if ($foundAt ne -1)
			{
				# get everything from the start up to the path (excluding the profile-name)
				$profilePath = substr($profilePath, 0, $foundAt - 1);
			}
		}
		
		my $isFtsProfile = isThisFtsProfile($profileName);
		if ($isFtsProfile != 1)
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("NOT_FTS_PROFILE", $profileName)."\n";
			next;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("PREPARING_UPGRADE", $profileName)."\n";
		
		my @profileInfoData = ();
		my $backupPath = "";
		my $profilePort = "";
		my $profMaxHeapSize = "";
		my $serviceStartupType = "";
		my $genericJvmArg = "";
		my $ftsHome = "";
		my $profHome = "";
		my $webXMLFile = "";
		my $serverXMLFile = "";
		
		# set to where the profile's upgrade data will be copied
		$backupPath = $gPrepUpgdFtsWASProfiles.$profileName."/";
		
		# get the port number of this profile
		if ($profileName ne "")
		{
			$profilePort = findPortNumberForProfile($profileName, 1);
		}
		
		# get the web.xml file
		if ($profilePort ne "")
		{
			$webXMLFile = findFtsWebXMLFolder($profileName);
			if ($webXMLFile ne "")
			{
				$webXMLFile .= "web.xml";
				$webXMLFile =~ s/\\/\//sg;
			}
		}
		
		# get the server.xml file
		if ($webXMLFile ne "")
		{
			$serverXMLFile = findFtsServerXMLFolder($profileName);
			if ($serverXMLFile ne "")
			{
				$serverXMLFile .= "server.xml";
				$serverXMLFile =~ s/\\/\//sg;
			}
		}
		
		# get the startup type
		if (($gIsWindows eq 1) and ($serverXMLFile ne ""))
		{
			$serviceStartupType = getServiceStartupType($profileName);
		}
		
		# get the max heap size (of the WAS profile) and generic JVM arguments
		if ($serverXMLFile ne "")
		{
			($genericJvmArg, $profMaxHeapSize, $ftsHome) = getHeapSizeAndJVMArg($serverXMLFile);
		}
		
		# get the profile's path
		if ($profilePath ne "")
		{
			$profHome = $profilePath;
		}
		
		# if we collected all the data we need, write it to the backup data file
		if ($serverXMLFile eq "")
		{
			dealWithError($langMsgText->maketext("CANNOT_GATHER_PROFILE", $profilePath, $profileName, $profilePort, $serverXMLFile, $webXMLFile)."\n", 0);
			next;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_BACKUP_OF_PROFILE", $profileName)."'\n";
		
		# make the directory for this profile's data
		printAndLog $msgLabel_INFO."mkpath -> [".$backupPath."]"."\n";
		File::Path::mkpath $backupPath;
		
		# delete any old state file (if any)
		unlink($backupPath."restore.state");
		
		# prepare data to write to the data file
		push(@profileInfoData, "FTS_HOME: ".$ftsHome);
		push(@profileInfoData, "PROF_HOME: ".$profHome);
		push(@profileInfoData, "PROF_NAME: ".$profileName);
		push(@profileInfoData, "PROF_PORT: ".$profilePort);
		push(@profileInfoData, "PROF_MAX_HEAP_SIZE: ".$profMaxHeapSize);		# from: server.xml
		push(@profileInfoData, "WIN_SERVICE_STARTUP_TYPE: ".$serviceStartupType);
		push(@profileInfoData, "GENERIC_JVM_ARG: ".$genericJvmArg);	# from: server.xml
		push(@profileInfoData, "WEB_XML: ".$webXMLFile);			# just restore the file
		push(@profileInfoData, "SERVER_XML: ".$serverXMLFile);		# don't restore this file, use set_solr_home and set_was_max_mem
		
		# write the data
		my $dataFile = $backupPath.$profileName.".txt";
		eval
		{
			$status = 1;	# assume success
			
			if (!open(FH, ">$dataFile"))
			{
				dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $dataFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
				$status = 0;
			}
			if ($status eq 1)
			{
				foreach my $item (@profileInfoData)
				{
					print FH $item."\n";
				}
				close FH;
			}
		};
		if ($@)
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $dataFile)."\n", 1);
			$status = 0;
		}
		
		# make a copy of web.xml and server.xml files into our backup directory
		if (($status eq 1) and !copy($webXMLFile, $backupPath))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $webXMLFile, $backupPath)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			$status = 0;
		}
		if (($status eq 1) and !copy($serverXMLFile, $backupPath))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $serverXMLFile, $backupPath)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			$status = 0;
		}
		
		# finaly, delete the WAS profile if they asked us to
		if (($status eq 1) and ($ftsDeleteFtsWASProfiles eq "true"))
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_FTS_WAS_PROFILE", $profileName)."'\n";
			deleteFtsWASProfile($profileName, 0);
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_PREP_UPGD_PROFILES", $stat_DONE)."\n", 1;
}


##############################################################################
# Restores CQ FTS WAS profiles
# --restore_was_profiles
sub Func_restore_was_profiles
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_RESTORE_PROFILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("restore_was_profiles");
	
	# make sure the backup directory exist, if not, terminate
	if (!(-d $gRestoreFtsWASProfiles))
	{
		dealWithError($langMsgText->maketext("NO_SUCH_DIR_FILE", $gRestoreFtsWASProfiles)."\n", 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("RESTORING_PROFILES_FROM", $gRestoreFtsWASProfiles)."'\n";
	
	# get list of CQ FTS WAS profiles
	my (@profilesDirList) = getListOfProfilesDir($gRestoreFtsWASProfiles);
	foreach my $item (@profilesDirList)
	{
		my $backupDataFile = $gRestoreFtsWASProfiles.$item."/".$item.".txt";
		my $stateFile = $gRestoreFtsWASProfiles.$item."/restore.state";
		my $ftsHome;
		my $profileName = "";
		my $profileHome = "";
		my $oldProfileHome = "";
		my $profilePort = "";
		my $profMaxHeapSize = "";
		my $serviceStartupType = "";
		my $genericJvmArg = "";
		my $webXMLFile = "";
		my $serverXMLFile = "";
		
		# if this profile is processed, move on to the next one
		if (-e $stateFile)
		{
			next;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("RESTORING_PROFILE", $backupDataFile)."'\n";
		
		if (!open(FH, $backupDataFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $backupDataFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			next;
		}
		
		my $line;
		while ($line = <FH>)
		{
			$line =~ s/\s+$//;
			
			if ($line =~ /FTS_HOME: /)
			{
				$ftsHome = $line;
				$ftsHome =~ s{FTS_HOME:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /PROF_HOME: /)
			{
				$profileHome = $line;
				$profileHome =~ s{PROF_HOME:\s(.*?)}{$1}x;
				$oldProfileHome = $profileHome;
			}
			elsif ($line =~ /PROF_NAME: /)
			{
				$profileName = $line;
				$profileName =~ s{PROF_NAME:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /PROF_PORT: /)
			{
				$profilePort = $line;
				$profilePort =~ s{PROF_PORT:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /PROF_MAX_HEAP_SIZE: /)
			{
				$profMaxHeapSize = $line;
				$profMaxHeapSize =~ s{PROF_MAX_HEAP_SIZE:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /WIN_SERVICE_STARTUP_TYPE: /)
			{
				$serviceStartupType = $line;
				$serviceStartupType =~ s{WIN_SERVICE_STARTUP_TYPE:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /GENERIC_JVM_ARG: /)
			{
				$genericJvmArg = $line;
				$genericJvmArg =~ s{GENERIC_JVM_ARG:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /WEB_XML: /)
			{
				$webXMLFile = $line;
				$webXMLFile =~ s{WEB_XML:\s(.*?)}{$1}x;
				$webXMLFile =~ s/\\/\//sg;
			}
			elsif ($line =~ /SERVER_XML: /)
			{
				$serverXMLFile = $line;
				$serverXMLFile =~ s{SERVER_XML:\s(.*?)}{$1}x;
				$serverXMLFile =~ s/\\/\//sg;
			}
		}
		close FH;
		
		if ($optFtsWasProfileHome ne "")	# did the user provide us with a new FTS WAS profile home?
		{
			# yes, use it instead of where it was originaly created
			$profileHome = $optFtsWasProfileHome;
		}
		
		if ($profileHome eq "")
		{
			# no profile home is giving, use the default
			$profileHome = $ftsWASProfileHomeDir;
		}
		
		# if we collected all the data we need, we can now restore the WAS profile
		if (($profileName eq "") or ($profilePort eq ""))	# CQ 8.0 does not require 'GENERIC_JVM_ARG' to exist
		{
			dealWithError($langMsgText->maketext("CANNOT_GATHER_PROFILE", $profileHome, $profileName, $profilePort, $genericJvmArg, $serverXMLFile, $serviceStartupType)."\n", 0);
			next;
		}
		
		my $status = 1;
		
		# when re-creating the old WAS profile, we must re-deploy the original Solr version into it
		my ($curDepSolrVer, $curDepDbSetName, $curDepUserDbName, $curCfgFile) = getCurrDeployInfoForSolrUpgrade($ftsHome);
		
		# recreate the WAS profile
		$status = createFtsWASProfile($ftsHome, $profileName, $profilePort, $serviceStartupType, $profileHome, $curDepSolrVer, 0);
		
		# restore the web.xml file
		if ($status eq 1)
		{
			my $fromWebXmlFile = $gRestoreFtsWASProfiles.$item."/"."web.xml";
		
			$webXMLFile = findFtsWebXMLFolder($profileName)."web.xml";
			
			if (!copy($fromWebXmlFile, $webXMLFile))
			{
				dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $fromWebXmlFile, $webXMLFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				$status = 0;
			}
		}
		
		# set WAS max memory
		if ($status eq 1)
		{
			if ($profMaxHeapSize ne "")
			{
				$status = startFtsWASProfile($profileName, 0);
				if ($status eq 1)
				{
					$status = setWASMaxMem($profileName, $profMaxHeapSize, 0);
				}
			}
		}
		
		# restart the WAS profile so all changes take effect
		if ($status eq 1)
		{
			$status = stopFtsWASProfile($profileName, 0);
			if ($status eq 1)
			{
				$status = startFtsWASProfile($profileName, 0);
			}
		}
		
		# update the cqtsadmin-*.xml file
		if (($status eq 1) and ($optFtsWasProfileHome ne ""))	# check to see if they changed the profile's home, if so update our config file
		{
			my $ftsCfg = $ftsHome."cqtsadmin"."-".$curDepDbSetName."-".$curDepUserDbName.".xml";
			
			if (!open(FH, $ftsCfg))
			{
				dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsCfg)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				$status = 0;
			}
			else
			{
				my $index = index($oldProfileHome, $profileName);
				if ($index != -1)
				{
					$oldProfileHome = substr($oldProfileHome, 0, $index);
				}
				
				# look for: <defaultValue>$oldProfileHome</defaultValue>
				# and repalce it with: <defaultValue>$profileHome</defaultValue>
				my $isUpdated = 0;
				my $data;
				my $line;
				while ($line = <FH>)
				{
					$index = index($line, "<defaultValue>".$oldProfileHome."</defaultValue>");
					if ($index == -1)
					{
						$index = index($line, "<newValue required=\"no\">".$oldProfileHome."</newValue>");
					}
					if ($index != -1)
					{
						$index = index($line, $oldProfileHome);
						substr($line, $index, length($oldProfileHome), $profileHome);
						$isUpdated = 1;
					}
					
					$data .= $line;
				}
				
				if ($isUpdated == 1)
				{
					$status = safeFileUpdate($ftsCfg, $data, 0);
				}
			}
		}
		
		# bring up to date cqtsadmin-<dbset>-<userdb>.xml file
		if ($status eq 1)
		{
			$status = updateFile_cqtsadmin_xml($ftsHome, $curDepDbSetName, $curDepUserDbName, 0);
		}
		
		# bring up to date Solr's solr.xml file
		if ($status eq 1)
		{
			$status = updateFile_solr_xml($ftsHome, $curDepDbSetName, $curDepUserDbName, 0);
		}
		
		# copy over 'typeahead' directory
		if ($status eq 1)
		{
			$status = copyDirectory_Solr_typeahead($ftsHome, $curDepDbSetName, $curDepUserDbName, 0);
		}
		
		# update the CQ FTS property file so it has the typeahead related data (if it is missing)
		if ($status eq 1)
		{
			my $propertyFile = $ftsHome."CQ-".$curDepDbSetName."-".$curDepUserDbName.".xml";
			my $entityFile =  $ftsHome."Entity-".$curDepDbSetName."-".$curDepUserDbName.".txt";
			
			$status = insertTypeaheadPropertyData($propertyFile, $entityFile, 0);
		}
		
		# set the state, for this profile that we are done
		if ($status eq 1)
		{
			if (!open(FH, ">$stateFile"))
			{
				dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $stateFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
		close FH;
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_RESTORE_PROFILES", $stat_DONE)."\n", 1;
}


##############################################################################
# --upgrade_solr_app
sub Func_upgrade_solr_app
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UPGRADE_SOLR_APP", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("upgrade_solr_app");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $upgradeWhichDeployment = $ftsDBSetName."_".$ftsUserDbName;
	
	my $allDeploymentsCheck = $ENV{"CQ_FTS_UpgradeAllDeployments"};	# undocumented
	if (defined($allDeploymentsCheck) and $allDeploymentsCheck eq "alldeployments")
	{
		# Undocumented; will let you upgrade Solr for all FTS deployments at once (a re-index might still be required)
		$upgradeWhichDeployment = "";
	}
	
	# get list of CQ FTS WAS profiles
	getListOfProfilesName(1);
	
	# for each profile name, find out if it's a FTS deployment and if it needs to be upgraded
	foreach my $profileName (@ftsExistingProfileNames)
	{
		my $status = 0;
		my $profilePath = "";
		
		($status, $profilePath) = getProfilePath($profileName, 0);
		if ($status eq 1)
		{
			# the profile's path which we got back ends with the profile-name, remove it
			my $foundAt = rindex($profilePath, $profileName) + 1;
			if ($foundAt ne -1)
			{
				# get everything from the start up to the path (excluding the profile-name)
				$profilePath = substr($profilePath, 0, $foundAt - 1);
			}
		}
		
		my $isFtsProfile = isThisFtsProfile($profileName);
		if ($isFtsProfile != 1)
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("NOT_FTS_PROFILE", $profileName)."\n";
			next;
		}
		
		my $serverXMLFile = "";
		my $genericJvmArg = "";
		my $profMaxHeapSize = "";
		my $ftsHome = "";
		
		# get the server.xml file
		$serverXMLFile = findFtsServerXMLFolder($profileName);
		if ($serverXMLFile ne "")
		{
			$serverXMLFile .= "server.xml";
			$serverXMLFile =~ s/\\/\//sg;
			
			($genericJvmArg, $profMaxHeapSize, $ftsHome) = getHeapSizeAndJVMArg($serverXMLFile);
		}
		
		my ($curDepSolrVer, $curDepDbSetName, $curDepUserDbName, $curCfgFile) = getCurrDeployInfoForSolrUpgrade($ftsHome);
		
		if ($curDepSolrVer eq $gSolrUpgradeToVersion)
		{
			printAndLog $msgLabel_INFO."\n", 1;
			printAndLog $msgLabel_INFO.$langMsgText->maketext("SOLR_IS_ALREADY_CURRENT", $curDepDbSetName, $curDepUserDbName, $curDepSolrVer)."\n";
			printAndLog $msgLabel_INFO."\n", 1;
			next;
		}
		
		# We will either upgrade one deployment or all of them
		if (($upgradeWhichDeployment eq "") or ($curDepDbSetName."_".$curDepUserDbName eq $upgradeWhichDeployment))
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("PREPARING_SOLR_UPGRADE", $profileName)."\n";
			
			upgradeSolrApp($curDepDbSetName, $curDepUserDbName, $profileName, $curCfgFile, $curDepSolrVer, $gSolrUpgradeToVersion, $dieOnError);
			
			if (mustReIndexSolrData($curDepSolrVer, $gSolrUpgradeToVersion) == 1)
			{
				$gShowMustReIndexMsg = 1;	# Re-index is required
				
				# Reset the batch indexer's pointer (by deleting the 'recovery' file) so that '--run_batch_idx' starts from the beginning
				my $theFile = $ftsHome."batch_recovery.properties";
				if (-e $theFile)
				{
					printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETE_BATCH_MARKER_FILE", $theFile)."\n";
					if (unlink($theFile) ne 1)
					{
						dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 0);
					}
				}
			}
		}
		else
		{
			if ($upgradeWhichDeployment eq "")
			{
				printAndLog $msgLabel_INFO.$langMsgText->maketext("SKIP_SOLR_UPGRADE", $curDepDbSetName, $curDepUserDbName, $curDepSolrVer)."\n";
			}
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UPGRADE_SOLR_APP", $stat_DONE)."\n", 1;
}


##############################################################################
# This function deletes a CQ FTS WAS profile.
# --delete_fts_was_profile
sub Func_delete_fts_was_profile
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DELETE_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("delete_fts_was_profile");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = 0;
	
	$status = deleteFtsWASProfile($ftsWASProfileName, $dieOnError);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DELETE_WAS_PROFILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This function creates a WAS profile.  Requires one of the following 
# parameter: manual, automatic, or disabled.  The parameter is only valid on 
# Windows.
# --create_fts_was_profile
sub Func_create_fts_was_profile
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CREATE_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("create_fts_was_profile");
	
	createFtsWASProfile($ftsDataDir, $ftsWASProfileName, $ftsFtsWASProfilePort, $gWinServiceStartUpType, $ftsWASProfileHomeDir, $ftsSolrVersion, 1);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CREATE_WAS_PROFILE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function runs the repair-mode record extractor
# --repair_records
sub Func_repair_records
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REINDEX_RECORDS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("repair_records");
	
	if ($ftsBatchIndexJVMParm ne "")
	{
		$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
	}
	
	my $cmdPrefix = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -repair";
	my $cmdPostFix = $Q.$ftsCQPropertyFile.$Q." -site \"".$ftsSiteName."\"";
	
	if (!(-e $ftsCQPropertyFile))
	{
		dealWithError($langMsgText->maketext("MISSING_CONFIG_FILE", $ftsCQPropertyFile)."\n", 1);
	}
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $file;
	my @files = <$ftsDataDir*.xml>;
	foreach $file (@files)
	{
		my $fileName;
		my $foundAt = rindex($file, "/") + 1;
		if ($foundAt ne -1)
		{
			$fileName = substr($file, $foundAt);
		}
		
		my $recordType;
		foreach $recordType (@$ftsListOfIndxedRecordTypes)
		{
			if ($fileName =~ m/^$recordType\d+\.xml/)
			{
				printAndLog $msgLabel_INFO.$langMsgText->maketext("RE_INDEXING_REPAIR", $file)."\n";
				
				my $cmd = $cmdPrefix." ".$Q.$file.$Q." ".$cmdPostFix;
				my @mustMatch = $langMsgText->maketext("REPAIR_RECORDS");
				my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, 0);
				
				if ($status)
				{
					# delete the file
					if (unlink($file) ne 1)
					{
						dealWithError($langMsgText->maketext("CANNOT_DELETE", $file)."\n", 0);
					}
				}
			}
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REINDEX_RECORDS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function optimizes the index
# --optimize_idx
sub Func_optimize_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_OPTIMIZE_INDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("optimize_idx");
	
	optimizeIndex("normal");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_OPTIMIZE_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function adds record-types
# --add_record_type
sub Func_add_record_type
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ADD_RECORD_TYPE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("");
	
	# the data is in "record1=field1,field2,field3;record2=field1,field2" format so our splits are baed on ";", "=" and ",".
	my @rawRecordTypesAndFieldsToAdd = split(/;/, $ftsAddRecordType);
	
	if (scalar(@rawRecordTypesAndFieldsToAdd) eq 0)
	{
		dealWithError($langMsgText->maketext("NO_RECORD_TYPES_TO_ADD")."\n", 1);
	}
	
	my @cleanRecordTypesAndFieldsToAdd;
	
	foreach my $recordTypeBlocks (@rawRecordTypesAndFieldsToAdd)
	{
		my @recordTypeAndItsFields = split(/=/, $recordTypeBlocks);
		
		my $recordTypeName;
		my $fieldsList = "";
		
		if (scalar(@recordTypeAndItsFields) eq 2)
		{
			$fieldsList = $recordTypeAndItsFields[1];
		}
		
		$recordTypeName = $recordTypeAndItsFields[0];
		
		$recordTypeName = trim($recordTypeName);
		$fieldsList = trim($fieldsList);
		
		if ($fieldsList eq "")
		{
			$fieldsList = addDefaultFieldsList($recordTypeName);
		}
		
		$fieldsList = cleanupFieldsList($fieldsList);
		
		isRecordTypeExist($recordTypeName, 1);
		isFieldListExist($recordTypeName, $fieldsList, 1);
		
		# check to see if this record type is already index, if so we die
		if (isThisRecordTypeIndexed($recordTypeName, 0) == 1)
		{
			dealWithError($langMsgText->maketext("RECORD_TYPE_ALREADY_INDEXED", $recordTypeName)."\n", 1);
		}
		
		push(@cleanRecordTypesAndFieldsToAdd, $recordTypeName."=".$fieldsList);
	}
	
	##########################################################################
	
	# check and set state
	my ($function, $state) = getState();
	
	if (($function ne "") and ($function ne "add_record_type"))
	{
		# we are in a state, and it's not our state
		dealWithError($langMsgText->maketext("IN_A_STATE_ALREADY", "add_record_type", $function)."\n", 0);
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
	elsif ($function eq "")
	{
		# we are not in any state, so set one to our state of 'setup_cq_fts'
		setState("add_record_type", "");
	}
	
	$gTopLevelSate = "add_record_type";
	
	##########################################################################
	
	# it's time to add record-type(s)
	my $status = 1;
	my $isIndexOk = 1;
	
	# create a restore poing of the files we are about to modify (in case we want to role back)
	# entity, property, intermediate, solr
	my @restoreFileList = ($ftsEntityFile, $ftsCQPropertyFile, $ftsSolrIntermediateFile, $ftsSolrFile);
	my $restoreExtension = ".restore.add_record_type";
	my $restoreFile;
	
	if (index($state, "restore_point") eq -1)
	{
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile, $restoreFile.$restoreExtension))
			{
				$status = 0;
				dealWithError($langMsgText->maketext("CANNOT_CREATE_RESTORE_FILE", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				last;
			}
		}
		
		if (! $status)
		{
			# we can't craete a restore point, so get out
			foreach $restoreFile (@restoreFileList)
			{
				if (unlink($restoreFile.$restoreExtension) ne 1)
				{
					dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n", 0);
				}
			}
			
			# we have an error, so terminate
			dealWithError($langMsgText->maketext("ADD_RECORD_ERROR")."\n", 1);
		}
		
		setDoneState("restore_point");
	}
	
	#
	# we have our restore point files in place, we are good to go
	#
	
	my $enableUpdateIdx = 0;
	
	# stop update mode
	if (($status) and (index($state, "stop_update_idx") eq -1))
	{
		Func_stop_update_idx(1);
		$enableUpdateIdx = 1;
		
		setDoneState("stop_update_idx");
	}
	
	##########################################################################
	# phase-1: add new record-type to: Entity, and schema.xml
	
	# add the new record-types to the existing Entity file
	if (($status) and (index($state, "add_to_entity") eq -1))
	{
		$status = addRecordTypeToEntity(\@cleanRecordTypesAndFieldsToAdd, 1);
		
		setDoneState("add_to_entity");
	}
	
	# re-generate intermediate Solr file
	if (($status) and (index($state, "add_to_intermediate_solr") eq -1))
	{
		$status = generateSolrIntermediateFile($ftsEntityFile, 0);
		
		setDoneState("add_to_intermediate_solr");
	}
	
	# customize the Solr file to include those new fields
	if (($status) and (index($state, "cust_solr_files") eq -1))
	{
		$status = Func_cust_solr_files(0, 1);
		
		setDoneState("cust_solr_files");
	}
	
	##########################################################################
	# phase-2: add new record-type to a specialy Entity, and CQ Search property file
	
	my $workingDir;
	
	$workingDir = $ftsDataDir."addRecordType.temp"."/";
	printAndLog $msgLabel_INFO."mkpath -> [".$workingDir."]"."\n";
	File::Path::mkpath $workingDir;
	
	# create a specialy Entity and CQ Search property file which will hold to-be-added record-types
	if (($status) and (index($state, "update_master_and_temp_property_file") eq -1))
	{
		$status = addRecordType_SetupFile($workingDir, \@cleanRecordTypesAndFieldsToAdd, 0);
		
		setDoneState("update_master_and_temp_property_file");
	}
	
	# restart WAS profile so that Solr picks up the changes to schema.xml
	if (($status) and (index($state, "restart_was_profile") eq -1))
	{
		$status = stopFtsWASProfile($ftsWASProfileName, 0);
		
		if ($status eq 1)
		{
			$status = startFtsWASProfile($ftsWASProfileName, 0);
			
			setDoneState("restart_was_profile");
		}
	}
	
	# send the new CQ search property file to CQ core 
	# (so that users can start searching on new record-types as they are added)
	if (($status) and (index($state, "enable_cqweb_fts") eq -1))
	{
		$status = Func_enable_cqweb_fts(0, 1);
		
		setDoneState("enable_cqweb_fts");
	}
	
	##########################################################################
	# phase-3: start indexing those new record-types
	
	# index newly added record-types
	if (($status) and (index($state, "run_batch_mode") eq -1))
	{
		if ($ftsBatchIndexJVMParm ne "")
		{
			$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
		}
		
		my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -batch ".$Q.$workingDir.$ftsCQPropertyFileName.$Q." -site "."\"".$ftsSiteName."\"";
		
		chdirAndValidate($ftsClearQuestHomeDir, 1);
		my @mustMatch = (": cqtsdbcrawer finished at ", ": cqtsdbcrawer  ", ": cqtsdbcrawer w_ ", ": cqtsdbcrawer  ");
		$status = executeExternalCommand($cmd, \@mustMatch, $gDieKind_must_match, 0);
		
		setDoneState("run_batch_mode");
	}
	
	##########################################################################
	
	# post record-type addition, deal with success or error
	if ($status)
	{
		# all is well, delete the restore files
		foreach $restoreFile (@restoreFileList)
		{
			if (unlink($restoreFile.$restoreExtension) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
		
		# delete the working directory (and the files in it)
		my @files = <$workingDir*>;
		foreach my $file (@files)
		{
			# clean up the temporary port file.
			if (unlink($file) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE", $file)."\n", 0);
			}
		}
		
		if (!rmdir($workingDir))
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE_FOLDER", $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
	}
	else
	{
		# some action failed, report an error (and terminate if the index was not modified, and it's OK)
		dealWithError($langMsgText->maketext("ERROR_DURING_ADD_RECORD")."\n", ! $isIndexOk);
		
		# restore from the backup
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile.$restoreExtension, $restoreFile))
			{
				dealWithError($langMsgText->maketext("FILE_RESTORE_ERROR", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
		
		# restore CQ Web search property
		Func_enable_cqweb_fts(0, 1);
		
		# if we run into an error and the index was modified, then we have an issue
		if (! $isIndexOk)
		{
			dealWithError($langMsgText->maketext("OUT_OF_SYNC_LUCENE_INDEX")."\n", $isIndexOk);
		}
	}
	
	##########################################################################
	
	# finally, re-start update mode (if all is well, or if we stopped earlier)
	if (($status or $enableUpdateIdx) and (index($state, "start_update_idx") eq -1))
	{
		Func_start_update_idx(1);	# start update mode
		
		setDoneState("start_update_idx");
	}
	
	setDoneState("");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ADD_RECORD_TYPE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function removes record-types
# --remove_record_type
sub Func_remove_record_type
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REMOVE_RECORD_TYPE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("");
	
	##########################################################################
	
	# check and set state
	my ($function, $state) = getState();
	my $newState = 0;
	
	if (($function ne "") and ($function ne "remove_record_type"))
	{
		# we are in a state, and it's not our state
		dealWithError($langMsgText->maketext("IN_A_STATE_ALREADY", "remove_record_type", $function)."\n", 0);
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
	elsif ($function eq "")
	{
		# we are not in any state, set it ours
		$newState = 1;
	}
	
	$gTopLevelSate = "remove_record_type";
	
	##########################################################################
	
	# validate that record-type names are valid and exist for removal
	my @recordTypesToRemove = split(/;/, $ftsRemoveRecordType);
	
	if (scalar(@recordTypesToRemove) eq 0)
	{
		dealWithError($langMsgText->maketext("NO_RECORD_TYPES_TO_REMOVE")."\n", 1);
	}
	
	if ($newState eq 1)
	{
		# this is not a continuation of the state, so do a simple check
		foreach my $recordType (@recordTypesToRemove)
		{
			isThisRecordTypeIndexed($recordType, 1);
		}
		
		setState("remove_record_type", "");
	}
	
	##########################################################################
	
	my $status = 1;
	my $isIndexOk = 1;
	
	# create a restore point of files we are about to modify:
	# entity, property, intermediate, solr
	my @restoreFileList = ($ftsEntityFile, $ftsCQPropertyFile, $ftsSolrIntermediateFile, $ftsSolrFile);
	my $restoreExtension = ".restore.remove_record_type";
	my $restoreFile;
	
	if (index($state, "restore_point") eq -1)
	{
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile, $restoreFile.$restoreExtension))
			{
				$status = 0;
				dealWithError($langMsgText->maketext("CANNOT_CREATE_RESTORE_FILE", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				last;
			}
		}
		
		if (! $status)
		{
			# we can't create a restore point, so get out
			foreach $restoreFile (@restoreFileList)
			{
				if (unlink($restoreFile.$restoreExtension) ne 1)
				{
					dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n", 0);
				}
			}
			
			# we have an error, so terminate
			dealWithError($langMsgText->maketext("REMOVE_RECORD_ERROR")."\n", 1);
		}
		
		setDoneState("restore_point");
	}
	
	#
	# we have our restore point files in place, we are good to go
	# everything we do from here on, will end up modifying files and settings
	#
	
	my $enableUpdateIdx = 0;
	
	# stop update mode
	if (($status) and (index($state, "stop_update_idx") eq -1))
	{
		Func_stop_update_idx(1);
		$enableUpdateIdx = 1;
		
		setDoneState("stop_update_idx");
	}
	
	# remove it from the Entity file
	if (($status) and (index($state, "remove_from_entity") eq -1))
	{
		$status = removeRecordTypeFromEntity(\@recordTypesToRemove);
		
		setDoneState("remove_from_entity");
	}
	
	# remove it from the CQ search property file
	if (($status) and (index($state, "remove_from_cq_property") eq -1))
	{
		$status = removeRecordTypeFromCQSearchProperty(\@recordTypesToRemove);
		
		setDoneState("remove_from_cq_property");
	}
	
	# re-generate intermediate Solr file
	if (($status) and (index($state, "remove_from_intermediate_solr") eq -1))
	{
		$status = generateSolrIntermediateFile($ftsEntityFile, 0);
		
		setDoneState("remove_from_intermediate_solr");
	}
	
	# customize the Solr file to include those new fields
	if (($status) and (index($state, "cust_solr_files") eq -1))
	{
		$status = Func_cust_solr_files(0, 1);
		
		setDoneState("cust_solr_files");
	}
	
	# send the new CQ search property file to CQ core
	if (($status) and (index($state, "enable_cqweb_fts") eq -1))
	{
		$status = Func_enable_cqweb_fts(0, 1);
		
		setDoneState("enable_cqweb_fts");
	}
	
	# remove records of this record-type fromt the index
	if (($status) and (index($state, "remove_from_idx") eq -1))
	{
		$status = removeRecordTypeFromIndex(\@recordTypesToRemove, "", 0);	# delete it from Lucene's index
		
		if (($status == 1) and ($ftsEnableTypeahead eq "true"))
		{
			$status = removeRecordTypeFromIndex(\@recordTypesToRemove, "typeahead", 0);	# delete it from Lucene's index
			
			$isIndexOk = $status;	# remeber if the index is OK or not
		}
		else
		{
			$isIndexOk = $status;	# remeber if the index is OK or not
		}
		
		setDoneState("remove_from_idx");
	}
	
	##########################################################################
	
	# post record-type removal, deal with success or error
	if ($status)
	{
		# all is well, delete the restore files
		foreach $restoreFile (@restoreFileList)
		{
			if (unlink($restoreFile.$restoreExtension) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n", 0);
			}
		}
	}
	else
	{
		# some action failed, report an error (and terminate if the index was not modified, and it's OK)
		dealWithError($langMsgText->maketext("ERROR_DURING_REMOVE_RECORD")."\n", ! $isIndexOk);
		
		# restore from the backup
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile.$restoreExtension, $restoreFile))
			{
				dealWithError($langMsgText->maketext("FILE_RESTORE_ERROR", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
		
		# restore CQ Web search property
		Func_enable_cqweb_fts(0, 1);
		
		# if we run into an error and the index was modified, then we have an issue
		if (! $isIndexOk)
		{
			dealWithError($langMsgText->maketext("OUT_OF_SYNC_LUCENE_INDEX")."\n", $isIndexOk);
		}
	}
	
	##########################################################################
	
	# finally, re-start update mode (if all is well, or if we stopped earlier)
	if (($status or $enableUpdateIdx) and (index($state, "start_update_idx") eq -1))
	{
		Func_start_update_idx(1);	# start update mode
		
		setDoneState("start_update_idx");
	}
	
	setDoneState("");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REMOVE_RECORD_TYPE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function forces re-execution of the update mode as if it's being run
# for the first time.
# --fresh_update_idx
sub Func_fresh_update_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_UPDATE_REINDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("fresh_update_idx");
	
	my $theFile = $ftsDataDir."update_recovery.properties";
	
	if (-e $theFile)
	{
		stopFtsWASProfile($ftsWASProfileName, 1);
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETE_OPLOG_MARKER_FILE", $theFile)."\n";
		if (unlink($theFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 1);
		}
		
		startFtsWASProfile($ftsWASProfileName, 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_UPDATE_REINDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function forces re-execution of the batch mode as if it's being run
# for the firest time.
# --fresh_batch_idx
sub Func_fresh_batch_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_BATCH_REINDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("fresh_batch_idx");
	
	my $theFile = $ftsDataDir."batch_recovery.properties";
	
	if (-e $theFile)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETE_BATCH_MARKER_FILE", $theFile)."\n";
		if (unlink($theFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 1);
		}
	}
	
	Func_run_batch_idx();
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_BATCH_REINDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function sets the JVM maximum memory of the WAS profile
# --set_was_max_mem
sub Func_set_was_max_mem
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_WAS_MAX_JVM_MEM", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("");
	
	if ($ftsWasProfileMaxHeapSize ne "")
	{
		setWASMaxMem($ftsWASProfileName, $ftsWasProfileMaxHeapSize, 1);
	}
	else
	{
		# they are removing JVM setting, the only way to do this is to actualy modify server.xml file
		my $serverXMLFile = findFtsServerXMLFolder($ftsWASProfileName);
		if ($serverXMLFile ne "")
		{
			$serverXMLFile .= "server.xml";
			$serverXMLFile =~ s/\\/\//sg;
			
			removeHeapSizeXMLTag($serverXMLFile); 
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_WAS_MAX_JVM_MEM", $stat_DONE)."\n", 1;
}


##############################################################################
# This function will update the CQ property file
# --update_fts_prop_files
sub Func_update_fts_prop_files
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UPDATE_FTS_FILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("update_fts_prop_files");
	
	customizeFtsFiles(1, $ftsCQPropertyFile, $ftsEntityFile, 1);		# this will sync the CQ Search property file with the CQ FTS config file
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UPDATE_FTS_FILES", $stat_DONE)."\n", 1;
}


##############################################################################
# This function generates the Entity file (if one is not provided), the Solr
# intermediate file, and the CQ Search property file.
# --gen_fts_files
sub Func_gen_fts_files
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GEN_FTS_FILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("gen_fts_files");
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	my $entityFile;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	# If they provided an entity file, then don't generate one and use theirs instead
	if ((defined $ftsUserDefinedEntityFile) and (length($ftsUserDefinedEntityFile) != 0))
	{
		$entityFile = $ftsUserDefinedEntityFile;
		printAndLog $msgLabel_INFO.$langMsgText->maketext("USING_USER_ENTITY_FILE", $entityFile)."\n";
	}
	else
	{
		my $entityParm = $parm." -o Entity";
		
		# If add "IncludeUsersGroups", add "-g" to cmd
		if($gIncludeUsersGroups eq 1)
		{
			$entityParm = $entityParm." -g";
		}
		
		my @mustMatch = $langMsgText->maketext("GEN_ENTITY_FILE");
		executeExternalCommand($cmd.$entityParm, @mustMatch, $gDieKind_must_match, 1);
		validateFileGenerated($cmd.$entityParm, $ftsClearQuestHomeDir.$ftsEntityFileName, 1);
		
		if (move($ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName) ne 1)
		{
			dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE", $ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName).
							"\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		
		validateFileGenerated("move ".$ftsClearQuestHomeDir.$ftsEntityFileName." ".$ftsDataDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName, 1);
		
		$entityFile = $ftsEntityFile;
	}
	
	generateSolrIntermediateFile($entityFile, 1);
	generateCQSearchPropertyFile($entityFile, $ftsDataDir.$ftsCQPropertyFileName, 1);
	
	setDoneState("gen_fts_files");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GEN_FTS_FILES", $stat_DONE)."\n", 1;
}


##############################################################################
# This function customizes the CQ Search property file
# --cust_fts_files
sub Func_cust_fts_files
{
	my ($isInUpdateMode) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_PROPERTY_FILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("cust_fts_files");
	
	if (!defined($isInUpdateMode))
	{
		$isInUpdateMode = 0;	# use default
	}
	
	customizeFtsFiles($isInUpdateMode, $ftsCQPropertyFile, $ftsEntityFile, 1);
	
	setDoneState("cust_fts_files");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_PROPERTY_FILE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function customizes the Solr schema.xml file
# --cust_solr_files
sub Func_cust_solr_files
{
	my ($dieOnError, $ignoreCheck) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_SOLR_FILE", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{
		dieIfCqFtsIsLocked();
		dieIfInState("cust_solr_files");
	}
	
	my $status;
	
	$status = customizeSolrFile($ftsSolrFile, $ftsSolrIntermediateFile, $dieOnError);
	
	# Also copy the schema to the typeahead Solr core.  We need to do this be it typeahead 
	# enabled or not, otherwise Solr will not start because the schema file won't be valid
	if ($status == 1)
	{
		$status = customizeSolrFile($ftsSolrTypeaheadFile, $ftsSolrIntermediateFile, $dieOnError);
	}
	
	if ($ignoreCheck eq 0)
	{
		setDoneState("cust_solr_files");
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_SOLR_FILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This functions sends the 'start' command to the WAS profile
# --start_fts_was_profile
sub Func_start_fts_was_profile
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("start_fts_was_profile");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = startFtsWASProfile($ftsWASProfileName, $dieOnError);
	
	setDoneState("start_fts_was_profile");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_WAS_PROFILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This functions sends the 'stop' command to the WAS profile
# --stop_fts_was_profile
sub Func_stop_fts_was_profile
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("stop_fts_was_profile");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = stopFtsWASProfile($ftsWASProfileName, $dieOnError);
	
	setDoneState("stop_fts_was_profile");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_WAS_PROFILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This function sets Solr's home environment via JVM's parameter
# --set_solr_home
sub Func_set_solr_home
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_SOLR_HOME", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("set_solr_home");
	
	setSolrHome($ftsWASProfileName, $ftsDataDir, 1);
	
	setDoneState("set_solr_home");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_SOLR_HOME", $stat_DONE)."\n", 1;
}


##############################################################################
# This functions enables CQ FTS by invoking the installutil command
# --enable_cqweb_fts
sub Func_enable_cqweb_fts
{
	my ($dieOnError, $ignoreCheck) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ENABLE_CQWEB_FTS", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{
		dieIfCqFtsIsLocked();
		dieIfInState("enable_cqweb_fts");
	}
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	enableDisableCQWebFTS($ENABLE_CQWEB_FTS, $dieOnError);
	
	if ($ignoreCheck eq 0)
	{
		setDoneState("enable_cqweb_fts");
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ENABLE_CQWEB_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This functions disables CQ FTS by invoking the installutil command
# --disable_cqweb_fts
sub Func_disable_cqweb_fts
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DISABLE_CQWEB_FTS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("disable_cqweb_fts");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	enableDisableCQWebFTS($DISABLE_CQWEB_FTS, $dieOnError);
	
	setDoneState("disable_cqweb_fts");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DISABLE_CQWEB_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function executes the batch mode record extractor
# --run_batch_idx
sub Func_run_batch_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_BATCH_INDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("run_batch_idx");
	
	runBatchIndex("both");
	
	setDoneState("run_batch_idx");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_BATCH_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function enables the update mode record extractor
# --start_update_idx
sub Func_start_update_idx
{
	my ($ignoreCheck) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_UPDATE_INDEX", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{	
		dieIfCqFtsIsLocked();
		dieIfInState("start_update_idx");
	}
	
	if ($ftsStartupdateIndex eq "true")
	{
		my $webXMLFile = findFtsWebXMLFolder($ftsWASProfileName);
		
		$webXMLFile .= "web.xml";
		
		$webXMLFile =~ s/\\/\//sg;
		
		my $line;
		my $data;
		if (!open(FH, $webXMLFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $webXMLFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			$data .= $line;
		}
		close FH;
		
		my $newValue;
		
		$newValue = HTML::Entities::encode($ftsCQPropertyFile);
		$data =~ s/<param-value>\S*\s*\D*\d*CQ-\S*\s*\D*\d*.xml<\/param-value>/<param-value>$newValue<\/param-value>/s;
		$data =~ s{
			(<param-name>config</param-name>)	# $1 = "sitename" element
			(.*?)								# $2 = stuff between elements
			(<param-value>)						# $3 = param-value start tag
			.*?									# discard param-value content
			(</param-value>)					# $4 = param-value end tag
		}{$1$2$3$newValue$4}sgx;
		
		$newValue = HTML::Entities::encode($ftsSiteName);
		$data =~ s{
			(<param-name>sitename</param-name>) # $1 = "sitename" element
			(.*?)								# $2 = stuff between elements
			(<param-value>)						# $3 = param-value start tag
			.*?									# discard param-value content
			(</param-value>)					# $4 = param-value end tag
		}{$1$2$3$newValue$4}sgx;
		
		$newValue = "<load-on-startup>1</load-on-startup>";
		$data =~ s/<!--\s*<load-on-startup>1<\/load-on-startup>\s*-->/$newValue/sm;
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("MODIFYING_WEBXML_FILE", $webXMLFile)."\n";
		
		safeFileUpdate($webXMLFile, $data, 1);	# now update web.xml file
		
		stopFtsWASProfile($ftsWASProfileName, 1);
		startFtsWASProfile($ftsWASProfileName, 1);
	}
	else
	{
		dealWithError($langMsgText->maketext("NOT_ENABLED_UPDATE_INDEX")."\n", 0);
	}
	
	if ($ignoreCheck eq 0)
	{
		setDoneState("start_update_idx");
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_UPDATE_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function stops the update mode record extractor.
# --stop_update_idx
sub Func_stop_update_idx
{
	my ($ignoreCheck) = @_;

	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_UPDATE_INDEX", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{	
		dieIfCqFtsIsLocked();
		dieIfInState("stop_update_idx");
	}
	
	my $webXMLFile = findFtsWebXMLFolder($ftsWASProfileName);
	
	$webXMLFile .= "web.xml";
	
	$webXMLFile =~ s/\\/\//sg;
	
	my $line;
	my $data;
	if (!open(FH, $webXMLFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $webXMLFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	while ($line = <FH>)
	{
		$data .= $line;
	}
	close FH;
	
	# If it's allready disabled, there is nothing to do
	if (($data =~ /<!--\s*<load-on-startup>1<\/load-on-startup>\s*-->/) eq "")
	{
		my $newValue;
		
		$newValue = "\n    <!--\n    <load-on-startup>1</load-on-startup>\n    -->"."\n";
		$data =~ s/\s*<load-on-startup>1<\/load-on-startup>\s*/$newValue/sm;
		
		safeFileUpdate($webXMLFile, $data, 1);	# now update web.xml file
		
		stopFtsWASProfile($ftsWASProfileName, 1);
		startFtsWASProfile($ftsWASProfileName, 1);
	}
	
	if ($ignoreCheck eq 0)
	{
		setDoneState("stop_update_idx");
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_UPDATE_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function reports if TypeAhead is broken or not
# --typeahead_cmd
sub Func_typeahead_cmd
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_TYPEAHEAD_CMD", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("");
	
	if ($ftsEnableTypeahead ne "true")
	{
		dealWithError($langMsgText->maketext("TYPEAHEAD_NOT_ENABLED")."\n", 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("TYPEAHEAD_CMD", $gTypeaheadCmd)."\n";
	
	my $typeaheadCmd = lc($gTypeaheadCmd);
	
	if ($typeaheadCmd eq "sync_idx")
	{
		syncTypeaheadIndex();
	}
	elsif ($typeaheadCmd eq "optimize_idx")
	{
		optimizeIndex("typeahead");
	}
	elsif ($typeaheadCmd eq "delete_idx")
	{
		deleteLuceneIndex("typeahead");
	}
	elsif ($typeaheadCmd eq "validate")
	{
		validateTypeahead();
	}
	else
	{
		my @cqtacfgActions = (["help", 1, 0], ["import", 0, 1], ["export", 0, 1], ["validate", 0, 1], ["remove", 0, 1]);
		my $ranIt = 0;
		
		foreach (@cqtacfgActions)
		{
			my @item = @$_;
			my $cmd = $item[0];
			if ($typeaheadCmd =~ /^$cmd?/)
			{
				executeTypeaheadScript_cqtacfg($gTypeaheadCmd, $item[0], $item[1], $item[2]);
				$ranIt = 1;
				last;
			}
		}
		
		if ($ranIt == 0)
		{
			dealWithError($langMsgText->maketext("INVALID_TYPEAHEAD_CMD", $gTypeaheadCmd)."\n", 1);
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_TYPEAHEAD_CMD", $stat_DONE)."\n", 1;
}


##############################################################################
# If a typeahead record-type is no longer enabled, it gets removed from the index
sub updateTypeaheadIndex
{
	my ($newTypeAheadData, $oldTypeAheadData) = @_;
	
	my $status = 1;
	
	my %old = ();
	my %new = ();
	
	# extract reocrd-type name and store them into a hash
	foreach my $item (@$oldTypeAheadData)
	{
		my @line = split("=", $item);
		$old{$line[0]} = "";
	}
	foreach my $item (@$newTypeAheadData)
	{
		my @line = split("=", $item);
		$new{$line[0]} = "";
	}
	
	# find out if a record-type no longer exists in the new
	# data if so, then remove remove it from the index
	for my $key (keys %old)
	{
		if (!(exists $new{$key}))
		{
			my @recordTypeToRemove;
			push(@recordTypeToRemove, $key);
			$status = removeRecordTypeFromIndex(\@recordTypeToRemove, "typeahead", 0);
		}
		
		if ($status == 0)
		{
			last;
		}
	}
	
	return $status;
}


##############################################################################
# Execute cqtacfg.pl script
sub internalExecuteTypeaheadScript_cqtacfg
{
	my ($cmdArg, $ignoreMatch, $dieOnError) = @_;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $cqtacfgCmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtacfgPL.$Q." ".$cmdArg." -dbset ".$ftsDBSetName." -userdb ".$ftsUserDbName." -username ".$ftsUserName." -password ".$ftsUserPassword;	
	
	my @mustMatch = ();
	if ($ignoreMatch == 0)
	{
		@mustMatch = ("ERROR: ", "Please see the IBM Rational ClearQuest Information Center for usage details of ClearQuest typeahead application.");
	}
	
	my $status = executeExternalCommand($cqtacfgCmd, \@mustMatch, $gDieKind_must_not_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Execute cqtacfg.pl script and also updates the typeahead index if the command
# being executed is 'import'
sub executeTypeaheadScript_cqtacfg
{
	my ($cqtacfgCmd, $cqtacfgCmdName, $ignoreMatch, $dieOnError) = @_;
	
	my $status = 1;
	
	my @oldTypeAheadData;
	my @newTypeAheadData;
	
	# if this is an 'import' operation, collect the current typeahead data to see if we have to also update the index
	if ($cqtacfgCmdName eq "import")
	{
		my $typeAheadFileName = $ftsDataDir."exported-typeahead-properties.txt";
		my $exportCmd = "export -file ".$typeAheadFileName;
		
		# get the data that is in the master-property right now: this is the old data
		$status = internalExecuteTypeaheadScript_cqtacfg($exportCmd, $ignoreMatch, $dieOnError);
		
		if ($status == 1)
		{
			if ((open my $FH, "<", $typeAheadFileName))
			{
				@oldTypeAheadData = <$FH>;
				close($FH);
			}
			else
			{
				dealWithError($langMsgText->maketext("CANNOT_OPEN", $typeAheadFileName)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				$status = 0;
			}
			
			# the temporary file is no longer needed, delete it
			if (unlink($typeAheadFileName) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE", $typeAheadFileName)."\n", 0);
			}
		}
		
		if ($status == 1)
		{
			# get the data user want us to set the master-property to: this is the new data
			$cqtacfgCmd =~ s{(.*)(-file\s)(.*)}{$1$2$3}sx;
			$typeAheadFileName = $3;
			if ((open my $FH, "<", $typeAheadFileName))
			{
				@newTypeAheadData = <$FH>;
				close($FH);
			}
			else
			{
				dealWithError($langMsgText->maketext("CANNOT_OPEN", $typeAheadFileName)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				$status = 0;
			}
		}
	}
	
	# now run the command if all is well
	if ($status == 1)
	{
		$status = internalExecuteTypeaheadScript_cqtacfg($cqtacfgCmd, $ignoreMatch, $dieOnError);
	}
	
	# if this was an 'import' operation, then check to see if we have to also update the 
	# typeahead index by removing any removed record-types
	if (($status == 1) && ($cqtacfgCmdName eq "import"))
	{
		updateTypeaheadIndex(\@newTypeAheadData, \@oldTypeAheadData);
	}
	
	return $status;
}


##############################################################################
# Sync the 'typeahead' index with the 'normal' index
sub syncTypeaheadIndex
{
	my ($function, $state) = getState();
	my $newState = 0;
	
	if (($function ne "") and ($function ne "typeahead_cmd_sync_idx"))
	{
		# we are in a state, and it's not our state
		dealWithError($langMsgText->maketext("IN_A_STATE_ALREADY", "typeahead_cmd_sync_idx", $function)."\n", 0);
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
	elsif ($function eq "")
	{
		# we are not in any state, set it ours
		$newState = 1;
	}
	
	$gTopLevelSate = "typeahead_cmd_sync_idx";
	
	if ($newState eq 1)
	{
		my $theFile = $ftsDataDir."batch_recovery.properties";
		
		if (-e $theFile)
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETE_BATCH_MARKER_FILE", $theFile)."\n";
			if (unlink($theFile) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 1);
			}
		}
		
		setState("typeahead_cmd_sync_idx", "");
	}
	
	if (index($state, "sync_fresh_batch_idx_continue") eq -1)
	{
		runBatchIndex("typeahead");
		
		setDoneState("");
	}
}


##############################################################################
# Optimizes the index
sub optimizeIndex
{
	my ($whichIndex) = @_;
	
	if ($ftsBatchIndexJVMParm ne "")
	{
		$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
	}
	
	if (!(-e $ftsCQPropertyFile))
	{
		dealWithError($langMsgText->maketext("MISSING_CONFIG_FILE", $ftsCQPropertyFile)."\n", 1);
	}
	
	if ($whichIndex ne "")
	{
		$whichIndex = " -sendto ".$whichIndex;
	}
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -optimize ".$Q.$ftsCQPropertyFile.$Q." -site \"".$ftsSiteName."\"".$whichIndex;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	my @mustMatch = $langMsgText->maketext("OPTIMIZE_IDX");
	executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, 1);
}


##############################################################################
# Deletes a Lucene index
sub deleteLuceneIndex
{
	my ($whichIndex) = @_;
	
	my $status = 1;
	
	if ($whichIndex ne "")
	{
		$whichIndex .= "/";
	}
	
	if ($ftsDeleteIndex ne "true")
	{
		dealWithError("\n".$langMsgText->maketext("NOT_ENABLED_DELETE_INDEX")."\n", 1);
	}
	
	# An example of what we need to send: http://localhost:14080/solr/update/<delete><query>*:*</query></delete>
	# Here "????" is the record-type name
	my $ua = LWP::UserAgent->new;
	
	my $httpPost = "http://".$ftsFtsServerName.":".$ftsFtsWASProfilePort."/solr/".$whichIndex."update";
	
	# Add POST data to HTTP request body
	my $solrDeleteQuery = "<delete><query>*:*</query></delete>";
	my $postData = { "stream.body" => $solrDeleteQuery, "commit" => "true" };
	my $postLogData = $httpPost."?"."stream.body=".$solrDeleteQuery."&commit=true";
	
	printAndLog $msgLabel_INFO."system -> [".$postLogData."]"."\n";
	
	my $resp = $ua->post($httpPost, $postData);
	if ($resp->is_error())
	{
		printAndLog $resp->status_line."\n", 1;
		dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$postLogData."]"."\n", 0);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Customizes the Solr schema.xml file
sub customizeSolrFile
{
	my ($solrFile, $solrIntermediateFile, $dieOnError) = @_;
	
	my $line;
	my $data;
	
	my $status = 1;
	
	# Read the current schema.xml file
	if (open(FH, $solrFile))
	{
		while ($line = <FH>)
		{
			$data .= $line;
		}
		close FH;
		
		# Reflect the user-db name in schema.xml
		$data =~ s/<schema name=".+" version=>/<schema name="$ftsUserDbName" version=>/;
		
		# Replace old <field>...</field> with new data based on user's DB
		my $solrFieldData = getSolrFieldData($solrIntermediateFile);
		$data =~ s/<fields>.+<\/fields>/<fields>\n$solrFieldData<\/fields>/s;
		
		# Replace old <copyField>...</copyField> with new data based on user's DB
		my $solrCopyFieldData = getSolrCopyFieldData($solrIntermediateFile);
		
		if ($data =~ /^\s+(<copyField source=".+" dest=".+"><\/copyField>)\s+$/smg)
		{
			# pre 7.1.0.2
			# <copyField source="Notes_Log" dest="RATL_CQ_ALL_FIELDS"></copyField>
			$data =~ s/^\s+(<copyField source=".+" dest=".+"><\/copyField>)\s+$/$solrCopyFieldData/smg;
		}
		else
		{
			# post 7.1.0.2
			# <copyField dest="RATL_CQ_ALL_FIELDS" source="Action_Types"/>
			$data =~ s/^\s+(<copyField dest=".+" source=".+"\/>)\s+$/$solrCopyFieldData/smg;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("MODIFYING_SOLR_SCHEMA_FILE", $solrFile)."\n";
		
		$status = safeFileUpdate($solrFile, $data, $dieOnError);
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $solrFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Executes the batch mode record extractor
sub runBatchIndex
{
	my ($whichIndex) = @_;
	
	if ($ftsDoBatchIndex eq "true")
	{
		# if we are in setup_cq_fts mode, update the CQ property file before we index so we pickup any sizes to batch size
		if ($optSetupCqFTS eq "1")
		{
			updateFTSPropFileIdxSetting($ftsCQPropertyFile, 1);
		}
		
		if ($ftsBatchIndexJVMParm ne "")
		{
			$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
		}
		
		if ($whichIndex ne "")
		{
			$whichIndex = " -sendto ".$whichIndex;
		}
		
		my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -batch ".$Q.$ftsCQPropertyFile.$Q." -site "."\"".$ftsSiteName."\"".$whichIndex;
		
		chdirAndValidate($ftsClearQuestHomeDir, 1);
		my @mustMatch = (": cqtsdbcrawer finished at ", ": cqtsdbcrawer  ", ": cqtsdbcrawer w_ ", ": cqtsdbcrawer  ");
		executeExternalCommand($cmd, \@mustMatch, $gDieKind_must_match, 1);
	}
	else
	{
		dealWithError($langMsgText->maketext("NOT_ENABLED_BATCH_INDEX")."\n", 0);
	}
}


##############################################################################
# Updates the cqtsadmin-<dbset>-<userdb>.xml file to insert missing XML blocks
sub updateFile_cqtsadmin_xml
{
	my ($ftsHome, $dbSetName, $userDbName, $dieOnError) = @_;
	
	my $ftsCfg = $ftsHome."cqtsadmin"."-".$dbSetName."-".$userDbName.".xml";
	
	my $status = 0;
	
	# Read the existing cqtsadmin-<dbset>-<userdb>.xml file
	if (!open(FH, $ftsCfg))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsCfg)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return $status;
	}
	
	# Read the file into memory
	my $toData;
	my $line;
	while ($line = <FH>)
	{
		$toData .= $line;
	}
	close FH;
	
	# update cqtsadmin-<dbset>-<userdb>.xml
	my $deleteIndexBlock;
	my $enableTypeaheadBlock;
	my $typeAheadCacheFillTriesBlock;
	my $typeAheadCacheSizeBlock;
	my $typeAheadPageSizeBlock;
	
	# read the language specific .xml file to extract the "<solrVersion>" xml block from it
	my $configFileName = $ftsClearQuestHomeDir."CQTsAdmin/L10N/".$langMsgText->maketext("LANGUAGE_CODE").".xml";
	if (open(FH, $configFileName))
	{
		my $line;
		my $fromData;
		while ($line = <FH>)
		{
			$fromData .= $line;
		}
		close FH;
		
		$fromData =~ s{(<deleteIndex>.*?</deleteIndex>)}{$1}sx;
		$deleteIndexBlock = "\n  \n  ".$1."\n  \n  ";
		
		$fromData =~ s{(<enableTypeahead>.*?</enableTypeahead>)}{$1}sx;
		$enableTypeaheadBlock = "\n  \n  ".$1."\n  \n  ";
		
		$fromData =~ s{(<typeAheadCacheFillTries>.*?</typeAheadCacheFillTries>)}{$1}sx;
		$typeAheadCacheFillTriesBlock = "\n  \n  ".$1."\n  \n  ";
		
		$fromData =~ s{(<typeAheadCacheSize>.*?</typeAheadCacheSize>)}{$1}sx;
		$typeAheadCacheSizeBlock = "\n  \n  ".$1."\n  \n  ";
		
		$fromData =~ s{(<typeAheadPageSize>.*?</typeAheadPageSize>)}{$1}sx;
		$typeAheadPageSizeBlock = "\n  \n  ".$1."\n  \n  ";
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $configFileName)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		return $status;
	}
	
	# insert the missing XML blocks (if any is missing)
	if (!($toData =~ m/<deleteIndex>.*?<\/deleteIndex>/s))
	{
		# insert after <deleteFtsWASProfiles/> XML block
		$toData =~ s{(.*?</deleteFtsWASProfiles>).*?(<.*?)}{$1$deleteIndexBlock$2}sgx;
	}
	
	# insert the missing XML blocks (if any is missing)
	if (!($toData =~ m/<enableTypeahead>.*?<\/enableTypeahead>/s))
	{
		# insert after <deleteIndex/> XML block
		$toData =~ s{(.*?</deleteIndex>).*?(<.*?)}{$1$enableTypeaheadBlock$2}sgx;
	}
	
	if (!($toData =~ m/<typeAheadCacheFillTries>.*?<\/typeAheadCacheFillTries>/s))
	{
		# insert after <enableTypeahead/> XML block
		$toData =~ s{(.*?</enableTypeahead>).*?(<.*?)}{$1$typeAheadCacheFillTriesBlock$2}sgx;
	}
	
	if (!($toData =~ m/<typeAheadCacheSize>.*?<\/typeAheadCacheSize>/s))
	{
		# insert after <typeAheadCacheFillTries/> XML block
		$toData =~ s{(.*?</typeAheadCacheFillTries>).*?(<.*?)}{$1$typeAheadCacheSizeBlock$2}sgx;
	}
	
	if (!($toData =~ m/<typeAheadPageSize>.*?<\/typeAheadPageSize>/s))
	{
		# insert after <typeAheadCacheSize/> XML block
		$toData =~ s{(.*?</typeAheadCacheSize>).*?(<.*?)}{$1$typeAheadPageSizeBlock$2}sgx;
	}
	
	$status = safeFileUpdate($ftsCfg, $toData, $dieOnError);
	
	return $status;
}


##############################################################################
# Updates Solr's solr.xml file to include 'typeahead' Solr core.  For now,
# this function copies over the existing solr.xml file, if we upgrade Solr
# from 3.6.1 to a later version, we will need to look into this function 
# to find out if a copy over is fine or not.
sub updateFile_solr_xml
{
	my ($ftsHome, $dbSetName, $userDbName, $dieOnError) = @_;
	
	my $status = 1;
	
	# if we are on Solr version 1.2.0, then we cannot copy solr.xml as it doesn't exist
	if ($ftsSolrVersion eq $gSolrVersion_120)
	{
		return $status;
	}
	
	my $fromFile = $ftsSourceDir."Solr/solr/solr.xml";
	my $toFile = $ftsHome."Solr/solr/solr.xml";
	
	if (!copy($fromFile, $toFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $fromFile, $toFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# If the directory <ftshome>\Solr\typeahead\ is missing then it is copied (including all
# the sub directories and files)
sub copyDirectory_Solr_typeahead
{
	my ($ftsHome, $dbSetName, $userDbName, $dieOnError) = @_;
	
	my $status = 1;
	
	# if we are on Solr version 1.2.0, then we cannot copy 'typeahead' directory as it doesn't exist
	if ($ftsSolrVersion eq $gSolrVersion_120)
	{
		return $status;
	}
	
	if (-d $ftsHome."Solr/typeahead/")
	{
		return $status;
	}
	
	my $tmp = $ftsTargetDir;	# remember it so we can restore it
	
	eval
	{
		$ftsTargetDir = $ftsHome;
		
		followDir($ftsSourceDir."Solr/typeahead/", \&copyDirOverwriteMkdir);
	};
	if ($@)
	{
		my $err = $@;
		dealWithError("\n".$err."\n", $dieOnError);
		$status = 0;
	}
	
	$ftsTargetDir = $tmp;	# restore it
	
	return $status;
}


##############################################################################
# Validates typeahead settings
sub validateTypeahead
{
	my $adminSessionObj = 0;
	my $TypeAheadProName = "TYPEAHEAD_PARAMS";
	my $TypeAheadProValue = "";
	my $pwd = $ftsUserPassword;
	my @TypeAheadProdata;
	my $isValid = 1;
		
	if ($pwd eq "\"\"")
	{
		$pwd = "";
	}
	eval
	{
		$adminSessionObj = CQAdminSession::Build();
		$adminSessionObj->Logon($ftsUserName, $pwd, $ftsDBSetName);
		$TypeAheadProValue = $adminSessionObj->GetMasterProperty($TypeAheadProName, $ftsUserDbName, "");
	};
	if ($@)
	{
		my $err = $@;
		dealWithError("\n".$err."\n", 1);
	}
	@TypeAheadProdata = split(/\r?\n/, $TypeAheadProValue);
	
	# Remove the blank items in the array
	@TypeAheadProdata = grep(/[^\s*]/, @TypeAheadProdata);
	
	foreach my $item (@TypeAheadProdata)
	{
		$item =~ s/\s*//g;	
		my @line = split("=", $item);
		my $recordType = $line[0];
		my @fields = split(",",$line[1]);
		
		if (isThisRecordTypeIndexed($recordType, 0))
		{
			foreach my $field (@fields)
			{
				if (isThisFieldIndexed($field, $recordType, 0) == 0)
				{
					printAndLog $msgLabel_WARN.$langMsgText->maketext("FIELD_NOT_INDEXED", $field, $recordType)."\n";
					$isValid = 0;
				}
			}
		}
		else
		{
			printAndLog $msgLabel_WARN.$langMsgText->maketext("RECORD_TYPE_NOT_INDEXED_2", $recordType)."\n";
			$isValid = 0;
		}
	}
	CQAdminSession::Unbuild($adminSessionObj);

	# Printout validation message
	my $msgLabel;
	my $msgKey;
	
	if ($isValid == 1)
	{
		$msgLabel = $msgLabel_INFO;
		$msgKey = "TYPEAHEAD_VALIDATION_SUCCESS";
	}
	else
	{
		$msgLabel = $msgLabel_WARN;
		$msgKey = "TYPEAHEAD_VALIDATION_FAILED";
	}
	
	printAndLog $msgLabel."************************************************************"."\n", 1;
	printAndLog $msgLabel.$langMsgText->maketext($msgKey)."\n", 1;
	printAndLog $msgLabel."************************************************************"."\n", 1;	
}


##############################################################################
# Deletes FTS WAS profile
sub deleteFtsWASProfile
{
	my ($profileName, $dieOnError) = @_;
	
	my $status = 0;
	my $profilePath = "";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_WAS_PROFILE", $profileName)."'\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("TIME_1_TO_5_MIN")."\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	
	($status, $profilePath) = getProfilePath($profileName, $dieOnError);
	
	stopFtsWASProfile($profileName, 0);		# stop it if it's running (otherwise the delete will fail)
	
	chdirAndValidate($ftsFtsWasAppSrvBinDir, $dieOnError);
	
	my @mustMatch = $langMsgText->maketext("DEL_WAS_PROFILE");
	my $cmd = $Q.$ftsFtsWasAppSrvBinDir."manageprofiles".$ext.$Q." -delete -profileName ".$profileName;
	$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_FOLDERS_FILES", $profilePath)."\n";
	File::Path::rmtree($profilePath);
	
	return $status;
}


##############################################################################
# Creates FTS WAS profile
sub createFtsWASProfile
{
	my ($ftsHome, $profileName, $profilePort, $serviceStartupType, $profileHomeDir, $solrVersion, $dieOnError) = @_;
	
	my $status = 1;
	
	# make sure there is no such profiel alrady
	if (profileExist($profileName) eq 1)
	{
		dealWithError($langMsgText->maketext("WAS_PROFILE_EXIST", $profileName)."\n", $dieOnError);
		$status = 0;
	}
	
	# make srue the port is good
	if ($status eq 1)
	{
		if (length($profilePort) < 5)
		{
			dealWithError($langMsgText->maketext("WAS_PORT_GT_5")."\n", $dieOnError);
			$status = 0;
		}
	}
	
	my $WASPropFile = $ftsDataDir."cqftsportdef.props";
	
	# prepare the WAS property file (to create the WAS profile)
	if ($status eq 1)
	{
		if ($optFtsWasProfPortsFile ne "")
		{
			if ($optFtsWasProfPortsFile ne $ftsWASProfilePortsDefPropFile)
			{
				printAndLog $msgLabel_WARN."************************************************************"."\n";
				printAndLog $msgLabel_WARN.$langMsgText->maketext("PORTS_FILE_OVERRIDE", $ftsWASProfilePortsDefPropFile)."\n", 1;
				printAndLog $msgLabel_WARN."************************************************************"."\n";
			}
			
			$WASPropFile = $optFtsWasProfPortsFile;
			
			printAndLog $msgLabel_INFO.$langMsgText->maketext("USER_PORT_PROFILE", $WASPropFile)."\n";
		}
		elsif ($ftsWASProfilePortsDefPropFile ne "")
		{
			$WASPropFile = $ftsWASProfilePortsDefPropFile;
			
			printAndLog $msgLabel_INFO.$langMsgText->maketext("USER_PORT_PROFILE", $WASPropFile)."\n";
		}
		else
		{
			my $wasProfileData = "WC_defaulthost=xxx080\nWC_adminhost=xxx060\nWC_defaulthost_secure=xxx443\nWC_adminhost_secure=xxx043\nBOOTSTRAP_ADDRESS=xxx809\nSOAP_CONNECTOR_ADDRESS=xxx880\nORB_LISTENER_ADDRESS=xxx00\nIPC_CONNECTOR_ADDRESS=xxx04"."\n";
			
			my $port = substr($profilePort, 0, length($profilePort) - 3);
			
			$wasProfileData =~ s/=xxx/=$port/smg;
			
			eval
			{
				printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_WAS_PROP_FILE", $WASPropFile)."\n";
				
				if (!open(FH, ">$WASPropFile"))
				{
					dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $WASPropFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
					$status = 0;
				}
				print FH $wasProfileData;
				close FH;
			};
			if ($@)
			{
				dealWithError($langMsgText->maketext("CANNOT_CREATE", $WASPropFile)."\n", $dieOnError);
				$status = 0;
			}
			
			printAndLog $msgLabel_INFO.$langMsgText->maketext("AUTO_GEN_PORT_PROFILE")."\n";
		}
	}
	
	# create the WAS profile
	if ($status eq 1)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_NEW_WAS_PROFILE", $profileName, $profilePort)."'\n";
		printAndLog $msgLabel_INFO."************************************************************"."\n";
		printAndLog $msgLabel_INFO.$langMsgText->maketext("TIME_5_TO_10_MIN")."\n";
		printAndLog $msgLabel_INFO."************************************************************"."\n";
		
		chdirAndValidate($ftsFtsWasAppSrvBinDir, 1);
		
		# Create a new WAS profile
		my $extraWindowsOpt = "";
		if ($gIsWindows eq 1)
		{
			$extraWindowsOpt = "-winserviceCheck true -winserviceStartupType ".$serviceStartupType;
		}
		else
		{
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_NOT_SUPPORTED_SRV_TYPE", $serviceStartupType)."\n", 1;
		} 
		
 		my $cmd = $Q.$ftsFtsWasAppSrvBinDir."manageprofiles".$ext.$Q." -create -templatePath ".$Q.$ftsFtsWasAppSrvDir."profileTemplates/default".$Q." -profileName ".$profileName." -profilePath ".$Q.$profileHomeDir.$profileName.$Q." ".$extraWindowsOpt." -hostName localhost -nodeName ".$profileName." -cellName dfltCell -portsFile ".$Q.$WASPropFile.$Q." -omitAction samplesInstallAndConfig defaultAppDeployAndConfig";
		
		my @mustMatch = $langMsgText->maketext("CRE_WAS_PROFILE");
		executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	}
	
	# delete WAS port file
	if (unlink($WASPropFile) ne 1)
	{
		# we run into a clean up issue, it's not nice but it's not a failur
		dealWithError($langMsgText->maketext("CANNOT_DELETE", $WASPropFile)."\n", 0);
	}

	
	# now deploy Solr
	if ($status eq 1)
	{
		my $solrWarFile = getSolrWARFile($solrVersion);
		
		$status = createDeplySolrScript($ftsDBSetName, $ftsUserDbName, $solrVersion, 0);
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -lang jython -profileName ".$profileName." -f ".$Q.$ftsDataDir."deploySolr.py".$Q." ".$Q.$solrWarFile.$Q." server1 solr ".$solrVersion;
			my @mustMatch = $langMsgText->maketext("DEPLOY_SOLR");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -c ".$Q."\$AdminConfig save".$Q;
			my @mustMatch = $langMsgText->maketext("DEPLOY_SOLR_SAVE");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
	}
	
	# depoly CqTsDbCrawler
	if ($status eq 1)
	{
		my $crawlerWarFile = $ftsClearQuestHomeDir."cqweb/apps/cqtsdbcrawler.war";
		
		if (!(-e $crawlerWarFile))
		{
			$crawlerWarFile = $ftsClearQuestHomeDir."cqtsdbcrawler.war";
		}
		
		if (-e $crawlerWarFile)
		{
			my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -c ".$Q."\$AdminApp install {".$crawlerWarFile."} { -contextroot cqtsdbcrawler -appname cqtsdbcrawler -MapWebModToVH {{cqtsdbcrawler cqtsdbcrawler.war,WEB-INF/web.xml default_host}}}".$Q;
			my @mustMatch = $langMsgText->maketext("DEPLOY_CRAWLER");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
		else
		{
			$status = 0;
		}
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -c ".$Q."\$AdminConfig save".$Q;
			my @mustMatch = $langMsgText->maketext("DEPLOY_CRAWLER_SAVE");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
	}
	
	# delete Solr deploy script
	deleteDeplySolrScript($ftsDBSetName, $ftsUserDbName, 0);
	
	# set Solr's home
	if ($status eq 1)
	{
		$status = setSolrHome($profileName, $ftsHome, $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Set the JVM maximum memory of the WAS profile
sub setWASMaxMem
{
	my ($profileName, $maxHeapSize, $dieOnError) = @_;
	
	my $cmd = $ftsClearQuestHomeDir."cqweb/scripts/";
	
	my $status = chdirAndValidate($cmd, $dieOnError);
	
	$cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -profile -profileName ".$profileName." ".$Q.$cmd."teamAdminUtils.jacl".$Q." -c ".$Q."setJvmMaxHeapSize ".$maxHeapSize.$Q;
	
	my @mustMatch = $langMsgText->maketext("SET_WAS_MAX_MEM");
	$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Prevent execution if the FTS deployment is locked
sub dieIfCqFtsIsLocked
{
	if (-f $ftsLockFile)
	{
		dealWithError("\n".$langMsgText->maketext("CQFTS_IS_LOCKED_ERROR")."\n", 1);
	}
}


##############################################################################
#  Prevent deployment if WAS is not 32 bit
sub dieIf64BitWAS
{
	my $WASjava = $ftsFtsWasAppSrvDir."java/bin/java";
	my $classpath = ($gIsWindows eq 1) ? "./cqtsdbcrawler.jar;./cqjni.jar" : "'./cqtsdbcrawler.jar:./cqjni.jar'";
	my $cmd = $Q.$WASjava.$Q." -classpath $classpath com.ibm.rational.clearquest.cqtscrawler.batch.ExtractorCommand -jvmarch";
	my @mustMatch = $langMsgText->maketext("MATCH_NOTHING");
	$gCaptureExeOutputData = "";

	executeExternalCommand($cmd, @mustMatch, $gDieKind_none, 0, 1);
	if ($gAllow64BitWAS ne "true")
	{
		if ($gCaptureExeOutputData =~ /sun.arch.data.model -> 64/)
		{
			dealWithError("\n".$langMsgText->maketext("INVALID_WAS_INSTALLED")."\n", 1);
		}
	}
}


##############################################################################
# Terminates if we are trying to execute a command-line option that will alter
# the current state
sub dieIfInState
{
	my ($cmd) = @_;
	
	if ($cmd eq "")
	{
		return;
	}
	
	my ($function, $state) = getState();
	
	if (($function ne "") and ($function ne $gTopLevelSate))
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("IN_A_STATE_ALREADY", $cmd, $function)."\n";
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
}


##############################################################################
# Runs the CQ installutil command (to enable / disable CQ FTS)
sub enableDisableCQWebFTS
{
	my ($action, $dieOnError) = @_;
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = 1;
	
	if ($ftsIsEnableCQFTS eq "true")
	{
		my $user = $ftsUserName;
		my $pwd = $ftsUserPassword;
		my $dbSet = $ftsDBSetName;
		my $dbName = $ftsUserDbName;
		
		if ($pwd eq "\"\"")
		{
			$pwd = "";
		}
		
		my $propertyName = "TEXT_SEARCH_PROPERTY";
		my $propertyValue = "";
		my @siteList = ();
		
		if ($action eq $ENABLE_CQWEB_FTS)
		{
			my @data;
			if (!open(FH, $ftsCQPropertyFile))
			{
				dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsCQPropertyFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				return 0;
			}
			@data = <FH>;
			close(FH);
			
			$propertyValue = join("", @data);	# bring it all into one buffer
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_PROP_DATA", "\n"."\n".$propertyName."=\"".$propertyValue."\""."\n")."\n";
		
		eval
		{
			my $adminSessionObj = CQAdminSession::Build();
			
			$adminSessionObj->Logon($user, $pwd, $dbSet);
			
			if ($ftsIsReplicated == 1)
			{
				# this is a replicated CQ DB
				if ($ftsEnabledAtReplicaNamesList ne "")
				{
					# we want the list from the XML config file first
					@siteList = split(",", $ftsEnabledAtReplicaNamesList);
				}
				elsif ($ftsReplicaNamesList ne "")
				{
					# no list in XML config file, so get the list from the DB
					@siteList = split(",", $ftsReplicaNamesList);
				}
			}
			else
			{
				# we are not replicated
				push(@siteList, "");
			}
			
			foreach my $siteName (@siteList)
			{
				if ($action eq $DISABLE_CQWEB_FTS)
				{
					printAndLog $msgLabel_INFO.$langMsgText->maketext("DISABLE_CQWEB_FTS", $siteName eq "" ? "<local>" : $siteName)."\n";
					$adminSessionObj->UnsetMasterProperty($propertyName, $dbName, $siteName);
				}
				else
				{
					printAndLog $msgLabel_INFO.$langMsgText->maketext("ENABLE_CQWEB_FTS", $siteName eq "" ? "<local>" : $siteName)."\n";
					$adminSessionObj->SetMasterProperty($propertyName, $propertyValue, $dbName, $siteName);
				}
			}
			
			CQAdminSession::Unbuild($adminSessionObj);
		};
		if ($@)
		{
			my $err = $@;
			dealWithError($langMsgText->maketext("ERROR_ENABLE_DISABLE_FTS")."\n".$err."\n", 1);
			$status = 0;
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("NOT_ENABLED_CQWEB_FTS")."\n", 0);
	}
	
	return $status;
}


##############################################################################
# Checks to see if the existing CQ search property file has the typeahead data
# if not, it will insert it
sub insertTypeaheadPropertyData
{
	my ($properityFileToUpdate, $entityFile, $dieOnError) = @_;
	
	# read the existing CQ search property file
	my @data;
	if (!open(FH, $properityFileToUpdate))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $properityFileToUpdate)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return 0;
	}
	@data = <FH>;
	close(FH);
	
	my $status = 1;		# assume all is well
	my $startSE = 0;
	my $endSE = 0;
	my $foundTAProp = 0;
	my $needsUpdate = 0;
	my $newData = "";
	my $line;
	
	foreach $line ( @data )
	{
		# match line containing "<searchrequest [anything] url="..."/>"
		if ($line =~ m/(.*<searchrequest .*\/>)/)
		{
			# see if it has: &amp;timeAllowed=%d, if not we need to insert it
			if (!($line =~ m/&amp;timeAllowed=%d/))
			{
				$line =~ s{(.*)(<searchrequest )(.*)(url)(=)(")(.*)(")(.*)(/>)}{$1$2$3$4$5$6$7$8$9$10}sgx;
				$line = $1.$2.$3.$4.$5.$6.$7."&amp;timeAllowed=%d".$8.$9.$10."\n";
				$needsUpdate = 1;
			}
		}

		# match line containing "<searchprop [anything]/>"
		if ($line =~ m/(.*<searchprop .*\/>)/)
		{
			# see if it has: timeallowed="0", if not we need to insert it
			if (!($line =~ m/timeallowed/))
			{
				$line =~ s{(.*)(<searchprop )(.*)(/>)}{$1$2$3$4}sgx;
				$line = $1.$2.$3." timeallowed=\"0\"".$4."\n";
				$needsUpdate = 1;
			}
		}
		
		# match line containing "<CQDbCrawler [anything] url="..."/>"
		if ($line =~ m/(.*<CQDbCrawler .*>)/)
		{
			# see if it has: &amp;timeAllowed=%d, if not we need to insert it
			if (!($line =~ m/typeahead=/))
			{
				# <CQDbCrawler sitename="&lt;local&gt;">
				$line =~ s{(.*)(<CQDbCrawler )(.*)(>)}{$1$2$3$4}sgx;
				$line = $1.$2.$3." typeahead=\"disable\"".$4."\n";
				$needsUpdate = 1;
			}
		}
		
		# match line containing "<typeaheadsearchprop [anything]/>"
		if ($line =~ m/(.*<typeaheadsearchprop .*\/>)/)
		{
			$foundTAProp = 1;
		}
		
		# starting block: match line containing "<SearchEngine " (yes, match the space at the end too)
		if ($line =~ m/(.*<SearchEngine .*>)/)
		{
			$startSE = 1;
		}
		
		# ending block: match line containing "</SearchEngine>"
		if ($line =~ m/(.*<\/SearchEngine>.*)/)
		{
			$endSE = 1;
			
			if (($startSE == 1) and ($foundTAProp == 0))
			{
				# we need to insert the typeahead XML data
				my $typeAheadData = "";
				
				($status, $typeAheadData) = extractTypeaheadPropertyData($entityFile, 0);
				
				if (($status == 1) and (length($typeAheadData) != 0))
				{
					# insert the typeahead data, pre-fix with 4 spaces for tabbing alignment
					$newData .= "    ".$typeAheadData."\n";
					$needsUpdate = 1;
				}
				else
				{
					# something went wrong, we could not find our typeahead data; exit the loop
					$status = 0;	# flip to failed state
					last;
				}
			}
			elsif ($startSE == 0)
			{
				# this is an edge case, it means the XML file is corrupt; it also means FTS isn't working, exit the loop
				dealWithError($langMsgText->maketext("MALFORMED_FILE", $properityFileToUpdate)."\n".">> [".$line."]"."\n", 0);
				$status = 0;	# flip to failed state
				last;
			}
		}
		
		$newData .= $line;
	}
	
	if (($needsUpdate == 1) and ($status == 1))
	{
		# the typeahead data was inserted, we need to save the updated CQ search property data
		$status = safeFileUpdate($properityFileToUpdate, $newData, 0);
		
		# upload the new CQ search property data to the DB
		if ($status == 1)
		{
			$status = enableDisableCQWebFTS($ENABLE_CQWEB_FTS, 0);
		}
	}
	
	return $status;
}


##############################################################################
# Extracts typeahead property data
sub extractTypeaheadPropertyData
{
	my ($entityFile, $dieOnError) = @_;
	
	my $status = 0;		# assume we will fail
	my $tempPropertyFile = $ftsCQPropertyFile.".inserrt-typeahead-data.temp";
	my $typeAheadData = "";
	
	# generate a temporary CQ search property file which will have the typeahead data
	$status = generateCQSearchPropertyFile($entityFile, $tempPropertyFile, 0);
	if ($status eq 0)
	{
		# the temporary file was not generated
		return $status, $typeAheadData;
	}
	
	# read the temporarily generated CQ search property file to extract from it the typeahead data
	my @data;
	if (!open(FH, $tempPropertyFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $tempPropertyFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return $status, $typeAheadData;
	}
	@data = <FH>;
	close(FH);
	
	$typeAheadData = join("", @data);	# bring it all into one buffer
	
	# extract the typeahead data by doing a double extraction
	# first the <SearchEngine/> block
	$typeAheadData =~ s{
						(<SearchEngine)
						(.*?)
						(</SearchEngine>)
					}{$1$2$3}sgx;
	
	$typeAheadData = $2;
	
	# now the <typeaheadsearchprop/> block
	$typeAheadData =~ s{
						(<typeaheadsearchprop)
						(.*?)
						(/>)
					}{$1$2$3}sgx;
	
	if (($1 eq "<typeaheadsearchprop") and ($3 eq "/>"))
	{
		# we have our match, build the typeahead data which we will return
		$typeAheadData = $1.$2.$3;
		$status = 1;
	}
	else
	{
		# something went wrong, we don't have our data
		$typeAheadData = "";
	}
	
	if (unlink($tempPropertyFile) ne 1)
	{
		# we run into a clean up issue, it's not nice but it's not a failur
		dealWithError($langMsgText->maketext("CANNOT_DELETE", $tempPropertyFile)."\n", 0);
	}
	
	return $status, $typeAheadData;
}


##############################################################################
# Customizes FTS files (for now, CQ Search properties file only)
sub updateFTSPropFileIdxSetting
{
	my ($cqProperityFile, $dieOnError) = @_;
	
	my @lines;
	
	if (!open(FH, $cqProperityFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $cqProperityFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	@lines = <FH>;
	close(FH);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("UPDATE_CQFTS_PROP_FILE", $cqProperityFile)."\n";
	
	my $data;
	
	foreach my $line ( @lines )
	{
		# change the size attribute of batch element lines if necessary
		if ($line =~ /^\s*<batch\s/)
		{
			$line =~ s/\bsize="\d+"/size="$ftsBatchIndexBatchSize"/;
			$line =~ s/\bdelay="\d+"/delay="$ftsBatchIndexDelay"/;
			$line =~ s/\bthread="\d+"/thread="$ftsBatchIndexThreads"/;
		}
		
		# change the size attribute of update element lines if necessary
		if ($line =~ /^\s*<update\s/)
		{
			$line =~ s/\bsize="\d+"/size="$ftsUpdateIndexBatchSize"/;
			$line =~ s/\bdelay="\d+"/delay="$ftsUpdateIndexDelay"/;
			$line =~ s/\blogin_iteration="\d+"/login_iteration="$ftsUpdateIndexLoginInterval"/;
		}
		
		$data .= $line;
	}
	
	safeFileUpdate($cqProperityFile, $data, 1);
}


##############################################################################
# Customizes FTS files (for now, CQ Search properties file only)
sub customizeFtsFiles
{
	my ($isInUpdateMode, $cqProperityFile, $cqEntityFile, $recreatePW) = @_;
	
	my $data;
	my $line;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("UPDATE_CQFTS_PROP_FILE", $cqProperityFile)."\n";
	
	# If a password is given, then generate an encrypted version of it, and a key file for it
	if (($ftsUserPassword ne "\"\"") and ($recreatePW eq 1))
	{
		my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q." -encrypt ";
		my $parm = " -u ".$ftsUserName." -p ".$ftsUserPassword." -s ".$ftsDBSetName." -d ".$ftsUserDbName." -kf ".$Q.$ftsKeyFile.$Q." -pf ".$Q.$ftsPWDFile.$Q;
		
		chdirAndValidate($ftsClearQuestHomeDir, 1);
		
		$cmd .= $parm;
		
		my @mustMatch = $langMsgText->maketext("MATCH_NOTHING");
		executeExternalCommand($cmd, @mustMatch, $gDieKind_none, 0);
		validateFileGenerated($cmd, $ftsKeyFile, 1);
		validateFileGenerated($cmd, $ftsPWDFile, 1);
		
		# Open the CQ Search property file and add to it the key and pwd info
		
		my $pwd;
		
		if (!open(FH, $ftsPWDFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsPWDFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			$pwd .= $line;
		}
		close FH;
		
		if (!open(FH, $cqProperityFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $cqProperityFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			$data .= $line;
		}
		close FH;
		
		$data =~ s/keyfile=".*?"/keyfile="$ftsKeyFile"/sg;
		$data =~ s/password=".*?"/password="$pwd"/sg;
		
		safeFileUpdate($cqProperityFile, $data, 1);
	}
	
	$data = "";
	
	#
	# Open the CQ Search property file and change the crawler setting based on new input (if any) from the user
	#
	
	my @lines;
	{
		if (!open(FH, $cqProperityFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $cqProperityFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		@lines = <FH>;
		close(FH);
	}
	
	my ($nu1, $nu2) = getEntityFileWithoutDispFieldInfo($cqEntityFile, 0, 1);	# all that we want is the list of display fields
	
	my $currRecordType = "";
	
	foreach $line ( @lines )
	{
		# change the type ahead values of element lines if necessary
		if ($line =~ /^\s*<typeaheadsearchprop\s/)
		{
			$line =~ s/\bcachefillmaxtries="\d+"/cachefillmaxtries="$ftsTypeAheadCacheFillTries"/;
			$line =~ s/\bcachesize="\d+"/cachesize="$ftsTypeAheadCacheSize"/;
			$line =~ s/\bpagesize="\d+"/pagesize="$ftsTypeAheadPageSize"/;
		}
		
		# change the size attribute of batch element lines if necessary
		if ($line =~ /^\s*<batch\s/)
		{
			$line =~ s/\bsize="\d+"/size="$ftsBatchIndexBatchSize"/;
			$line =~ s/\bdelay="\d+"/delay="$ftsBatchIndexDelay"/;
			$line =~ s/\bthread="\d+"/thread="$ftsBatchIndexThreads"/;
		}
		
		# change the size attribute of update element lines if necessary
		if ($line =~ /^\s*<update\s/)
		{
			$line =~ s/\bsize="\d+"/size="$ftsUpdateIndexBatchSize"/;
			$line =~ s/\bdelay="\d+"/delay="$ftsUpdateIndexDelay"/;
			$line =~ s/\blogin_iteration="\d+"/login_iteration="$ftsUpdateIndexLoginInterval"/;
		}

		# change the typeahead value to enable / disable typeahead
		if ($line =~ /^\s*<CQDbCrawler\s/)
		{
			my $val = "disable";
			if ($ftsEnableTypeahead eq "true")
			{
				$val = "enable";
			}
			$line =~ s/\btypeahead="\w+"/typeahead="$val"/;
		}
		
		# change the port and FTS server name
		if (($line =~ /^\s*<searchrequest\s/) or ($line =~ /^\s*<batch\s/) or ($line =~ /^\s*<update\s/))
		{
			$line =~ s/:\d*080\//:$ftsFtsWASProfilePort\//;
			$line =~ s/\/\/.*?:/\/\/$ftsFtsServerName:/sg;
		}
		
		# Set the "dbid" field as the default display field
		if (($isInUpdateMode eq 0) and (($line =~ /^\s*<field\s/) and ($line =~ /name="dbid"/)))
		{
			$line =~ s/disp_order="0"/disp_order="9999"/;
		}
		
		# set the display-field to whatever the user want it to be (set via the Entity file)
		unless (not %gDispOrdFieldList)
		{
			if ($line =~ m/(.*<CQEntity.*name=.*)/)
			{
				while (my ($recordType, $field) = each (%gDispOrdFieldList))
				{
					if ($line =~ m/(.*<CQEntity.*name="$recordType".*)/)
					{
						$currRecordType = $recordType;
					}
				}
			}
			if ($line =~ m/(.*<\/CQEntity>)/)
			{
				$currRecordType = "";
			}
			if ($currRecordType ne "")
			{
				while (my ($recordType, $field) = each (%gDispOrdFieldList))
				{
					if ($currRecordType eq $recordType)
					{
						if ($line =~ /^\s*<field\s/)
						{
							# we matched a field line
							if ($line =~ /name="$field"/)
							{
								# we matched out field name, change it to be the display field
								$line =~ s/disp_order="0"/disp_order="1"/;
							}
							else
							{
								# it's not our display field, reset it to '0' in case if it was set to '1'
								$line =~ s/disp_order="1"/disp_order="0"/;
							}
						}
					}
				}
			}
		}
		
		$data .= $line;
	}
	
	safeFileUpdate($cqProperityFile, $data, 1);
}


##############################################################################
# This function tries to create a new deployment of CQ FTS; it checks to see 
# if one already exists at the current location, if not then it will create
# the directory and push the 2 functions to be executed to complete the 
# operation.  The 2 functions are: Func_copy_fts_template() and 
# Func_gen_entity_file()
# --init_cq_fts
sub initCqFTS
{
	dieIfCqFtsIsLocked();
	dieIfInState("init_cq_fts");
	
	dieIf64BitWAS();
	
	if (!(-f $optCqFtsHomeDir) and !(-d $optCqFtsHomeDir))
	{
		# the location they want to initalize CQ FTS at is not a config file or an existing 
		# folder; this is an initial setup so create our inital folders and files
		createNewCqFtsSetup();
	}
	elsif (-d $optCqFtsHomeDir)
	{
		# yes, folder exist, go after CQ FTS XML config file
		
		my $optCqFtsHomeDir_org = $optCqFtsHomeDir;
		
		$optCqFtsHomeDir = findExistingCqFtsSetup();	# try where ever the user asked us to look
		if ($optCqFtsHomeDir eq "")
		{
			# we can't find our CQ FTS config file, so revert back to new setup
			$optCqFtsHomeDir = $optCqFtsHomeDir_org;
			createNewCqFtsSetup();
		}
		else
		{
			my $dieMsg = $langMsgText->maketext("FTS_ALREADY_EXIST", $ftsDBSetName, $ftsUserDbName, $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName."/");
			dealWithError("\n".$dieMsg."\n", 1);
		}
	}
	else
	{
		my $dieMsg = $langMsgText->maketext("CANNOT_BE_HERE");
		dealWithError("\n".$dieMsg."\n", 1);
	}
	
	push(@gCmdToExecute, ["Func_"."copy_fts_template"]);		# --copy_fts_template
	push(@gCmdToExecute, ["Func_"."gen_entity_file"]);			# --gen_entity_file
}


##############################################################################
# Returns the state we are in
sub getState
{
	if (-f $ftsStateFile)
	{
		my $data;
		
		# the file exist, read it and return the data
		if (open(FH, $ftsStateFile))
		{
			my $line;
			
			while ($line = <FH>)
			{
				$data .= $line;
			}
			close FH;
		}
		
		if (!defined($data))
		{
			$data = "=";
		}
		
		my ($function, $state) = split(/=/, $data);
		
		$function = trim($function);
		$state = trim($state);
		
		return ($function, $state);
	}
	else
	{
		return ("", "");
	}
}


##############################################################################
# Sets the state we should be in
sub setState
{
	my ($function, $state) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SET_STATE", $function, $state)."\n";
	
	my ($functionOld, $stateOld) = getState();
	
	if (($function eq "") and ($state eq ""))
	{
		# the state is being reset to nothing, so just delete the state file
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CLEARING_STATE")."\n";
		if ((-f $ftsStateFile) and (unlink($ftsStateFile) ne 1))
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $ftsStateFile)."\n", 1);
		}
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CURRENT_STATE", $function, $state)."\n";
	
		# the state for the current function is being change
		if (($functionOld ne "") and ($function ne $functionOld))
		{
			# somehow, we are trying to overwrite an existing state, this is an error
			dealWithError($langMsgText->maketext("UNEXPECTED_STATE_ERROR", $function, $functionOld)."\n", 1);
		}
		
		if ($stateOld ne "")
		{
			$stateOld .= ",";
		}
		
		$stateOld .= $state;
		
		my $data = $function."=".$stateOld;
		
		if (open(FH, ">$ftsStateFile"))
		{
			print FH $data;
			close FH;
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_WRITE_TO_FILE", $ftsStateFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
}


##############################################################################
# Completes the state transaction
sub setDoneState
{
	my ($state) = @_;
	
	if ($gTopLevelSate eq "setup_cq_fts")
	{
		if ($state eq "")
		{
			setState("", "");	# this means we are done, so reset the state
		}
		else
		{
			setState("setup_cq_fts", $state);
		}
	}
	elsif ($gTopLevelSate eq "remove_record_type")
	{
		if ($state eq "")
		{
			setState("", "");	# this means we are done, so reset the state
		}
		else
		{
			setState("remove_record_type", $state);
		}
	}
	elsif ($gTopLevelSate eq "add_record_type")
	{
		if ($state eq "")
		{
			setState("", "");	# this means we are done, so reset the state
		}
		else
		{
			setState("add_record_type", $state);
		}
	}
	elsif ($gTopLevelSate eq "typeahead_cmd_sync_idx")
	{
		if ($state eq "")
		{
			setState("", "");	# this means we are done, so reset the state
		}
		else
		{
			setState("typeahead_cmd_sync_idx", $state);
		}
	}
}


##############################################################################
# Checks to see if we are in a state or not
sub isInState
{
	my ($function, $state) = getState();
	
	if ($function ne "")
	{
		return 0;
	}
	else
	{
		return 1;
	}
}


##############################################################################
# Pushes the required functions to perform end-to-end setup of CQ FTS.  It 
# also sets the 'state' file so that re-execution of the this function (in 
# the event of a failure) knowns where to continue from.
# --setup_cq_fts
sub setupCQFts
{
	# 1) if the state file doesn't exist, create it
	# 2) if the state file exist, based on it's state push functions to be called
	
	dieIfCqFtsIsLocked();
	
	my ($function, $state) = getState();
	
	if (($function ne "") and ($function ne "setup_cq_fts"))
	{
		# we are in a state, and it's not our state
		dealWithError($langMsgText->maketext("IN_A_STATE_ALREADY", "setup_cq_fts", $function)."\n", 0);
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
	elsif ($function eq "")
	{
		# we are not in any state, so set one to our state of 'setup_cq_fts'
		setState("setup_cq_fts", "");
	}
	elsif ($function eq "setup_cq_fts")
	{
		# we are in our state for which we need to continue 
		setState("setup_cq_fts", "");
	}
	
	$gTopLevelSate = "setup_cq_fts";
	
	# we are going to do a default end-to-end setup
	
	if (index($state, "gen_fts_files") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."gen_fts_files"]);				# --gen_fts_files
	}
	if (index($state, "cust_fts_files") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."cust_fts_files", 0]);			# --cust_fts_files
	}
	if (index($state, "cust_solr_files") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."cust_solr_files", 1]);			# --cust_solr_files
	}
	if (index($state, "stop_fts_was_profile") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."stop_fts_was_profile", 0]);		# --stop_fts_was_profile
	}
	if (index($state, "set_solr_home") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."set_solr_home"]);				# --set_solr_home
	}
	if (index($state, "start_fts_was_profile") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."start_fts_was_profile", 1]);		# --start_fts_was_profile
	}
	if (index($state, "enable_cqweb_fts") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."enable_cqweb_fts", 1]);			# --enable_cqweb_fts
	}
	if (index($state, "run_batch_idx") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."run_batch_idx"]);				# --run_batch_idx
	}
	if (index($state, "start_update_idx") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."start_update_idx"]);				# --start_update_idx
	}
	if (index($state, "clear_state") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."clear_state"]);					# --clear_state
	}
}


##############################################################################
# Starts a WAS profile
sub startFtsWASProfile
{
	my ($profileName, $dieOnError) = @_;
	
	my @mustMatch = $langMsgText->maketext("START_WAS_PROFILE");
	my $cmd = $Q.$ftsFtsWasAppSrvBinDir."startServer".$ext.$Q." server1 -profileName ".$profileName;
	my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Stops a WAS profile
sub stopFtsWASProfile
{
	my ($profileName, $dieOnError) = @_;
	
	my @mustMatch = $langMsgText->maketext("STOP_WAS_PROFILE");
	my $cmd = $Q.$ftsFtsWasAppSrvBinDir."stopServer".$ext.$Q." server1 -profileName ".$profileName;
	my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Sets Solr's home directory
sub setSolrHome
{
	my ($profileName, $dataDir, $dieOnError) = @_;
	
	my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -lang jython -profileName ".$profileName." -f ".$Q.$ftsSetSolrHomePY.$Q." -interactive:no ".$Q.$dataDir."Solr/solr".$Q;
	
	my @mustMatch = $langMsgText->maketext("SET_SOLR_HOME");
	my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Output file about running multiutil on Windows
sub runMultiutilOnUnix
{
	my $multiutilOnUnixData = $langMsgText->maketext("UNIX_MULTIUTIL_INFO", 
									$ftsDBSetName, $ftsCQPropertyFile, $optCqFtsHomeDir, 
									$ftsUserName, $ftsDBSetName, $ftsUserDbName, $ftsTargetDir);
	
	if ($gIsWindows eq 0)
	{
		if (!(-f $multiutilMsgFile))
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("UNIX_MULTIUTIL_MSG_FILE", $multiutilMsgFile)."\n";
		
			if (open(FH2, ">:utf8", $multiutilMsgFile))
			{
				print FH2 $multiutilOnUnixData;
				close FH2;
			}
			else
			{
				dealWithError($langMsgText->maketext("CANNOT_CREATE", $multiutilMsgFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
			}
		}
		
		if ($ftsIsEnableCQFTS eq "true")
		{
			printAndLog $msgLabel_WARN.""."\n", 1;
			printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_NOT_SUPPORTED_SCRUB")."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_SEE_FILE", $multiutilMsgFile)."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_HOW_TO_SCRUB")."\n", 1;
			printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
			printAndLog $msgLabel_WARN.""."\n", 1;
		}
	}
}


##############################################################################
# Finds the web.xml folder of a WAS profile
sub findFtsWebXMLFolder
{
	my ($profileName) = @_;
	
	my $status = 0;
	my $profilePath = "";
	
	($status, $profilePath) = getProfilePath($profileName, 1);
	
	my $lookingInCellsDir = $profilePath."config/cells/";
	
	if (index($lookingInCellsDir, ' ') ne -1)
	{
		$lookingInCellsDir = $Q.$lookingInCellsDir.$Q;
	}
	
	my $webXMLDir = "";
	
	my @files = <$lookingInCellsDir*>;
	foreach my $cellDir (@files)
	{
		my $temp = $cellDir."/applications/cqtsdbcrawler.ear/deployments/cqtsdbcrawler/cqtsdbcrawler.war/WEB-INF/";
		
		if (-d $temp)
		{
			$webXMLDir = $temp;
			last;
		}
	}
	
	return $webXMLDir;
}


##############################################################################
# Finds the server.xml folder of a WAS profile
sub findFtsServerXMLFolder
{
	my ($profileName) = @_;
	
	my $status = 0;
	my $profilePath = "";
	
	($status, $profilePath) = getProfilePath($profileName, 1);
	
	my $lookingInCellsDir = $profilePath."config/cells/";	
	
	if (index($lookingInCellsDir, ' ') ne -1)
	{
		$lookingInCellsDir = $Q.$lookingInCellsDir.$Q;
	}
	
	my $serverXMLDir = "";
	
	my @files = <$lookingInCellsDir*>;
	foreach my $cellDir (@files)
	{
		my $lookingInNodesDir = $cellDir."/nodes/";
		
		if (-d $lookingInNodesDir)
		{
			if (index($lookingInNodesDir, ' ') ne -1)
			{
				$lookingInNodesDir = $Q.$lookingInNodesDir.$Q;
			}
			
			my @nodeDir = <$lookingInNodesDir*>;
			foreach my $nodeDir (@nodeDir)
			{
				my $temp = $nodeDir."/servers/server1/";
				if (-d $temp)
				{
					$serverXMLDir = $temp;
					last;
				}				
			}
		}
	}
	
	return $serverXMLDir;
}


##############################################################################
# Finds out the service start up type (Windows only)
sub getServiceStartupType
{
	my ($profileName) = @_;
	
	my $foundStartupType = "manual";	# default to 'manual'
	
	if ($gIsWindows eq 1)
	{
		my $cmd = "sc qc \"IBMWAS61Service - ".$profileName."\"";	# sc qc "IBMWAS61Service - xyz"
		
		# execute the command
		if (open(FH, "$cmd 2>&1 |"))
		{
			my $line;
			
			while ($line = <FH>)
			{
				if ($line =~ /START_TYPE/)
				{
					if ($line =~ /AUTO_START/)
					{
						$foundStartupType = "automatic";
					}
					elsif ($line =~ /DEMAND_START/)
					{
						$foundStartupType = "manual";
					}
					elsif ($line =~ /DISABLED/)
					{
						$foundStartupType = "disabled";
					}
				}
			}
			close FH;
		}
	}
	
	return $foundStartupType;
}


##############################################################################
# Finds out the JVM heap size and the parameter
sub getHeapSizeAndJVMArg
{
	my ($serverXMLName) = @_;
	
	my $genericJvmArg = "";
	my $maxHeapSize = "";
	my $ftsHome = "";
	
	$serverXMLName =~ s/$Q//sg;
	
	if (open(FH, $serverXMLName))
	{
		my $line;
		while ($line = <FH>)
		{
			$line = trim($line);
			
			# With CQ 8.0 'solr.solr.home' is stored in 'systemProperties' as name/value pairs
			if ($line =~ /solr.solr.home/)
			{
				# <systemProperties ... ... name="solr.solr.home" value="C:/CQ.Search/TextSearch_SAMPL/Solr/solr"/>
				$ftsHome = $line;
				$ftsHome =~ s{
					<systemProperties.*?
					name
					.*?
					=.*?"solr.solr.home"
					.*?
					value
					.*?
					=.*?"(.*?)"
					.*?
					>
				}{$1}x;
				
				# remove the "Solr/solr" part from the 'value' part that we extracted above
				$ftsHome =~ s{
					(.*?)(Solr/solr)
				}{$1}x;
			}
			
			if ($line =~ /genericJvmArguments/)
			{
				# matching: <jvmEntries ... ... genericJvmArguments="XXXXX" ... ...>
				$genericJvmArg = $line;
				$genericJvmArg =~ s{
					<jvmEntries.*?
					genericJvmArguments
					.*?
					=.*?"(.*?)"
					.*?
					>
				}{$1}x;
				
				# matching: -Dsolr.solr.home=C:/CQ.Search/TextSearch_SAMPL/Solr/solr
				$genericJvmArg =~ s{
					(-Dsolr.solr.home=)(.*?)
				}{$2}x;
				
				# Starting with CQ 8.0, 'solr.solr.home' is store in 'systemProperties'.
				# So if we have not yet processed 'solr.solr.home' from 'systemProperties'
				# see if it's in 'genericJvmArguments' and process it.
				if ($ftsHome eq "")
				{
					# matching: C:/CQ.Search/TextSearch_SAMPL/Solr/solr
					$ftsHome = $genericJvmArg;
					$ftsHome =~ s{
						(.*?)(Solr/solr)
					}{$1}x;
				}
			}
			
			if ($line =~ /maximumHeapSize/)
			{
				# matching: <jvmEntries ... ... maximumHeapSize="XXXXX" ... ...>
				$maxHeapSize = $line;
				$maxHeapSize =~ s{
					<jvmEntries.*?
					maximumHeapSize
					.*?
					=.*?"(.*?)"
					.*?
					>
				}{$1}x;
			}
		}
		close FH;
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $serverXMLName)."\n", 0);
	}
	
	if ($ftsHome ne "")
	{
		$ftsHome =~ s/\\/\//sg;
		if (substr($ftsHome, -1) ne "/")
		{
			$ftsHome .= "/";
		}
	}
	
	return ($genericJvmArg, $maxHeapSize, $ftsHome);
}


##############################################################################
# Removes the heap size setting from the server.xml file
sub removeHeapSizeXMLTag
{
	my ($serverXMLName) = @_;
	
	my $status = 1;		# assume we will not fail
	
	if (open(FH, $serverXMLName))
	{
		my $serverXMLData = "";
		my $wasChanged = 0;
		
		my $line;
		while ($line = <FH>)
		{
			if ($line =~ /maximumHeapSize/)
			{
				# matching: <jvmEntries ... ... maximumHeapSize="XXXXX" ... ...>
				# to remove: maximumHeapSize="XXXXX"
				$line =~ s{
					(.*?)(<jvmEntries)(.*?)
					(maximumHeapSize)(.*?)(=)(.*?)(")(.*?)(")
					(.*?)(>)(.*?)
				}{$1$2$3$11$12$13}x;
				
				$wasChanged = 1;
			}
			
			$serverXMLData .= $line;
		}
		close FH;
		
		if ($wasChanged eq 1)
		{		
			$status = safeFileUpdate($serverXMLName, $serverXMLData, 0);
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $serverXMLName)."\n", 0);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Creates a script to be executed which will report if the CMD is in elevated
# adminstrative mode or not
sub createAdminElevationTestScript
{
	my ($dieOnError) = @_;
	
	my $scriptData = <<EOF;
echo OFF
NET SESSION >nul 2>&1
IF %ERRORLEVEL% EQU 0 (
    ECHO Administrator PRIVILEGES Detected!
) ELSE (
    ECHO NOT AN ADMIN!
)
EOF
	
	my $theFile;
	
	eval
	{
		$theFile = $ftsDataDir."adminTest".$ext;
		
		if (!open(FH, ">$theFile"))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $theFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		}
		print FH $scriptData;
		close FH;
	};
	if ($@)
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $theFile)."\n", $dieOnError);
	}
}


##############################################################################
# Delete elevated privilage check script
sub deleteAdminElevationTestScript
{
	my $theFile = $ftsDataDir."adminTest".$ext;
	
	if (-e $theFile)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_FILE", $theFile)."\n";
		if (unlink($theFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 0);
		}
	}
}


##############################################################################
# Creates the Solr deployment (under WAS) script
sub createDeplySolrScript
{
	my ($dbSetName, $userDbName, $solrVer, $dieOnError) = @_;
	
	my $status = 1;
	
	my $scriptData = <<EOF;

#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2008, 2011.  All Rights Reserved
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#

import sys, java

#-----------------------------------------------------------------
# Deploy Solr
#-----------------------------------------------------------------
def deploySolr(appLocation, serverName, contextRoot, solrVer):

   global AdminApp
   global AdminConfig

   print "deploySolr, running with: "
   print "	Application Location: " + appLocation
   print "	Server name: " + serverName
   print "	Solr ver: " + solrVer
   
   apacheSolr = "apache-solr-" + solrVer
   
   appArgs = "[-server " + serverName + " -contextroot " + contextRoot + " -appname " + apacheSolr + " -MapWebModToVH [[" + apacheSolr + ".war " + apacheSolr + ".war,WEB-INF/web.xml default_host]]]"

   AdminApp.install(appLocation, appArgs)

   AdminConfig.save()
   # the server should be restarted before trying to use the datasource


#-----------------------------------------------------------------
# Main
#-----------------------------------------------------------------
if (len(sys.argv) != 4):
   print "deploySolr: this script requires 3 parameters: " 
   print "	  1) The name of the node on which the application will be deployed,"
   print "	  2) The name of the server under which to create resources,"
   print "	  3) The contextRoot,"
   print "	  4) The Solr version"

   print ""
   print "e.g.:     deploySolr C:/myApps/myApp1.war server1 contextRoot"  
else:
   deploySolr(sys.argv[0], sys.argv[1], sys.argv[2], sys.argv[3])
EOF
	
	my $theFile;
	
	eval
	{
		$theFile = $ftsTargetDir.$dbSetName."_".$userDbName."/"."deploySolr.py";
		printAndLog $msgLabel_INFO.$langMsgText->maketext("WAS_PROFILE_DEPOLY_FILE", $theFile)."\n";
		
		if (!open(FH, ">$theFile"))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $theFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
			$status = 0;
		}
		print FH $scriptData;
		close FH;
	};
	if ($@)
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $theFile)."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Deletes the Solr deploy file
sub deleteDeplySolrScript
{
	my ($dbSetName, $userDbName, $dieOnError) = @_;
	
	my $theFile = $ftsTargetDir.$dbSetName."_".$userDbName."/"."deploySolr.py";
	
	if (-e $theFile)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_FILE", $theFile)."\n";
		if (unlink($theFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", $dieOnError);
		}
	}
}


##############################################################################
# Figures out the CQ Web profile path (if it exist on this machine)
sub findProfilePath_cqweb
{
	my $status = 0;
	my $cqWebProfilePath = "";
	my $cqWebProfileName = "";
	
	getListOfProfilesName(0);
	
	my $foundIt = 0;
	
	foreach my $profileName (@ftsExistingProfileNames)
	{
		($status, $cqWebProfilePath) = getProfilePath($profileName, 0);
		
		my $lookingInCellsDir = $cqWebProfilePath."config/cells/";
		
		if (index($lookingInCellsDir, ' ') ne -1)
		{
			$lookingInCellsDir = $Q.$lookingInCellsDir.$Q;
		}
		
		my @files = <$lookingInCellsDir*>;
		foreach my $cellDir (@files)
		{
			my $temp = $cellDir."/applications/TeamEAR.ear/TeamEAR.ear";
			
			if (-e $temp)
			{
				$foundIt = 1;
				last;
			}
		}
		
		if ($foundIt == 1)
		{
			$cqWebProfileName = $profileName;
			last;
		}
	}
	
	if ($foundIt == 0)
	{
		$cqWebProfilePath = "";
		$cqWebProfileName = "";
	}
	
	return ($cqWebProfileName, $cqWebProfilePath);
}


##############################################################################
# Creates the CQ FTS admin configuration file
sub createFTSConfigXMLFile
{
	my ($configFile) = @_;
	
	# based on the local, read the specific configuration file
	my $configFileName = $ftsClearQuestHomeDir."CQTsAdmin/L10N/".$langMsgText->maketext("LANGUAGE_CODE").".xml";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("READING_FTS_CONFIG_XML_FILE", $configFileName)."\n";
	
	if (!open FH, $configFileName)
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $configFileName)."\n", 1);
	}
	my @temp = <FH>;
	close FH;
	
	my $portsFileToUse = "";
	
	# if they gave us a ports file, copy it to our FTS home directory and use it
	if ($optFtsWasProfPortsFile ne "")
	{
		my $foundAt;
		my $portsFileName;
		
		$foundAt = rindex($ftsWASProfilePortsDefPropFile, "/") + 1;
		if ($foundAt ne -1)
		{
			$portsFileName = substr($ftsWASProfilePortsDefPropFile, $foundAt);
		}
		
		if (!copy($ftsWASProfilePortsDefPropFile, $ftsDataDir.$portsFileName))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $ftsWASProfilePortsDefPropFile, $ftsDataDir.$portsFileName)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		
		$portsFileToUse = $ftsDataDir.$portsFileName;
	}
	
	my $configData = join("",@temp);
	
	my $creationDate = "".localtime();
	$configData =~ s/\[date\]/$creationDate/s;
	
	# my $configData = $langMsgText->maketext("CONFIG_FILE_DATA", "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>", "".localtime());
	my $entityFile = $ftsDataDir."Entity-".$ftsDBSetName."-".$ftsUserDbName.".txt";
	my $siteName = "";
	
	if (($ftsIsReplicated eq 1) and ($ftsLocalReplicaName ne "") and ($ftsLocalReplicaName ne "<local>"))
	{
		$siteName = HTML::Entities::decode($ftsLocalReplicaName);
	}
	elsif (($ftsIsReplicated eq 0) and ($ftsSiteName ne ""))
	{
		$siteName = HTML::Entities::decode($ftsSiteName);
	}
	
	$configData =~ s/\[destination\]/$ftsTargetDir/s;
	$configData =~ s/\[dbSet\]/$ftsDBSetName/s;
	$configData =~ s/\[userDB\]/$ftsUserDbName/s;
	$configData =~ s/\[localhost\]/$ftsFtsServerName/s;
	$configData =~ s/\[entity-file\]/$entityFile/s;
	$configData =~ s/\[siteName\]/$siteName/s;
	$configData =~ s/\[was-profile-home\]/$ftsWASProfileHomeDir/s;
	$configData =~ s/\[was-profile-ports-file\]/$portsFileToUse/s;
	
	if ($optCreateCqFtsWasProfile eq "1")
	{
		my $wasProfileNamePrefix = "cqsearchprofile_";
		
		# we are being asked for a new profile creation during initial setup
		$ftsWASProfileName = $wasProfileNamePrefix.$ftsDBSetName."_".$ftsUserDbName;
		if (length($ftsWASProfileName) > $gMaxWASProfileLen)
		{
			my $oldName = $ftsWASProfileName;
			
			$wasProfileNamePrefix = "cqfts_";	# try a shorter name
			$ftsWASProfileName = $wasProfileNamePrefix.$ftsDBSetName."_".$ftsUserDbName;
			if (length($ftsWASProfileName) > $gMaxWASProfileLen)
			{
				dealWithError($langMsgText->maketext("PROFILE_NAME_TOO_LONG", $ftsWASProfileName, $gMaxWASProfileLen)."\n", 1);
			}
			else
			{
				printAndLog $msgLabel_WARN.$langMsgText->maketext("WAS_PROFILE_NAME_TOO_LONG", $gMaxWASProfileLen, $ftsWASProfileName, $oldName)."\n", 1;
			}
		}
		
		$ftsFtsWASProfilePort = findNextAvailablePortNumber(1);
		
		$configData =~ s/<newValue required="no">cqsearchprofile_\[db-set-name\]_\[user-db-name\]<\/newValue>/<newValue required="no">$ftsWASProfileName<\/newValue>/s;
		$configData =~ s/<newValue required="no">\[\*\]<\/newValue>/<newValue required="no">$ftsFtsWASProfilePort<\/newValue>/s;
	}
	else
	{
		# give them the default (i.e.: use the default profile and port)
		$configData =~ s/<newValue required="no">cqsearchprofile_\[db-set-name\]_\[user-db-name\]<\/newValue>/<newValue required="no"><\/newValue>/s;
		$configData =~ s/<newValue required="no">\[\*\]<\/newValue>/<newValue required="no"><\/newValue>/s;
	}
	
	eval
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_FTS_CONFIG_XML_FILE", $configFile)."\n";
		
		if (!open(FH, ">$configFile"))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $configFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		print FH $configData;
		close FH;
	};
	if ($@)
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $configFile)."\n", 1);
	}
}


##############################################################################
# Gets the copy-field data off the generated Solr file
sub getSolrCopyFieldData
{
	my ($fileName) = @_;
	
	my $line;
	my @lines;
	my $data = "";
	
	if (!open FH, $fileName)
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $fileName)."\n", 1);
	}
	@lines = <FH>;
	close FH;
	
	foreach $line ( @lines )
	{
		if ($line =~ /<copyField source=".+" dest="RATL_CQ_ALL_FIELDS"><\/copyField>/)
		{
			$data .= $line;
		}
	}
	
	if ($data eq "")
	{
		# we are in 7.1.1 and later
		foreach $line ( @lines )
		{
			if ($line =~ /<copyField dest="RATL_CQ_ALL_FIELDS" source=".+"\/>/)
			{
				$data .= $line;
			}
		}
	}
	
	return $data;
}


##############################################################################
# Gets the fields data from Solr's schema.xml file
sub getSolrFieldData
{
	my ($fileName) = @_;

	my $line;
	my @lines;
	my $data = "";
	
	open FH, $fileName;
	@lines = <FH>;
	close FH;
	
	foreach $line ( @lines )
	{
		if ($line =~ /<field name=".+" type="(text|string)" indexed="(true|false)" stored="(true|false)" required="(true|false)" multiValued="(true|false)"><\/field>/)
		{
			$data .= $line;
		}
	}
	
	if ($data eq "")
	{
		# we are in 7.1.1 and later
		foreach $line ( @lines )
		{
			if ($line =~ /<field indexed="(true|false)" multiValued="(true|false)" name=".+" required="(true|false)" stored="(true|false)" type="(text|string)"\/>/)
			{
				$data .= $line;
			}
		}
	}
	
	return $data;
}


##############################################################################
# Reads CQ FTS admin configuration file
sub readConfigData
{
	my ($dataFile) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("READING_FTS_CONFIG_XML_FILE", $dataFile)."\n";
	
	if (!(-f $dataFile))
	{
		dealWithError($langMsgText->maketext("NO_FTS_AT_FTSHOME", $ftsDBSetName, $ftsUserDbName, $ftsTargetDir)."\n", 1);
	}
	
	my $parser = XML::Parser->new(ErrorContext => 3);
	
	$parser->setHandlers(	Start	=>	\&startElement,
							End		=>	\&endElement,
							Char	=>	\&characterData	);
	
	eval
	{
		$parser->parsefile($dataFile);
	};
	
	if ($@)
	{
		# report any error that stopped parsing, or announce success
		$@ =~ s/at \/.*?$//s;
		my $err = $@;
		dealWithError($langMsgText->maketext("ERROR_IN_CFG_FILE", $dataFile)."\n".$err."\n", 1);
	}
}


##############################################################################
# Creates an Entity file without the '&' in it (if the source has it)
sub getEntityFileWithoutDispFieldInfo
{
	my ($entityFile, $createCleanEntityFile, $dieOnError) = @_;
	
	if (!open(FH, $entityFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $entityFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
	}
	
	my @newEntityData;
	
	my $line;
	while ($line = <FH>)
	{
		# Project=ratl_mastership,ratl_keysite,dbid,is_active,version,lock_version,record_type,&Name,Description,
		if ($line =~ /&/)
		{
			my $foundAt = 0;
			
			# get the record-type name
			$foundAt = index($line, "=");
			if ($foundAt ne -1)
			{
				my $recordTypeName;
				my $fieldName;
				
				$recordTypeName = trim(substr($line, 0, $foundAt));
				
				# now get the list of fields
				$foundAt = 0;
				while ($foundAt ne -1)
				{
					$foundAt = index($line, "&", $foundAt);
					if ($foundAt ne -1)
					{
						$foundAt = $foundAt + 1;
						
						my $endAt = index($line, ",", $foundAt);
						if ($endAt eq -1)
						{
							$fieldName = trim(substr($line, $foundAt));
						}
						else
						{
							$fieldName = trim(substr($line, $foundAt, $endAt - $foundAt));
						}
						
						$gDispOrdFieldList{$recordTypeName} = $fieldName;
					}
				}
				
				$line =~ s/&//g;
				
				push(@newEntityData, $line);
			}
		}
		else
		{
			# this is a record-type with no display field setting
			my $foundAt = index($line, "=");
			if ($foundAt ne -1)
			{
				$gDispOrdFieldList{trim(substr($line, 0, $foundAt))} = "";
				
				push(@newEntityData, $line);
			}
		}
	}
	
	close FH;
	
	my $entityFileClean;
	my $isNew = 0;
	
	if ((@newEntityData ne 0) and ($createCleanEntityFile eq 1))
	{
		$entityFileClean = $entityFile."-clean";
	
		if (!open(FH, ">".$entityFileClean))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $entityFileClean)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		}
		
		foreach my $line (@newEntityData)
		{
			print FH $line;
		}
		
		close FH;
		
		$entityFile = $entityFileClean;
		$isNew = 1;
	}
	
	return ($entityFile, $isNew);
}


##############################################################################
# Generates Solr intermediate file
sub generateSolrIntermediateFile
{
	my ($entityFile, $dieOnError) = @_;
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $status = 1;
	
	my ($entityFileClean, $isUseNewFile) = getEntityFileWithoutDispFieldInfo($entityFile, 1, $dieOnError);
	
	my $solrParm = $parm." -o Solr -f ".$entityFileClean;
	
	my @mustMatch = $langMsgText->maketext("GEN_INTER_SOLR_FILE");
	executeExternalCommand($cmd.$solrParm, @mustMatch, $gDieKind_must_match, $dieOnError);
	validateFileGenerated($cmd.$solrParm, $ftsClearQuestHomeDir.$ftsSolrIntermediateFileName, $dieOnError);
	if (move($ftsClearQuestHomeDir.$ftsSolrIntermediateFileName, $ftsDataDir.$ftsSolrIntermediateFileName) ne 1)
	{
		dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE", $ftsClearQuestHomeDir.$ftsSolrIntermediateFileName, $ftsDataDir.$ftsSolrIntermediateFileName)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	if ($isUseNewFile eq 1)
	{
		# delete the temporary 'clean' Entity file
		if (unlink($entityFileClean) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $entityFileClean)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
	}
	
	return $status;
}


##############################################################################
# Generates CQ Search property file
sub generateCQSearchPropertyFile
{	
	my ($entityFile, $moveTo, $dieOnError) = @_;
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $status = 1;
	
	my ($entityFileClean, $isUseNewFile) = getEntityFileWithoutDispFieldInfo($entityFile, 1, $dieOnError);
	
	my $cqParm = $parm." -o CQ -f ".$entityFileClean;
	
	my @mustMatch = $langMsgText->maketext("GEN_CQ_PROP_XML_FILE");
	executeExternalCommand($cmd.$cqParm, @mustMatch, $gDieKind_must_match, $dieOnError);
	validateFileGenerated($cmd.$cqParm, $ftsClearQuestHomeDir.$ftsCQPropertyFileName, $dieOnError);
	if (move($ftsClearQuestHomeDir.$ftsCQPropertyFileName, $moveTo) ne 1)
	{
		dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE_TO", $ftsClearQuestHomeDir.$ftsCQPropertyFileName, $moveTo).
						"\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	if ($isUseNewFile eq 1)
	{
		# delete the temporary 'clean' Entity file
		if (unlink($entityFileClean) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $entityFileClean)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
	}
	
	return $status;
}


##############################################################################
# Find where WebSphere is installed
sub findWASAppServerDir
{
	# IM's WAS AppServer file is in: %RATIONAL_HOME%\was_info.txt"
	my $appSrvFile = $ftsClearQuestHomeDir."was_info.txt";
	my $foundAppSrvPath = "";
	
	if (-f $appSrvFile)
	{
		my @data;
		if (open(FH, $appSrvFile))
		{
			@data = <FH>;
			close(FH);
		}
		
		# we are looking for one of those two lines in the file 'was_info.txt':
		#   WAS_HOME=C:\Program Files\IBM\WebSphere\AppServer
		#   WAS_PROFILE=C:\Program Files\IBM\WebSphere\AppServer\profiles\AppSrv01
		
		my $line;
		foreach $line (@data)
		{
			$line = trim($line);
			if ($line =~ m/WAS_HOME=/)
			{
				# we found: WAS_HOME=C:\Program Files\IBM\WebSphere\AppServer
				$line =~ s/WAS_HOME=//;
				$foundAppSrvPath = trim($line);
			}
			elsif ($line =~ /WAS_PROFILE=/)
			{
				# we found: WAS_PROFILE=C:\Program Files\IBM\WebSphere\AppServer\profiles\AppSrv01
				$line =~ s/WAS_PROFILE=//;
				$foundAppSrvPath = trim($line);
			}
			
			# do we have our path, and is it the correct one?
			if (($foundAppSrvPath ne "") and (-f $foundAppSrvPath."/bin/manageprofiles".$ext))
			{
				last;
			}
			else
			{
				$foundAppSrvPath = "";	# keep looking or we haven't found it yet
			}
		}
	}
	
	if ($foundAppSrvPath eq "")
	{
		dealWithError($langMsgText->maketext("NEED_WAS_APPSERVER_PATH", "CQFTS_AppServer_HOME")."\n", 1);
	}
	
	return $foundAppSrvPath;
}


##############################################################################
# Builds a scenarios data map for scenarios help
my %scenarios = ();
sub buildScenariosFromXMLFile
{
	my ($tag, $current) = @_;
	
	my $type;
	for (my $k = 0; $k < @$current; $k++)
	{
		if ($type = ref $$current[$k])
		{
			if ($type eq "ARRAY")
			{
				buildScenariosFromXMLFile($tag, $$current[$k]);
			}
			elsif ($type eq "HASH")
			{
				my $temp = $$current[$k];
				if (scalar keys %$temp ne 0)
				{
					die "\n\n\nFATAL ERROR: should not be here!! -- displayhash()\n\n";
				}
			}
		}
		else
		{
			my $info = $$current[$k];
			
			if ($info eq "Scenario")
			{
				if (scalar keys %scenarios ne 0)
				{
					$gFtsScenarios{$scenarios{"id"}} = {%scenarios};
					%scenarios = ();
				}
			}
			
			if (($info eq "id") or ($info eq "headline") or ($info eq "body"))
			{
				$tag = trim($info);
			}
			elsif (($tag ne "") and ($info ne 0) and (trim($info) ne ""))
			{
				$scenarios{$tag} = trim($info);
			}
		}
	}
	
	if ((scalar keys %scenarios ne 0) and ($tag eq ""))
	{
		if (!(exists $gFtsScenarios{$scenarios{"id"}}))
		{
			# the last element was not pushed into the hash
			$gFtsScenarios{$scenarios{"id"}} = {%scenarios};
			%scenarios = ();
		}
	}
}


##############################################################################
# Returns the full path to where FTS template files are
sub getCQSolrTemplateDir
{
	my ($solrVer) = @_;
	
	my $cqSolrSourceDir = $gCQSolrVersionInfo{$solrVer};
	
	if (!(defined @$cqSolrSourceDir[0]) or (length(@$cqSolrSourceDir[0]) == 0))
	{
		dealWithError($langMsgText->maketext("UNDEFINED_SOLR_VER", $solrVer)."\n", 1);
	}
	
	return $ftsClearQuestHomeDir.@$cqSolrSourceDir[0];
}


##############################################################################
# Returns the full path as well as the Solr WAR file
sub getSolrWARFile
{
	my ($solrVer) = @_;
	
	my $solrWarFile = $gCQSolrVersionInfo{$solrVer};
	
	if (!(defined @$solrWarFile[1]) or (length(@$solrWarFile[1]) == 0))
	{
		dealWithError($langMsgText->maketext("UNDEFINED_SOLR_VER", $solrVer)."\n", 1);
	}
	
	return $ftsClearQuestHomeDir.@$solrWarFile[1];
}


##############################################################################
# Determines information about the current deployment used to upgrade Solr
sub getCurrDeployInfoForSolrUpgrade
{
	my ($ftsHome) = @_;
	
	my $DH;
	my $cfgFile;
	my $dbSetName = "";
	my $userDbName = "";
	my $solrVersion = "";
	
	# From FTS home, extract the dbset and userdb info.
	my $foundAt = rindex($ftsHome, "/", length($ftsHome) - 2) + 1;
	if ($foundAt ne -1)
	{
		my $tmp = substr($ftsHome, $foundAt, -1);
		$tmp =~ s{(.*)_(.*)}{$1$2}x;
		$dbSetName = $1;
		$userDbName = $2;
	}
	else
	{
		# cannot be here: the value in $ftsHome is wrong
		printAndLog "\n [".$ftsHome."]", 1;
		dealWithError("\n".$langMsgText->maketext("CANNOT_BE_HERE")."\n", 1);
	}
	
	$cfgFile = $ftsHome."cqtsadmin-".$dbSetName."-".$userDbName.".xml";
	
	# If we found our file, then we have the dbset and userdb, we now need to find out Solr's version
	if ((defined $cfgFile) and (length($cfgFile) != 0))
	{
		my $line;
		my $data;
		
		if (!open(FH, $cfgFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $cfgFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			$line = trim($line);
			$data .= $line;
		}
		close FH;
		
		# Find the <solrVersion> block, and extract the Solr version info out of it
		my $defaultValueSolrVer = "";
		my $newValueSolrVer = "";
		
		$data =~ s{<solrVersion>(.*?)</solrVersion>}{$1}x;
		my $solrVersionBlock = $1;
		if (defined $solrVersionBlock)
		{
			my $tmp;
		
			# We did find "<solrVersion>", pull out the Solr version from it
			$tmp = $solrVersionBlock;
			$tmp =~ s{.*?<newValue.*?required="no">(.*?)</newValue>.*?}{$1}sgx;
			$newValueSolrVer = $1;
			
			$tmp = $solrVersionBlock;
			$tmp =~ s{.*?<defaultValue>(.*?)</defaultValue>.*?}{$1}sgx;
			$defaultValueSolrVer = $1;
			
			if ((defined $newValueSolrVer) and (length($newValueSolrVer) != 0))
			{
				$solrVersion = $newValueSolrVer;
			}
			elsif ((defined $defaultValueSolrVer) and (length($defaultValueSolrVer) != 0))
			{
				$solrVersion = $defaultValueSolrVer;
			}
		}
		else
		{
			# We did not find "<solrVersion>", this means we are on pre CQ 8.0.0.5 so default to Solr 1.2.0
			$solrVersion = $gSolrVersion_120;
		}
	}
	
	return ($solrVersion, $dbSetName, $userDbName, $cfgFile);
}


##############################################################################
sub upgradeSolrApp
{
	my ($dbSetName, $userDbName, $profileName, $cfgFile, $fromSolrVer, $toSolrVer, $dieOnError) = @_;
	
	my $status = 1;
	
	my $curFtsDataDir = $ftsTargetDir.$dbSetName."_".$userDbName."/";
	
	my $solrSchemaFile = $curFtsDataDir."Solr/solr/conf/schema.xml";
	my $solrSchemaFileBackup = $solrSchemaFile."upgradeSolrApp";
	
	# Create backup of the existing schema.xml, and solrconfig.xml files
	if ($status eq 1)
	{
		if (!copy($solrSchemaFile, $solrSchemaFileBackup))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE_BACKUP_FILE", $solrSchemaFile, $solrSchemaFileBackup)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		}
	}
	
	# Un-deploy current Solr
	if ($status eq 1)
	{
		my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -lang jython -c \"AdminApp.uninstall('apache-solr-".$fromSolrVer."')\"";
		my @mustMatch = $langMsgText->maketext("UNDEPLOY_SOLR");
		$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, 1);
	}
	
	# Deploy latest Solr
	if ($status eq 1)
	{
		my $solrWarFile = getSolrWARFile($toSolrVer);
		
		$status = createDeplySolrScript($dbSetName, $userDbName, $toSolrVer, 0);
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -lang jython -profileName ".$profileName." -f ".$Q.$curFtsDataDir."deploySolr.py".$Q." ".$Q.$solrWarFile.$Q." server1 solr ".$toSolrVer;
			my @mustMatch = $langMsgText->maketext("DEPLOY_SOLR");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsFtsWasAppSrvBinDir."wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -c ".$Q."\$AdminConfig save".$Q;
			my @mustMatch = $langMsgText->maketext("DEPLOY_SOLR_SAVE");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
		
		deleteDeplySolrScript($dbSetName, $userDbName, 0);
	}
	
	# Copy over the new Solr template
	if ($status eq 1)
	{
		$ftsSourceDir = getCQSolrTemplateDir($toSolrVer);
		my $tmp = $ftsTargetDir;	# remember it so we can restore it
		
		$ftsTargetDir = $curFtsDataDir;
		
		followDir($ftsSourceDir, \&copyDirOverwriteMkdir);
		
		$ftsTargetDir = $tmp;
	}
	
	# Update schema.xml
	if ($status eq 1)
	{
		# from $solrSchemaFileBackup copy over to $solrSchemaFile "<fields>" and "<copyField>"
		if (open(FH, $solrSchemaFile))
		{
			my $line;
			my $data;
			while ($line = <FH>)
			{
				$data .= $line;
			}
			close FH;
			
			# Reflect the user-db name in schema.xml
			$data =~ s/<schema name=".+" version=>/<schema name="$userDbName" version=>/;
			
			# Replace old <field>...</field> with new data based on user's DB (off their old schema.xml file)
			my $solrFieldData = getSolrFieldData($solrSchemaFileBackup);
			$data =~ s/<fields>.+<\/fields>/<fields>\n$solrFieldData<\/fields>/s;
			
			# Replace old <copyField>...</copyField> with new data based on user's DB (off their old schema.xml file)
			my $solrCopyFieldData = getSolrCopyFieldData($solrSchemaFileBackup);
			
			if ($data =~ /^\s+(<copyField source=".+" dest=".+"><\/copyField>)\s+$/smg)
			{
				# pre 7.1.0.2
				# <copyField source="Notes_Log" dest="RATL_CQ_ALL_FIELDS"></copyField>
				$data =~ s/^\s+(<copyField source=".+" dest=".+"><\/copyField>)\s+$/$solrCopyFieldData/smg;
			}
			else
			{
				# post 7.1.0.2
				# <copyField dest="RATL_CQ_ALL_FIELDS" source="Action_Types"/>
				$data =~ s/^\s+(<copyField dest=".+" source=".+"\/>)\s+$/$solrCopyFieldData/smg;
			}
			
			printAndLog $msgLabel_INFO.$langMsgText->maketext("MODIFYING_SOLR_SCHEMA_FILE", $solrSchemaFile)."\n";
			
			$status = safeFileUpdate($solrSchemaFile, $data, $dieOnError);
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $solrSchemaFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
			$status = 0;
		}
	}
	
	# Update cqtsadmin-<dbset>-<userdb>.xml
	if ($status eq 1)
	{
		my $solrVersionBlock;
		
		# Read the language specific .xml file to extract the "<solrVersion>" xml block from it
		my $configFileName = $ftsClearQuestHomeDir."CQTsAdmin/L10N/".$langMsgText->maketext("LANGUAGE_CODE").".xml";
		if (open(FH, $configFileName))
		{
			my $line;
			my $data;
			while ($line = <FH>)
			{
				$data .= $line;
			}
			close FH;
			
			$data =~ s{(<solrVersion>.*?</solrVersion>)}{$1}sx;
			$solrVersionBlock = $1;
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $configFileName)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
			$status = 0;
		}
		
		# Read the existing cqtsadmin-<dbset>-<userdb.xml file to insert into it the "<solrVersion>" xml block
		if (open(FH, $cfgFile))
		{
			my $line;
			my $data;
			while ($line = <FH>)
			{
				$data .= $line;
			}
			close FH;
			
			if ($data =~ m/<solrVersion>.*?<\/solrVersion>/s)
			{
				# Replace existing "<solrVersion>...</solrVersion>" XML block
				$data =~ s/<solrVersion>.*?<\/solrVersion>/$solrVersionBlock/s;
			}
			else
			{
				# Insert the new "<solrVersion>" XML block
				$solrVersionBlock = "  ".$solrVersionBlock."\n\n";
				$data =~ s{(.*?)(</CQFTSSetup>)}{$1$solrVersionBlock$2}sgx;
			}
			
			printAndLog $msgLabel_INFO.$langMsgText->maketext("MODIFYING_FTS_CFG_FILE", $cfgFile)."\n";
			
			$status = safeFileUpdate($cfgFile, $data, $dieOnError);
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $cfgFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
			$status = 0;
		}
	}
	
	# Restart the WAS profile so the change takes effect
	if ($status eq 1)
	{
		$status = stopFtsWASProfile($profileName, 0);
		if ($status eq 1)
		{
			$status = startFtsWASProfile($profileName, 0);
		}
	}
	
	return $status;
}


##############################################################################
# Check to see if this upgrade requires re-indexing
sub mustReIndexSolrData
{
	my ($fromSolrVer, $toSolrVer) = @_;
	
	my $reIndexVer = $gCQSolrVersionInfo{$toSolrVer}[2];
	my @vers = @$reIndexVer;
	my $mustReIndex = 0;
	
	foreach my $ver (@vers)
	{
		$ver = trim($ver);
		if ($ver eq $fromSolrVer)
		{
			$mustReIndex = 1;
		}
	}
	
	return $mustReIndex;
}


##############################################################################
# Sets default values which can be overridden by command-line option or the 
# configuration file
sub setDefaultValues
{
	if (!exists $defValues{ "destination" })
	{
		$defValues{ "destination" } = ($gIsWindows eq 1) ? "C:/CQ.Search" : "/CQ.Search";
	}
	if (!exists $defValues{ "dbSet" })
	{
		$defValues{ "dbSet" } = "TextSearch";
	}
	if (!exists $defValues{ "userDB" })
	{
		$defValues{ "userDB" } = "SAMPL";
	}
	if (!exists $defValues{ "siteName" })
	{
		$defValues{ "siteName" } = "&lt;local&gt;";
	}
	if (!exists $defValues{ "enableAtSites" })
	{
		$defValues{ "enableAtSites" } = "";
	}
	if (!exists $defValues{ "entityFile" })
	{
		$defValues{ "entityFile" } = "";
	}
	if (!exists $defValues{ "ftsWASProfileName" })
	{
		$defValues{ "ftsWASProfileName" } = "cqsearchprofile";
	}
	if (!exists $defValues{ "ftsWASProfileHome" })
	{
		$defValues{ "ftsWASProfileHome" } = "";
	}
	if (!exists $defValues{ "ftsWASProfilePort" })
	{
		$defValues{ "ftsWASProfilePort" } = "14080";
	}
	if (!exists $defValues{ "ftsWASProfilePortsDefPropFile" })
	{
		$defValues{ "ftsWASProfilePortsDefPropFile" } = "";
	}
	if (!exists $defValues{ "ftsServerName" })
	{
		$defValues{ "ftsServerName" } = "localhost";
	}
	if (!exists $defValues{ "enableCQFTS" })
	{
		$defValues{ "enableCQFTS" } = "true";
	}
	if (!exists $defValues{ "runBatchIndex" })
	{
		$defValues{ "runBatchIndex" } = "true";
	}
	if (!exists $defValues{ "startUpdateIndex" })
	{
		$defValues{ "startUpdateIndex" } = "true";
	}
	if (!exists $defValues{ "batchIndexJVMParm" })
	{
		$defValues{ "batchIndexJVMParm" } = "";
	}
	if (!exists $defValues{ "ftsWASProfileMaxHeapSize" })
	{
		$defValues{ "ftsWASProfileMaxHeapSize" } = "";
	}
	if (!exists $defValues{ "batchIndexBatchSize" })
	{
		$defValues{ "batchIndexBatchSize" } = "250";
	}
	if (!exists $defValues{ "batchIndexDelay" })
	{
		$defValues{ "batchIndexDelay" } = "30";
	}
	if (!exists $defValues{ "batchIndexThreads" })
	{
		$defValues{ "batchIndexThreads" } = "5";
	}
	if (!exists $defValues{ "updateIndexBatchSize" })
	{
		$defValues{ "updateIndexBatchSize" } = "250";
	}
	if (!exists $defValues{ "updateIndexDelay" })
	{
		$defValues{ "updateIndexDelay" } = "30";
	}
	if (!exists $defValues{ "updateIndexLoginInterval" })
	{
		$defValues{ "updateIndexLoginInterval" } = "1";
	}
	if (!exists $defValues{ "addRecordType" })
	{
		$defValues{ "addRecordType" } = "";
	}
	if (!exists $defValues{ "removeRecordType" })
	{
		$defValues{ "removeRecordType" } = "";
	}
	if (!exists $defValues{ "deleteFtsWASProfiles" })
	{
		$defValues{ "deleteFtsWASProfiles" } = "";
	}
	if (!exists $defValues{ "deleteIndex" })
	{
		$defValues{ "deleteIndex" } = "";
	}
	if (!exists $defValues{ "enableTypeahead" })
	{
		$defValues{ "enableTypeahead" } = "";
	}
	if (!exists $defValues{ "typeAheadCacheFillTries" })
	{
		$defValues{ "typeAheadCacheFillTries" } = "";
	}
	if (!exists $defValues{ "typeAheadCacheSize" })
	{
		$defValues{ "typeAheadCacheSize" } = "";
	}
	if (!exists $defValues{ "typeAheadPageSize" })
	{
		$defValues{ "typeAheadPageSize" } = "";
	}
	if (!exists $defValues{ "solrVersion" })
	{
		$defValues{ "solrVersion" } = $gSolrVersion_120;	# Default to Solr 1.2.0 if 'solrVersion' tag is not found in cqtsadmin-<dbset>-<userdb>.xml
	}
	
	$ftsCQPerl = ($gIsWindows eq 1) ? $ftsClearQuestHomeDir."cqperl" : "cqperl";
}


##############################################################################
# Gets the user's value from the configuration file
sub getUserValue
{
	my ($key) = @_;
	
	my $value = $newValues{$key};
	
	if ((!defined $value) or (length($value) == 0))
	{
		$value = $defValues{$key};
	}
	
	if ((!defined $value) or (length($value) == 0))
	{
		return ""; # there is no value, so assume none; this maybe OK for some cases, but not all
	}
	
	return $value;
}


##############################################################################
# Navigates between directories
sub followDir
{
	my ($path, $code_ref) = @_;
	my ($DH, $file);
	
	if (-d $path)
	{
		$code_ref->($path);
		
		unless (opendir $DH, $path)
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $path)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		
		while ($file = readdir $DH)
		{
			next if $file eq '.' || $file eq '..';
			followDir("$path/$file", $code_ref);
		}
	}
	else
	{
		my $toWhere = $path;
		
		substr($toWhere, 0, length($ftsSourceDir), $ftsTargetDir);
		if (!copy($path, $toWhere))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $path, $toWhere)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
}


##############################################################################
# Copies a directory (and it's content)
sub copyDir
{
	my ($copy) = @_;
	
	substr($copy, 0, length($ftsSourceDir), $ftsTargetDir);
	
	unless (-d $copy)
	{
		if (! mkdir($copy))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $copy)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
	else
	{
		if ($copy ne $ftsDataDir)
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_ALREADY_EXIST", $copy, $ftsTargetDir)."\n", 1);
		}
	}
}


##############################################################################
# Makes a directory if it doesn't exist
sub copyDirOverwriteMkdir
{
	my ($copy) = @_;
	
	substr($copy, 0, length($ftsSourceDir), $ftsTargetDir);
	
	unless (-d $copy)
	{
		if (! mkdir($copy))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $copy)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
}


##############################################################################
# Handles an element start event of XML data parsing
sub startElement
{
	my ($parseinst, $element, %attrs) = @_;
	
	SWITCH:
	{
		if ($element eq "destination")
		{
			$tag = "destination";
			last SWITCH;
		}
		if ($element eq "dbSet")
		{
			$tag = "dbSet";
			last SWITCH;
		}
		if ($element eq "userDB")
		{
			$tag = "userDB";
			last SWITCH;
		}
		if ($element eq "siteName")
		{
			$tag = "siteName";
			last SWITCH;
		}
		if ($element eq "enableAtSites")
		{
			$tag = "enableAtSites";
			last SWITCH;
		}
		if ($element eq "entityFile")
		{
			$tag = "entityFile";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileName")
		{
			$tag = "ftsWASProfileName";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileHome")
		{
			$tag = "ftsWASProfileHome";
			last SWITCH;
		}
		if ($element eq "ftsWASProfilePort")
		{
			$tag = "ftsWASProfilePort";
			last SWITCH;
		}
		if ($element eq "ftsWASProfilePortsDefPropFile")
		{
			$tag = "ftsWASProfilePortsDefPropFile";
			last SWITCH;
		}
		if ($element eq "ftsServerName")
		{
			$tag = "ftsServerName";
			last SWITCH;
		}
		if ($element eq "enableCQFTS")
		{
			$tag = "enableCQFTS";
			last SWITCH;
		}
		if ($element eq "runBatchIndex")
		{
			$tag = "runBatchIndex";
			last SWITCH;
		}
		if ($element eq "startUpdateIndex")
		{
			$tag = "startUpdateIndex";
			last SWITCH;
		}
		if ($element eq "batchIndexJVMParm")
		{
			$tag = "batchIndexJVMParm";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileMaxHeapSize")
		{
			$tag = "ftsWASProfileMaxHeapSize";
			last SWITCH;
		}
		if ($element eq "batchIndexBatchSize")
		{
			$tag = "batchIndexBatchSize";
			last SWITCH;
		}
		if ($element eq "batchIndexDelay")
		{
			$tag = "batchIndexDelay";
			last SWITCH;
		}
		if ($element eq "batchIndexThreads")
		{
			$tag = "batchIndexThreads";
			last SWITCH;
		}
		if ($element eq "updateIndexBatchSize")
		{
			$tag = "updateIndexBatchSize";
			last SWITCH;
		}
		if ($element eq "updateIndexDelay")
		{
			$tag = "updateIndexDelay";
			last SWITCH;
		}
		if ($element eq "updateIndexLoginInterval")
		{
			$tag = "updateIndexLoginInterval";
			last SWITCH;
		}
		if ($element eq "addRecordType")
		{
			$tag = "addRecordType";
			last SWITCH;
		}
		if ($element eq "removeRecordType")
		{
			$tag = "removeRecordType";
			last SWITCH;
		}
		if ($element eq "deleteFtsWASProfiles")
		{
			$tag = "deleteFtsWASProfiles";
			last SWITCH;
		}
		if ($element eq "deleteIndex")
		{
			$tag = "deleteIndex";
			last SWITCH;
		}
		if ($element eq "enableTypeahead")
		{
			$tag = "enableTypeahead";
			last SWITCH;
		}
		if ($element eq "typeAheadCacheFillTries")
		{
			$tag = "typeAheadCacheFillTries";
			last SWITCH;
		}
		if ($element eq "typeAheadCacheSize")
		{
			$tag = "typeAheadCacheSize";
			last SWITCH;
		}
		if ($element eq "typeAheadPageSize")
		{
			$tag = "typeAheadPageSize";
			last SWITCH;
		}		
		if ($element eq "solrVersion")
		{
			$tag = "solrVersion";
			last SWITCH;
		}
		
		if ($element eq "defaultValue")
		{
			$isInDefaultValue = 1;
			last SWITCH;
		}
		if ($element eq "newValue")
		{
			$isInNewValue = 1;
			last SWITCH;
		}
	}
}


##############################################################################
# Handles an element end event of XML data parsing
sub endElement
{
	my ($parseinst, $element, %attrs) = @_;
	
	SWITCH:
	{
		if ($element eq "destination")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "dbSet")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "userDB")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "siteName")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "enableAtSites")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "entityFile")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileName")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileHome")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfilePort")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfilePortsDefPropFile")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsServerName")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "enableCQFTS")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "runBatchIndex")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "startUpdateIndex")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexJVMParm")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileMaxHeapSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexBatchSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexDelay")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexThreads")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "updateIndexBatchSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "UpdateIndexDelay")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "updateIndexLoginInterval")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "addRecordType")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "removeRecordType")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "deleteFtsWASProfiles")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "deleteIndex")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "enableTypeahead")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "typeAheadCacheFillTries")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "typeAheadCacheSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "typeAheadPageSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "solrVersion")
		{
			$tag = "";
			last SWITCH;
		}
		
		if ($element eq "defaultValue")
		{
			$isInDefaultValue = 0;
			$defaultValue = "";
			last SWITCH;
		}
		if ($element eq "newValue")
		{
			$isInNewValue = 0;
			$newValue = "";
			last SWITCH;
		}
	}
}


##############################################################################
# Handles character data event of XML data parsing
sub characterData
{
	my ($parseinst, $data) = @_;
	
	if ($isInDefaultValue == 1)
	{
		$defValues{ $tag } .= trim(HTML::Entities::decode($data));
	}
	elsif ($isInNewValue == 1)
	{
		$newValues{ $tag } .= trim(HTML::Entities::decode($data));
	}
}


##############################################################################
# Removes white characters from start and end of a string
sub trim
{
	my ($string) = @_;
	
	if (defined($string))
	{
		$string =~ s/^\s+//;
		$string =~ s/\s+$//;
	}
	
	return $string;
}


##############################################################################
# Compares the data of 2 arrays
sub compareArrays
{
	my ($a1, $a2) = @_;
	
	my @aa1 = sort(split(",", $a1));
	my @aa2 = sort(split(",", $a2));
	
	return 0 unless @aa1 == @aa2;
	
	for (my $i = 0; $i < @aa1; $i++)
	{
		return 0 if trim($aa1[$i]) ne trim($aa2[$i]);
	}
	return 1;
}


##############################################################################
# Clears an array list of Entity data
sub cleanArrayList
{
	my ($s) = @_;
	
	my @a = split(",", $s);
	my $result = "";
	
	@a = sort(@a);
	
	foreach my $i (@a)
	{
		if ($result ne "")
		{
			$result .= ",";
		}
		$result .= trim($i);
	}
	
	return $result;
}


##############################################################################
# For debugging, prints the key=value paris
sub printValues
{
	for my $key ( keys %defValues )
	{
		my $value = $defValues{$key};
		
		printAndLog "Key =======> ".$key."\n";
		printAndLog "val [def] => ".$value."\n";
		printAndLog "val [new] => ".$newValues{$key}."\n";
	}
}


##############################################################################
# Prints the small help for the tool
sub printMiniUsage
{
	my ($dieMsg, $doExit) = @_;
	
	printAndLog $dieMsg, 1;
	
	printAndLog $langMsgText->maketext("MINI_HELP_MSG", $gVersion_cqtsadmin), 1;
	
	if ($doExit)
	{
		exit $doExit
	}
}


##############################################################################
# Prints the long help for the tool
sub printLongUsage
{
	my ($dieMsg, $doExit) = @_;
	
	printAndLog $dieMsg, 1;
	
	printAndLog $langMsgText->maketext("FULL_HELP_MSG"), 1;
	
	if ($doExit)
	{
		exit $doExit
	}
}


##############################################################################
# Prints the usage for the tool
sub Usage
{
	my ($dieMsg, $doExit) = @_;
	
	printAndLog $dieMsg, 1;
	
	printMiniUsage("", 0);
	printLongUsage("", 0);
	
	if ($doExit)
	{
		exit $doExit
	}
}


##############################################################################
# Updates the Master Properties list so all needed dbname and site name combos are checked
sub updateMasterPropertiesList
{
	my (@listOfPropertyKeys) = @_;
	
	my @tempMasterPropArray = ();	
	my $isReplicated = 1;
	
	if ($ftsLocalReplicaName eq "<local>")
	{
		$isReplicated = 0;
	} 
	
	# go through the original array and copy it and other values to check into a new array
	foreach my $subValues (@listOfPropertyKeys)
	{
		push(@tempMasterPropArray, $subValues);
		
		# has user-db or site-id
		if (@$subValues[0] eq 1 or @$subValues[2] eq 1)
		{
			push(@tempMasterPropArray, [0,@$subValues[1],0]);
		}
		
		# has both
		if (@$subValues[0] eq 1 and @$subValues[2] eq 1 and $isReplicated eq 1)
		{
			push(@tempMasterPropArray, [1,@$subValues[1],0]);
			push(@tempMasterPropArray, [0,@$subValues[1],1]);
		}
	}
	
	return @tempMasterPropArray;
}